module Meqval
!======================================================================================
   use Mdef,only: dp,iu,one,zero,nfermi
   use Mlinalg,only: EigHE,EigvalsHE,util_rotate_cc
   use Mroot,only: brent
   implicit none
!--------------------------------------------------------------------------------------
   private
   public :: EqFromHam,DTrAB,GetMuChem,num_part,GetEnOcc

   interface GetEnOcc
      module procedure GetEnOcc_single, GetEnOcc_array
   end interface GetEnOcc

   interface EqFromHam
      module procedure EqFromHam_single, EqFromHam_array
   end interface EqFromHam 

   interface GetMuChem
      module procedure GetMuChem_single, GetMuChem_array
   end interface GetMuChem 
!--------------------------------------------------------------------------------------
contains
!--------------------------------------------------------------------------------------
   subroutine GetEnOcc_single(nst,Beta,Mu,H,En,occ)
      integer,intent(in)        :: nst
      real(dp),intent(in)       :: Beta,Mu
      complex(dp),intent(in)    :: H(:,:)
      real(dp),intent(inout)    :: En(:)
      real(dp),intent(inout)    :: occ(:)

      call EigvalsHE(H,En)
      occ = nfermi(Beta,En-Mu)

   end subroutine GetEnOcc_single
!--------------------------------------------------------------------------------------
   subroutine GetEnOcc_array(nst,Nk,Beta,Mu,H,En,occ)
      integer,intent(in)        :: nst,Nk
      real(dp),intent(in)       :: Beta,Mu
      complex(dp),intent(in)    :: H(:,:,:)
      real(dp),intent(inout)    :: En(:,:)
      real(dp),intent(inout)    :: occ(:,:)
      integer  :: ik

      do ik=1,Nk
         call EigvalsHE(H(:,:,ik),En(:,ik))
      end do

      occ = nfermi(Beta,En-Mu)

   end subroutine GetEnOcc_array
!--------------------------------------------------------------------------------------
   subroutine EqFromHam_single(nst,Beta,Mu,H,Rho,large_size)
      integer,intent(in)        :: nst
      real(dp),intent(in)       :: Beta,Mu
      complex(dp),intent(in)    :: H(:,:)
      complex(dp),intent(inout) :: Rho(:,:)
      logical,intent(in),optional :: large_size
      logical :: use_zgemm
      real(dp) :: En(nst)
      complex(dp) :: Q(nst,nst),RhoB(nst,nst)
      integer  :: i

      use_zgemm = .false.
      if(present(large_size)) use_zgemm = large_size

      call EigHE(H,En,Q)
      RhoB = zero
      forall(i=1:nst) RhoB(i,i) = nfermi(Beta,En(i)-Mu)

      Rho = util_rotate_cc(nst,Q,RhoB,large_size=use_zgemm)

   end subroutine EqFromHam_single
!--------------------------------------------------------------------------------------
   subroutine EqFromHam_array(nst,Nk,Beta,Mu,H,Rho,large_size)
      integer,intent(in)        :: nst,Nk
      real(dp),intent(in)       :: Beta,Mu
      complex(dp),intent(in)    :: H(:,:,:)
      complex(dp),intent(inout) :: Rho(:,:,:)
      logical,intent(in),optional :: large_size
      logical :: use_zgemm
      real(dp) :: En(nst)
      complex(dp) :: Q(nst,nst),RhoB(nst,nst)
      integer  :: i,ik

      use_zgemm = .false.
      if(present(large_size)) use_zgemm = large_size

      Rho = zero
      do ik=1,Nk
         call EigHE(H(:,:,ik),En,Q)
         RhoB = zero
         forall(i=1:nst) RhoB(i,i) = nfermi(Beta,En(i)-Mu)
         Rho(:,:,ik) = util_rotate_cc(nst,Q,RhoB,large_size=use_zgemm)
      end do

   end subroutine EqFromHam_array
!--------------------------------------------------------------------------------------
   pure real(dp) function DTrAB(A,B)
      complex(dp),intent(in) :: A(:,:),B(:,:)
      integer :: n

      n = size(A,1)
      DTrAB = real(sum(A(1:n,1:n)*transpose(B(1:n,1:n))),dp)

   end function DTrAB
!--------------------------------------------------------------------------------------
   function GetMuChem_single(Nst,Beta,H,npart,Emin,Emax,Tol) result(Mu)
      integer,intent(in)     :: Nst
      real(dp),intent(in)    :: Beta
      complex(dp),intent(in) :: H(:,:)
      real(dp),intent(in)    :: npart
      real(dp),intent(in),optional :: Emin,Emax,Tol
      real(dp) :: Mu
      real(dp) :: Emin_,Emax_,tol_=1.0e-8_dp
      real(dp) :: Eps(Nst)

      call EigvalsHE(H,Eps)
      Emin_ = minval(Eps); Emax_ = maxval(Eps)
      if(present(Emin)) Emin_ = Emin
      if(present(Emax)) Emax_ = Emax
      if(present(Tol)) tol_ = Tol

      Mu = brent(part_func,Emin_,Emax_,tol_)

   !.............................................
   contains
   !.............................................
   real(dp) function part_func(mu)
      real(dp),intent(in) :: mu
      integer :: i

      part_func = -npart
      do i=1,Nst
         part_func = part_func + nfermi(Beta,Eps(i)-mu)
      end do

   end function part_func
   !.............................................
   end function GetMuChem_single
!--------------------------------------------------------------------------------------
   real(dp) function Num_part(Rhok)
      complex(dp),intent(in) :: Rhok(:,:,:)
      integer :: nbnd,Nk,i,ik

      nbnd = size(Rhok,1)
      Nk = size(Rhok,3)
      Num_part = 0.0_dp
      do ik=1,Nk
         do i=1,nbnd
            Num_part = Num_part + dble(Rhok(i,i,ik))/Nk
         end do
      end do

   end function Num_part
!--------------------------------------------------------------------------------------
   function GetMuChem_array(Nst,Nk,Beta,H,npart,Emin,Emax,Tol) result(Mu)
      integer,intent(in)     :: Nst,Nk
      real(dp),intent(in)    :: Beta
      complex(dp),intent(in) :: H(:,:,:)
      real(dp),intent(in)    :: npart
      real(dp),intent(in),optional :: Emin,Emax,Tol
      real(dp) :: Mu
      real(dp) :: Emin_,Emax_,tol_=1.0e-8_dp
      integer :: ik
      real(dp),allocatable :: Eps(:,:)

      allocate(Eps(Nst,Nk))
      do ik=1,Nk
         call EigvalsHE(H(:,:,ik),Eps(:,ik))
      end do
      Emin_ = minval(Eps); Emax_ = maxval(Eps)
      if(present(Emin)) Emin_ = Emin
      if(present(Emax)) Emax_ = Emax
      if(present(Tol)) tol_ = Tol

      Mu = brent(part_func,Emin_,Emax_,tol_)

      deallocate(Eps)

   !.............................................
   contains
   !.............................................
   real(dp) function part_func(mu)
      real(dp),intent(in) :: mu
      integer :: i,ik

      part_func = -npart
      do ik=1,Nk
         do i=1,Nst
            part_func = part_func + nfermi(Beta,Eps(i,ik)-mu)/Nk
         end do
      end do

   end function part_func
   !.............................................
   end function GetMuChem_array
!--------------------------------------------------------------------------------------


!======================================================================================
end module Meqval