module Mgregory 
!====================================================================================
   use Mdef,only:&
      dp,zero
   use Mcoefficients,only:&
      StartPolyW, &
      OmegaW,&
      AdamsMoultonW
   use Mgrid1d,only:&
      grid1d_t
   implicit none
!------------------------------------------------------------------------------------
   private
   public :: &
      GenGridWeights,&
      GregoryIntegrate,&
      GregoryCumIntegrate
!------------------------------------------------------------------------------------
contains
!------------------------------------------------------------------------------------
   subroutine GenGregWeights(Npts,Ord,w)
      integer,intent(in)   :: Npts,Ord
      real(dp),intent(out) :: w(0:)
      !............................
      integer::r,i,j
      real(dp)::cj
      real(dp)::b(0:Npts),c(0:Npts+1)
      !............................

      r=min(Ord,Npts)

      b=0.0_dp

      b(0)=1.0_dp
      c(0)=1.0_dp
      c(1)=-0.5_dp
      b(Npts)=0.0_dp
      w=1.0_dp
      w(0)=0.5_dp
      w(Npts)=0.5_dp

      ! .. trapzoidal rule ..
      if(Npts==1) return

      ! .. Simpon's rule ..
      if(Npts==2) then
         w(0)=1.0_dp/3.0_dp
         w(1)=4.0_dp/3.0_dp
         w(2)=1.0_dp/3.0_dp
         return
      end if

      ! .. Simpon's 3/8 rule ..
      if(Npts==3) then
         w(0)=3.0_dp/8.0_dp
         w(1)=9.0_dp/8.0_dp
         w(2)=9.0_dp/8.0_dp
         w(3)=3.0_dp/8.0_dp
         return
      end if

      ! .. composite Simpon's rule ..
      if(Npts==4) then
         w(0)=1.0_dp/3.0_dp
         w(1)=4.0_dp/3.0_dp
         w(2)=2.0_dp/3.0_dp
         w(3)=4.0_dp/3.0_dp
         w(4)=1.0_dp/3.0_dp
         return
      end if

      ! .. general Gregory weights ..
      do j=1,r
         cj=0.5_dp*c(j)
         do i=j,1,-1
            b(i)=b(i)-b(i-1)
         end do
         do i=3,j+2
            cj=cj+c(j+2-i)/real(i,dp)
         end do
         c(j+1)=-cj
         do i=0,Npts
            w(i)=w(i)-cj*(b(Npts-i)+b(i))
         end do
      end do

   end subroutine GenGregWeights
!------------------------------------------------------------------------------------
   subroutine GenGridWeights(GregOrder,tmesh,weights,transp)
      integer,intent(in)::GregOrder
      type(grid1d_t),intent(in)::tmesh
      real(dp),allocatable,intent(out)::weights(:,:)
      logical,intent(in),optional :: transp
      logical :: transpose_
      integer::n,i
      real(dp),allocatable,dimension(:)::w

      transpose_ = .false.
      if(present(transp)) transpose_ = transp

      n=tmesh%Npts
      if(.not.allocated(weights)) then
         allocate(weights(0:n,0:n))
      end if

      allocate(w(0:n))
      weights=0.0_dp
      if(transpose_) then
         do i=1,n
            call GenGregWeights(i,GregOrder,w)
            weights(0:i,i)=w(0:i)
         end do
      else
         do i=1,n
            call GenGregWeights(i,GregOrder,w)
            weights(i,0:i)=w(0:i)
         end do
      end if

      weights=tmesh%dX*weights

      deallocate(w)

   end subroutine GenGridWeights
!------------------------------------------------------------------------------------
   function GregoryIntegrate(n,h,f) result(fint)
      integer,parameter :: p=5
      integer,intent(in)   :: n
      real(dp),intent(in)  :: h
      real(dp),intent(in)  :: f(0:)
      real(dp) :: fint
      integer :: j,k
      real(dp) :: w
      real(dp) :: wstart(0:2*p-2,0:p-1)
      real(dp) :: omega(0:p-1)
      real(dp) :: wam(0:p-1)

      fint = 0.0_dp
      wstart(0:p-1,0:p-1) = StartPolyW(p)
      wam = AdamsMoultonW(p)
      do k=0,p-2
         do j=0,k
            wstart(p+k,j) = wstart(p+k-1,j)
         end do
         do j=k+1,p-1
            wstart(p+k,j) = wstart(p+k-1,j) + wam(j-k-1)
         end do
      end do

      if(n <= 2*p-2) then
         do j=0,p-1
            w = wstart(n,j)
            fint = fint + w * f(j)
         end do
      else
         do j=0,p-1
            w = wstart(2*p-2,j)
            fint = fint + w * f(j)
         end do
      end if

      if(n >= p) then
         omega = OmegaW(p)
         do j=max(n-p+1,p),n
            w = omega(n-j)
            fint = fint + w * f(j)
         end do
         do j=p,n-p
            fint = fint + f(j)
         end do
      end if

      fint = h * fint

   end function GregoryIntegrate
!------------------------------------------------------------------------------------
   function GregoryCumIntegrate(nmax,h,f) result(fint)
      integer,parameter :: p=5
      integer,intent(in)   :: nmax
      real(dp),intent(in)  :: h
      real(dp),intent(in)  :: f(0:)
      real(dp) :: fint(0:nmax)
      integer :: n,j,k
      real(dp) :: w
      real(dp) :: wstart(0:2*p-2,0:p-1)
      real(dp) :: omega(0:p-1)
      real(dp) :: wam(0:p-1)


      fint = 0.0_dp
      wstart(0:p-1,0:p-1) = StartPolyW(p)
      wam = AdamsMoultonW(p)
      do k=0,p-2
         do j=0,k
            wstart(p+k,j) = wstart(p+k-1,j)
         end do
         do j=k+1,p-1
            wstart(p+k,j) = wstart(p+k-1,j) + wam(j-k-1)
         end do
      end do

      do n=0,nmax
         if(n <= 2*p-2) then
            do j=0,p-1
               w = wstart(n,j)
               fint(n) = fint(n) + w * f(j)
            end do
         else
            do j=0,p-1
               w = wstart(2*p-2,j)
               fint(n) = fint(n) + w * f(j)
            end do
         end if
         if(n >= p) then
            omega = OmegaW(p)
            do j=max(n-p+1,p),n
               w = omega(n-j)
               fint(n) = fint(n) + w * f(j)
            end do
            do j=p,n-p
               fint(n) = fint(n) + f(j)
            end do
         end if
      end do

      fint = h * fint

   end function GregoryCumIntegrate
!------------------------------------------------------------------------------------

!====================================================================================
end module Mgregory