module Mpropagate
!======================================================================================
! This module defines routines for solving the EOM for the density matrix based
! on the generalized Kadanoff-Baym ansatz (GKBA).
!======================================================================================
   use Mdebug
   use Mdef,only:&
      dp,iu,one,zero,nfermi
   use Mgrid1d,only:&
      grid1d_t
   use Mlinalg,only:&
      Eye,&
      EigHE,&
      blas_zgemm,&
      util_rotate,&
      util_rotate_cc
   use Mrungekutta,only:&
      ODE_step_RK5
   use Mevol,only:&
      GenU_CF2,&
      GenU_CF4,&
      UnitaryStepFW,&
      UnitaryStepBW,&
      UnitaryStepFBW
   !---------------------
   !    KBKIT modules
   !---------------------
   use kbkit_greenfun
   !---------------------
   implicit none
!--------------------------------------------------------------------------------------  
   private
   public &
      Propagate_OBE,&
      Timestep_OBE,&
      Timestep_Houston,&
      Propagate_Houston,&
      Timestep_Houston_ext,&
      Propagate_Houston_ext,&
      Timestep_Houston_multi,&
      Propagate_Houston_multiband,&
      Propagate_OBE_1bnd,&
      Propagate_Houston_1bnd
!--------------------------------------------------------------------------------------
contains
!--------------------------------------------------------------------------------------
   subroutine Propagate_OBE_1bnd(tgrd,feq,T1,H,Rho)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: feq
      real(dp),intent(in)                 :: T1
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      integer :: n
      real(dp) :: ft(2)

      Rho%ret(:,:,0) = Rho%Mats
      do n=0,tgrd%Npts-1
         ft(1) = dble(Rho%ret(1,1,n))
         call TimeStep_OBE_1bnd(n,tgrd,feq,T1,H,ft(1),ft(2))
         Rho%ret(1,1,n+1) = ft(2)
      end do

   end subroutine Propagate_OBE_1bnd
!--------------------------------------------------------------------------------------
   subroutine TimeStep_OBE_1bnd(tstp,tgrd,feq,T1,H,ft,ft_new)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: feq
      real(dp),intent(in)              :: T1
      type(CntrFun_1time_t),intent(in) :: H
      real(dp),intent(in)              :: ft
      real(dp),intent(out)             :: ft_new

      ft_new = ODE_step_RK5(tstp,tgrd%dX,ft_deriv,ft)
      !......................................................
      contains
      !......................................................
      function ft_deriv(t,yt) result(dydt)
         real(dp),intent(in) :: t 
         real(dp),intent(in) :: yt
         real(dp) :: dydt

         dydt = -0.5_dp * (yt - feq) / T1

      end function ft_deriv
      !......................................................

   end subroutine TimeStep_OBE_1bnd
!--------------------------------------------------------------------------------------
   subroutine Propagate_OBE(tgrd,feq,T1,T2,H,Rho,band_basis)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: feq(:)
      real(dp),intent(in)                 :: T1,T2
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      logical,intent(in),optional         :: band_basis
      logical :: band_basis_
      integer :: n
      real(dp),allocatable :: eps0(:)
      complex(dp),allocatable :: vect0(:,:)

      band_basis_ = .true.
      if(present(band_basis)) band_basis_ = band_basis

      Rho%ret(:,:,0) = Rho%Mats
      if(band_basis_) then
         do n=0,tgrd%Npts-1
            call TimeStep_OBE(n,tgrd,feq,T1,T2,H,Rho%ret(:,:,n),Rho%ret(:,:,n+1))
         end do
      else
         allocate(eps0(H%Nstates),vect0(H%Nstates,H%Nstates))
         call EigHE(H%Mats,eps0,vect0)
         do n=0,tgrd%Npts-1
            call TimeStep_OBE(n,tgrd,feq,T1,T2,H,Rho%ret(:,:,n),Rho%ret(:,:,n+1),rot_mat=vect0)
         end do
         deallocate(eps0,vect0)
      end if

   end subroutine Propagate_OBE
!--------------------------------------------------------------------------------------
   subroutine TimeStep_OBE(tstp,tgrd,feq,T1,T2,H,Rho,Rho_new,rot_mat)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: feq(:)
      real(dp),intent(in)              :: T1,T2
      type(CntrFun_1time_t),intent(in) :: H
      complex(dp),intent(in)           :: Rho(:,:)
      complex(dp),intent(inout)        :: Rho_new(:,:)
      complex(dp),intent(in),optional  :: rot_mat(:,:)
      logical :: rot_basis
      integer :: nst
      real(dp) :: g12

      nst = H%Nstates
      g12 = 0.5_dp / T1 + 1.0_dp / T2

      rot_basis = present(rot_mat)

      Rho_new = ODE_step_RK5(nst,tstp,tgrd%dX,deriv,Rho)
      !......................................................
      contains
      !......................................................
      function deriv(nst,t,yt) result(dydt)
         integer,intent(in) :: nst
         real(dp),intent(in) :: t 
         complex(dp),intent(in) :: yt(:,:)
         complex(dp) :: dydt(nst,nst)
         integer :: i,j
         complex(dp),dimension(nst,nst) :: h0_t,RhoB,scatt_B,scatt_H

         call H%Interpolate(tgrd,t,h0_t)
         dydt = -iu*(matmul(h0_t,yt) - matmul(yt,h0_t))

         if(rot_basis) then
            rhoB = util_rotate(nst,rot_mat,yt,large_size=.false.)
         else
            rhoB = yt
         end if

         scatt_B(1,1) = - 0.5_dp * (rhoB(1,1) - feq(1)) / T1
         scatt_B(2,2) = - 0.5_dp * (rhoB(2,2) - feq(2)) / T1
         scatt_B(1,2) = - g12 * rhoB(1,2)
         scatt_B(2,1) = - g12 * rhoB(2,1)

         scatt_H = util_rotate_cc(nst,rot_mat,scatt_B,large_size=.false.)

         dydt = dydt + scatt_H

      end function deriv
      !......................................................

   end subroutine TimeStep_OBE
!--------------------------------------------------------------------------------------
   subroutine Propagate_Houston_1bnd(tgrd,Beta,Mu,T1,H,Rho)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: Beta,Mu
      real(dp),intent(in)                 :: T1
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      integer :: n
      real(dp) :: ft(2)

      Rho%ret(:,:,0) = Rho%Mats
      do n=0,tgrd%Npts-1
         ft(1) = dble(Rho%ret(1,1,n))
         call Timestep_Houston_1bnd(n,tgrd,Beta,Mu,T1,H,ft(1),ft(2))
         Rho%ret(1,1,n+1) = ft(2)
      end do

   end subroutine Propagate_Houston_1bnd
!--------------------------------------------------------------------------------------
   subroutine TimeStep_Houston_1bnd(tstp,tgrd,Beta,Mu,T1,H,ft,ft_new)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: Beta,Mu
      real(dp),intent(in)              :: T1
      type(CntrFun_1time_t),intent(in) :: H
      real(dp),intent(in)              :: ft
      real(dp),intent(out)             :: ft_new

      ft_new = ODE_step_RK5(tstp,tgrd%dX,deriv,ft)
      !......................................................
      contains
      !......................................................
      function deriv(t,yt) result(dydt)
         real(dp),intent(in) :: t 
         real(dp),intent(in) :: yt
         real(dp) :: dydt
         complex(dp) :: h0_t(1,1)
         real(dp) :: eps_t,feq

         call H%Interpolate(tgrd,t,h0_t)
         eps_t = dble(h0_t(1,1))
         feq = nfermi(beta,eps_t - mu)

         dydt = -(yt - feq) / T1

      end function deriv
      !......................................................

   end subroutine TimeStep_Houston_1bnd
!--------------------------------------------------------------------------------------
   subroutine Propagate_Houston(tgrd,Beta,Mu,T1,T2,H,Rho)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: Beta,Mu
      real(dp),intent(in)                 :: T1,T2
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      integer :: n

      Rho%ret(:,:,0) = Rho%Mats
      do n=0,tgrd%Npts-1
         call Timestep_Houston(n,tgrd,Beta,Mu,T1,T2,H,Rho%ret(:,:,n),Rho%ret(:,:,n+1))
      end do

   end subroutine Propagate_Houston
!--------------------------------------------------------------------------------------
   subroutine Timestep_Houston(tstp,tgrd,beta,mu,T1,T2,H,Rho,Rho_new)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: beta,mu
      real(dp),intent(in)              :: T1,T2
      type(CntrFun_1time_t),intent(in) :: H
      complex(dp),intent(in)           :: Rho(:,:)
      complex(dp),intent(inout)        :: Rho_new(:,:)
      integer :: nst

      nst = H%Nstates

      Rho_new = ODE_step_RK5(nst,tstp,tgrd%dX,deriv,Rho)
      !......................................................
      contains
      !......................................................
      function deriv(nst,t,yt) result(dydt)
         integer,intent(in) :: nst
         real(dp),intent(in) :: t 
         complex(dp),intent(in) :: yt(:,:)
         complex(dp) :: dydt(nst,nst)
         integer :: i,j
         real(dp),dimension(nst) :: eps_t,feq
         complex(dp),dimension(nst,nst) :: h0_t,vect_t,rho_H
         complex(dp),dimension(nst,nst) :: scatt_H,scatt_B

         call H%Interpolate(tgrd,t,h0_t)
         dydt = -iu*(matmul(h0_t,yt) - matmul(yt,h0_t))

         call EigHE(h0_t,eps_t,vect_t)
         feq = nfermi(beta,eps_t - mu)

         rho_H = util_rotate(nst,vect_t,yt,large_size=.false.)
         scatt_H(1,1) = (rho_H(1,1) - feq(1)) / T1
         scatt_H(2,2) = (rho_H(2,2) - feq(2)) / T1
         scatt_H(1,2) = rho_H(1,2) / T2
         scatt_H(2,1) = rho_H(2,1) / T2

         scatt_B = util_rotate_cc(nst,vect_t,scatt_H,large_size=.false.)

         dydt = dydt - scatt_B

      end function deriv
      !......................................................

   end subroutine Timestep_Houston
!--------------------------------------------------------------------------------------
   subroutine Timestep_Houston_ext(tstp,tgrd,beta,mu,Tmat,H,Rho,Rho_new,large_size)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: beta,mu
      real(dp),intent(in)              :: Tmat(:,:)
      type(CntrFun_1time_t),intent(in) :: H
      complex(dp),intent(in)           :: Rho(:,:)
      complex(dp),intent(inout)        :: Rho_new(:,:)
      logical,intent(in),optional      :: large_size
      logical :: large_size_
      integer :: nst

      large_size_ = .false.
      if(present(large_size)) large_size_ = large_size

      nst = H%Nstates

      Rho_new = ODE_step_RK5(nst,tstp,tgrd%dX,deriv,Rho)
      !......................................................
      contains
      !......................................................
      function deriv(nst,t,yt) result(dydt)
         integer,intent(in) :: nst
         real(dp),intent(in) :: t 
         complex(dp),intent(in) :: yt(:,:)
         complex(dp) :: dydt(nst,nst)
         integer :: i,j
         real(dp),dimension(nst) :: eps_t,feq
         complex(dp),dimension(nst,nst) :: h0_t,vect_t,rho_H
         complex(dp),dimension(nst,nst) :: scatt_H,scatt_B

         call H%Interpolate(tgrd,t,h0_t)
         dydt = -iu*(matmul(h0_t,yt) - matmul(yt,h0_t))

         call EigHE(h0_t,eps_t,vect_t)
         feq = nfermi(beta,eps_t - mu)

         rho_H = util_rotate(nst,vect_t,yt,large_size=large_size_)
         do i=1,nst
            scatt_H(i,i) = (rho_H(i,i) - feq(i)) / Tmat(i,i)
         end do

         do i=1,nst
            do j=1,nst
               if(i == j) cycle
               scatt_H(i,j) = rho_H(i,j) / Tmat(i,j)
            end do
         end do

         scatt_B = util_rotate_cc(nst,vect_t,scatt_H,large_size=large_size_)

         dydt = dydt - scatt_B

      end function deriv
      !......................................................

   end subroutine Timestep_Houston_ext
!--------------------------------------------------------------------------------------
   subroutine Propagate_Houston_ext(tgrd,Beta,Mu,Tmat,H,Rho)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: Beta,Mu
      real(dp),intent(in)                 :: Tmat(:,:)
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      integer :: n

      Rho%ret(:,:,0) = Rho%Mats
      do n=0,tgrd%Npts-1
         call Timestep_Houston_ext(n,tgrd,Beta,Mu,Tmat,H,Rho%ret(:,:,n),Rho%ret(:,:,n+1),&
            large_size=(H%Nstates >= 8))
      end do

   end subroutine Propagate_Houston_ext
!--------------------------------------------------------------------------------------
   subroutine Timestep_Houston_multi(tstp,tgrd,beta,mu,T1,T2,H,Rho,Rho_new,large_size)
      integer,intent(in)               :: tstp
      type(grid1d_t),intent(in)        :: tgrd
      real(dp),intent(in)              :: beta,mu
      real(dp),intent(in)              :: T1,T2
      type(CntrFun_1time_t),intent(in) :: H
      complex(dp),intent(in)           :: Rho(:,:)
      complex(dp),intent(inout)        :: Rho_new(:,:)
      logical,intent(in),optional      :: large_size
      logical :: large_size_
      integer :: nst
      real(dp) :: gm1,gm2,gm12

      large_size_ = .false.
      if(present(large_size)) large_size_ = large_size

      nst = H%Nstates

      gm1 = 1.0_dp / T1
      gm2 = 1.0_dp / T2
      gm12 = gm1 - gm2

      Rho_new = ODE_step_RK5(nst,tstp,tgrd%dX,deriv,Rho)
      !......................................................
      contains
      !......................................................
      function deriv(nst,t,yt) result(dydt)
         integer,intent(in) :: nst
         real(dp),intent(in) :: t 
         complex(dp),intent(in) :: yt(:,:)
         complex(dp) :: dydt(nst,nst)
         integer :: i,j
         real(dp),dimension(nst) :: eps_t,feq
         complex(dp),dimension(nst,nst) :: h0_t,vect_t,rho_H,ht_yt,yt_ht
         complex(dp),dimension(nst,nst) :: scatt_H,scatt_B

         call H%Interpolate(tgrd,t,h0_t)

         call EigHE(h0_t,eps_t,vect_t)
         feq = nfermi(beta,eps_t - mu)

         rho_H = util_rotate(nst,vect_t,yt,large_size=large_size_)
         do i=1,nst
            scatt_H(i,i) = - gm1 * (rho_H(i,i) - feq(i))
         end do

         do i=1,nst
            do j=1,nst
               if(i == j) cycle
               scatt_H(i,j) = gm12 * rho_H(i,j)
            end do
         end do

         scatt_B = util_rotate_cc(nst,vect_t,scatt_H,large_size=large_size_)
         ht_yt = matmul(h0_t,yt)
         yt_ht = matmul(yt,h0_t)
         dydt = -iu * (ht_yt - yt_ht) + scatt_B

      end function deriv
      !......................................................

   end subroutine Timestep_Houston_multi
!--------------------------------------------------------------------------------------
   subroutine Propagate_Houston_multiband(tgrd,Beta,Mu,T1,T2,H,Rho)
      type(grid1d_t),intent(in)           :: tgrd
      real(dp),intent(in)                 :: Beta,Mu
      real(dp),intent(in)                 :: T1,T2
      type(CntrFun_1time_t),intent(in)    :: H
      type(CntrFun_1time_t),intent(inout) :: Rho
      integer :: n

      Rho%ret(:,:,0) = Rho%Mats
      do n=0,tgrd%Npts-1
         call Timestep_Houston_multi(n,tgrd,Beta,Mu,T1,T2,H,Rho%ret(:,:,n),Rho%ret(:,:,n+1),&
            large_size=(H%Nstates >= 8))
      end do

   end subroutine Propagate_Houston_multiband
!--------------------------------------------------------------------------------------


!======================================================================================
end module Mpropagate
