module kbkit_gkba
!======================================================================================
! This module defines routines for solving the EOM for the density matrix based
! on the generalized Kadanoff-Baym ansatz (GKBA).
!======================================================================================
  use Mdebug
  use Mdef,only:&
       dp,iu,one,zero,nfermi,large_size
  use Mlinalg,only:&
       EigHE,&
       Eye,&
       Trace,&
       blas_zgemm,&
       util_zgemm,&
       util_rotate,&
       util_rotate_cc
  use Mgrid1d,only:&
       grid1d_t
  use Mrungekutta,only:&
       ODE_step_RK5
  use Mcoefficients,only:&
       CAdBash,&
       CAdMoul
  use Mevol,only:&
       UnitaryStepFW,&
       UnitaryStepBW,&
       UnitaryStepFBW,&
       GenU_CF2,&
       GenU_CF4
  use kbkit_greenfun,only:&
       CntrFun_1time_t,&
       CntrFun_2time_t,&
       CntrFun_timeslice_t
  use kbkit_integration,only:&
       CntrIntegrals_t,&
       GregoryIntegral
  implicit none
!--------------------------------------------------------------------------------------
  private
  public &
       InitIntegrator,&
       gkba_t,&
       GKBA_Start,&
       GKBA_Predictor,&
       GKBA_Corrector,&
       GKBA_Step_AdamsMoulton,&
       GKBA_Step_AdamsMoulton_direct,&
       Init_GKBA_array,&
       Clean_GKBA_array
!-------------------------------------------------------------------------------------
  type gkba_t
     logical,private              :: dephasing_mode=.false.
     integer,public               :: nst               !! number of states (matrix rank)
     integer,public               :: Nt                !! number of time points
     integer,private              :: order=5           !! order of method (quadrature/integration)
     integer,dimension(2),public  :: stencil2p=[1,2]   !! internal two-point stencil
     integer,allocatable,public   :: stencil(:)        !! internal stencil
     real(dp),public              :: dt                !! step size
     real(dp),allocatable,private :: Tdeph_mat(:,:)
     complex(dp),allocatable      :: H(:,:,:)          !! Hamiltonian (MF), saved over stencil
     complex(dp),allocatable      :: C(:,:,:)          !! [H, Rho], saved over stencil
     complex(dp),allocatable      :: Rho(:,:,:)        !! density matrix (with memory)
     complex(dp),allocatable      :: Uprop(:,:,:,:)    !! time evolution operator
     complex(dp),allocatable      :: CollInt(:,:,:)    !! collision integral
   contains
     procedure,public  :: Init => Init_GKBA
     procedure,public  :: Clean => Clean_GKBA
     procedure,public  :: SetEqFromEps => GKBA_EqFromEps
     procedure,public  :: SetEqFromHam => GKBA_EqFromHam
     procedure,public  :: SetDephasing => GKBA_SetDephasing
     procedure,public  :: UpdateH => GKBA_UpdateH
     procedure,public  :: UpdateC => GKBA_UpdateC
     procedure,public  :: FillStart => GKBA_FillStart
     procedure,public  :: ExpectVal => GKBA_ExpectVal
     procedure,public  :: GetLesser => GKBA_GetLesser
     procedure,public  :: GetGreater => GKBA_GetGreater
     procedure,public  :: Ecorr => GKBA_Ecorr
     procedure,public  :: Reconstruct_Green => GKBA_Reconstruct_Green
     procedure,public  :: Reconstruct_Green_step => GKBA_Reconstruct_Green_step
     procedure,public  :: GetU_NewStep => GKBA_GetU_new
     procedure,public  :: GetU_OldStep => GKBA_GetU_old
     procedure,public  :: SetU_NewStep => GKBA_SetU_new
     procedure,public  :: SetU_OldStep => GKBA_SetU_old
     procedure,public  :: SwitchToCorrector => ShiftStencil
     procedure,public  :: SwitchToNextStep => ShiftStencil2p
     procedure,public  :: GenCollInt => Gen_GKBA_CollInt
  end type gkba_t
!--------------------------------------------------------------------------------------
   type(CntrIntegrals_t) :: Cntrinteg 
!--------------------------------------------------------------------------------------
contains
!--------------------------------------------------------------------------------------
   subroutine InitIntegrator(GregorgyOrder,t_grid)
      integer,intent(in) :: GregorgyOrder
      type(grid1d_t),intent(in) :: t_grid

      call Cntrinteg%Init(GregorgyOrder,t_grid)

   end subroutine InitIntegrator
!--------------------------------------------------------------------------------------
   subroutine Zmult(n,A,B,C)
      integer,intent(in) :: n
      complex(dp),intent(in) :: A(:,:),B(:,:)
      complex(dp),intent(inout) :: C(:,:)

      if(n <= large_size) then
         C(1:n,1:n) = matmul(A(1:n,1:n),B(1:n,1:n))
      else
         call util_zgemm(A(1:n,1:n), B(1:n,1:n), C(1:n,1:n))
      end if

   end subroutine Zmult
!--------------------------------------------------------------------------------------
   subroutine ZMM(n,al,bt,A,B,C)
      integer,intent(in) :: n
      complex(dp),intent(in) :: al,bt
      complex(dp),intent(in) :: A(:,:),B(:,:)
      complex(dp),intent(inout) :: C(:,:)

      if(n <= large_size) then
         C(1:n,1:n) = al * matmul(A(1:n,1:n),B(1:n,1:n)) + bt * C(1:n,1:n)
      else
         call blas_zgemm('n','n',n,n,n,al,A,n,B,n,bt,C,n)
      end if

   end subroutine ZMM
!--------------------------------------------------------------------------------------
   function HermConjg(A) result(Acc)
      complex(dp),intent(in) :: A(:,:)
      complex(dp) :: Acc(size(A,dim=2),size(A,dim=1))

      Acc = conjg(transpose(A))

   end function HermConjg
!--------------------------------------------------------------------------------------
   subroutine ZHEpart(n,A,Ah)
      integer,intent(in) :: n
      complex(dp),intent(in) :: A(:,:)
      complex(dp),intent(inout) :: Ah(:,:)

      Ah(1:n,1:n) = A(1:n,1:n) + HermConjg(A(1:n,1:n))

   end subroutine ZHEpart
!--------------------------------------------------------------------------------------
  subroutine Init_GKBA(gkba,nst,tgrid,order)
    class(gkba_t)                    :: gkba
    integer,intent(in)               :: nst
    type(grid1d_t),intent(in)        :: tgrid
    integer,intent(in),optional      :: order
    integer :: i

    if(present(order)) gkba%order = order
    gkba%order = min(gkba%order,6)

    gkba%nst = nst
    gkba%Nt = tgrid%Npts
    gkba%dt = tgrid%dX

    allocate(gkba%H(nst,nst,gkba%order)) ; gkba%H=zero
    allocate(gkba%C(nst,nst,gkba%order)) ; gkba%C=zero
    allocate(gkba%Rho(nst,nst,0:gkba%Nt)) ; gkba%Rho=zero
    allocate(gkba%Uprop(nst,nst,0:gkba%Nt,2)) ; gkba%Uprop=zero
    allocate(gkba%CollInt(nst,nst,gkba%order)) ; gkba%CollInt=zero

    allocate(gkba%stencil(gkba%order))
    gkba%stencil = [(i,i=1,gkba%order)]

  end subroutine Init_GKBA
!--------------------------------------------------------------------------------------
  subroutine Clean_GKBA(gkba)
    class(gkba_t)::gkba

    if(allocated(gkba%H)) deallocate(gkba%H)
    if(allocated(gkba%C)) deallocate(gkba%C)
    if(allocated(gkba%Rho)) deallocate(gkba%Rho)
    if(allocated(gkba%Uprop)) deallocate(gkba%Uprop)
    if(allocated(gkba%CollInt)) deallocate(gkba%CollInt)
    if(allocated(gkba%stencil)) deallocate(gkba%stencil)

  end subroutine Clean_GKBA
!--------------------------------------------------------------------------------------
  subroutine GKBA_SetDephasing(gkba,Tdeph,Tdeph_mat)
    class(gkba_t) :: gkba
    real(dp),intent(in),optional :: Tdeph
    real(dp),intent(in),optional :: Tdeph_mat(:,:)
    integer :: i

    allocate(gkba%Tdeph_mat(gkba%nst,gkba%nst)); gkba%Tdeph_mat = 1.0e10_dp

    if(present(Tdeph)) then
        gkba%Tdeph_mat = Tdeph
    end if

    if(present(Tdeph_mat)) then
        gkba%Tdeph_mat(1:gkba%nst,1:gkba%nst) = Tdeph_mat(1:gkba%nst,1:gkba%nst)
    end if

    gkba%dephasing_mode = .true.

  end subroutine GKBA_SetDephasing
!--------------------------------------------------------------------------------------
  subroutine GKBA_EqFromEps(gkba,Beta,Mu,Eps)
    class(gkba_t)::gkba
    real(dp),intent(in)::Beta,Mu,Eps(:)
    integer::i

    gkba%Rho(:,:,0) = zero
    forall(i=1:gkba%nst) gkba%Rho(i,i,0)=nfermi(Beta,Eps(i)-Mu)

  end subroutine GKBA_EqFromEps
!--------------------------------------------------------------------------------------
  subroutine GKBA_EqFromHam(gkba,Beta,Mu,H)
    class(gkba_t)::gkba
    real(dp),intent(in)::Beta,Mu
    complex(dp),intent(in)::H(:,:)
    integer::i
    real(dp),allocatable::En(:)
    complex(dp),allocatable::Q(:,:),Qcc(:,:)

    allocate(En(gkba%nst),Q(gkba%nst,gkba%nst),Qcc(gkba%nst,gkba%nst))

    call EigHE(H,En,Q)
    Qcc=HermConjg(Q)
    gkba%Rho(:,:,0)=zero
    do i=1,gkba%nst
       gkba%Rho(i,i,0)=nfermi(Beta,En(i)-Mu)
    end do
    gkba%Rho(:,:,0)=matmul(Q,matmul(gkba%Rho(:,:,0),Qcc))

    deallocate(En,Q,Qcc)

  end subroutine GKBA_EqFromHam
!--------------------------------------------------------------------------------------
  pure subroutine ShiftStencil(gkba)
    class(gkba_t),intent(inout)::gkba

    select case(gkba%order)
    case(2)
       gkba%Stencil=&
            [gkba%Stencil(2),&
            gkba%Stencil(1)]
    case(3)
       gkba%Stencil=&
            [gkba%Stencil(3),&
            gkba%Stencil(1),&
            gkba%Stencil(2)]
    case(4)
       gkba%Stencil=&
            [gkba%Stencil(4),&
            gkba%Stencil(1),&
            gkba%Stencil(2),&
            gkba%Stencil(3)]
    case(5)
       gkba%Stencil=&
            [gkba%Stencil(5),&
            gkba%Stencil(1),&
            gkba%Stencil(2),&
            gkba%Stencil(3),&
            gkba%Stencil(4)]
    case(6)
       gkba%Stencil=&
            [gkba%Stencil(6),&
            gkba%Stencil(1),&
            gkba%Stencil(2),&
            gkba%Stencil(3),&
            gkba%Stencil(4),&
            gkba%Stencil(5)]
    end select

  end subroutine ShiftStencil
!--------------------------------------------------------------------------------------
  pure subroutine ShiftStencil2p(gkba)
    class(gkba_t),intent(inout)::gkba

    gkba%Stencil2p = [gkba%Stencil2p(2),gkba%Stencil2p(1)]

  end subroutine ShiftStencil2p
!--------------------------------------------------------------------------------------
  pure subroutine GKBA_UpdateH(gkba,H)
    class(gkba_t),intent(inout)::gkba
    complex(dp),intent(in)::H(:,:)

    gkba%H(1:gkba%nst,1:gkba%nst,gkba%stencil(1)) = H(1:gkba%nst,1:gkba%nst)

  end subroutine GKBA_UpdateH
!--------------------------------------------------------------------------------------
  subroutine GKBA_UpdateC(gkba,tstp,large_size)
    class(gkba_t),intent(inout) :: gkba
    integer,intent(in)          :: tstp
    logical,intent(in),optional :: large_size
    logical :: use_zgemm

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    if(use_zgemm) then
      call blas_zgemm('n','n',gkba%nst,gkba%nst,gkba%nst,-iu,gkba%H(:,:,gkba%stencil(1)),&
        gkba%nst,gkba%Rho(:,:,tstp),gkba%nst,zero,gkba%C(:,:,gkba%stencil(1)),gkba%nst)
      call blas_zgemm('n','n',gkba%nst,gkba%nst,gkba%nst,iu,gkba%Rho(:,:,tstp),&
        gkba%nst,gkba%H(:,:,gkba%stencil(1)),gkba%nst,one,gkba%C(:,:,gkba%stencil(1)),gkba%nst)
    else
      gkba%C(:,:,gkba%stencil(1)) = -iu * &
        (matmul(gkba%H(:,:,gkba%stencil(1)),gkba%Rho(:,:,tstp)) &
        - matmul(gkba%Rho(:,:,tstp),gkba%H(:,:,gkba%stencil(1))) )
    end if
 
  end subroutine GKBA_UpdateC
!--------------------------------------------------------------------------------------
  subroutine AddDephasing(nst,Rho,H,Tdeph_mat,IC,large_size)
    integer,intent(in)          :: nst
    complex(dp),intent(in)      :: Rho(:,:)
    complex(dp),intent(in)      :: H(:,:)
    real(dp),intent(in)         :: Tdeph_mat(:,:)
    complex(dp),intent(inout)   :: IC(:,:)
    logical,intent(in),optional :: large_size
    logical :: use_zgemm
    integer :: i,j
    real(dp),dimension(nst) :: eps
    complex(dp),dimension(nst,nst) :: vect,rho_H,scatt_H,scatt_B

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    call EigHE(H,eps,vect)
    rho_H = util_rotate(nst,vect,Rho,large_size=use_zgemm)

    scatt_H = zero
    do i=1,nst
        do j=1,nst
            if(i == j) cycle
            scatt_H(i,j) = - rho_H(i,j) / Tdeph_mat(i,j)
        end do
    end do

    scatt_B = util_rotate_cc(nst,vect,scatt_H,large_size=use_zgemm)
    IC = IC + scatt_B

  end subroutine AddDephasing
!--------------------------------------------------------------------------------------
  subroutine GKBA_FillStart(gkba,H)
    class(gkba_t)           :: gkba
    complex(dp),intent(in)  :: H(:,:)
    integer :: it,jt,k
    complex(dp),dimension(gkba%nst,gkba%nst) :: Udt,Uold

    gkba%CollInt = zero


    do k=1,gkba%order
       gkba%H(:,:,gkba%stencil(k)) = H
    end do

    call GenU_CF2(gkba%dt,H,Udt)

    do it=1,gkba%order
       gkba%Rho(:,:,it) = zero
       call UnitaryStepFBW(gkba%nst,Udt,gkba%Rho(:,:,it-1),gkba%Rho(:,:,it))
    end do

    do jt=0,gkba%order-2
       call GenU_CF2((gkba%order-1-jt)*gkba%dt,H,Uold)
       call gkba%SetU_oldstep(jt,Uold)
       call gkba%SetU_newstep(jt,Uold)
    end do

    Uold=Eye(gkba%nst)
    call gkba%SetU_oldstep(gkba%order-1,Uold)
    call gkba%SetU_newstep(gkba%order-1,Uold)

  end subroutine GKBA_FillStart
!--------------------------------------------------------------------------------------
  subroutine GKBA_start(gkba,H,SGM,G,Err,first_iter)
    integer,parameter  :: MaxIter=10
    real(dp),parameter :: c1 = 0.5_dp - sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: c2 = 0.5_dp + sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: ErrTol=1.0e-14_dp
    type(gkba_t),intent(inout)          :: gkba !! GKBA type to propagate
    type(CntrFun_1time_t),intent(in)    :: H    !! Hamiltonian, initialized 0..p-1
    type(CntrFun_2time_t),intent(in)    :: SGM  !! self-energy, initialized 0..p-1,0..p-1
    type(CntrFun_2time_t),intent(inout) :: G    !! Green's function, initialized 0..p-1,0..p-1
    real(dp),intent(out)                :: Err  !! norm difference of density matrix between iterations
    logical,intent(in),optional         :: first_iter !! if true, initial guess will be applied
    logical :: init_rho
    integer :: n,j,n1,n2,iter
    real(dp) :: t1,t2,err_conv
    complex(dp),dimension(gkba%nst,gkba%nst) :: H1,H2,Udt,XR,XL,xrho,ICj
    complex(dp),allocatable,dimension(:,:,:) :: Rho_old,Rho_tmp,IC
    complex(dp),allocatable,dimension(:,:,:,:) :: Ut

    call assert(gkba%nst == H%Nstates)
    call assert(gkba%nst == SGM%Nstates)

    init_rho = .false.
    if(present(first_iter)) init_rho = first_iter

    if(init_rho) then
       call gkba%FillStart(H%Ret(:,:,0))
    end if

    ! .. copy density matrix for error assessment ..
    allocate(Rho_old(gkba%nst,gkba%nst,0:gkba%order-1))
    Rho_old = gkba%Rho(:,:,0:gkba%order-1)

     ! .. compute time evolution operator ..
    allocate(Ut(gkba%nst,gkba%nst,0:gkba%order-1,0:gkba%order-1))
    do n=0,gkba%order-1
       Ut(:,:,n,n) = Eye(gkba%nst)
    end do
    do n=0,gkba%order-2
       t1 = (n + c1) * gkba%dt
       t2 = (n + c2) * gkba%dt
       forall(n1=1:gkba%nst,n2=1:gkba%nst)
          H1(n1,n2) = Interpol1D(gkba%order-1,gkba%dt,H%Ret(n1,n2,0:),t1)
          H2(n1,n2) = Interpol1D(gkba%order-1,gkba%dt,H%Ret(n1,n2,0:),t2)
       end forall
       call GenU_CF4(gkba%dt,H1,H2,Udt)
       Ut(:,:,n+1,n) = matmul(Udt,Ut(:,:,n,n))
    end do
    do n=2,gkba%order-1
       do j=0,n-2
          Ut(:,:,n,j) = matmul(Ut(:,:,n,n-1),Ut(:,:,n-1,j))
       end do
    end do

    do n=1,gkba%order-1
       do j=0,n-1
          Ut(:,:,j,n) = HermConjg(Ut(:,:,n,j))
       end do
    end do

    allocate(IC(gkba%nst,gkba%nst,0:gkba%order-1)); IC = zero
    allocate(Rho_tmp(gkba%nst,gkba%nst,0:gkba%order-1))

    ITERLOOP:do iter=1,MaxIter

       Rho_tmp(1:gkba%nst,1:gkba%nst,0:gkba%order-1) = gkba%rho(1:gkba%nst,1:gkba%nst,0:gkba%order-1)

       ! .. compute collision integrals ..
       do n=1,gkba%order-1
          ICj = GregoryIntegral(gkba%nst,n,CollInt_term_start,Cntrinteg)
          IC(:,:,n) = ICj + HermConjg(ICj)
          if(gkba%dephasing_mode) then
            call AddDephasing(gkba%nst,Rho_tmp(:,:,n),H%ret(:,:,n),gkba%Tdeph_mat,IC(:,:,n),&
                large_size=.false.)
          end if
       end do

       ! .. propagate with Runge-Kutta-Fehlberg ..
       do n=0,gkba%order-2
          gkba%Rho(:,:,n+1) = ODE_step_RK5(gkba%nst,n,gkba%dt,RhoEOM,gkba%Rho(:,:,n))
       end do

       ! .. check error ..
       err_conv = sum(abs(Rho_tmp - gkba%rho(1:gkba%nst,1:gkba%nst,0:gkba%order-1)))
       if(err_conv < ErrTol) exit ITERLOOP
       
    end do ITERLOOP

    ! .. update Green's functions ..
    do n=0,gkba%order-1
       do j=0,n
          xrho = Eye(gkba%nst) - gkba%Rho(:,:,j)
          G%Lesser(:,:,n,j) = iu*matmul(Ut(:,:,n,j),gkba%Rho(:,:,j))
          G%Greater(:,:,n,j) = -iu*matmul(Ut(:,:,n,j),xrho)
          G%Lesser(:,:,j,n) = -HermConjg(G%Lesser(:,:,n,j))
          G%Greater(:,:,j,n) = -HermConjg(G%Greater(:,:,n,j))
       end do
    end do

    do j=0,gkba%order-1
       call gkba%SetU_oldstep(j,Ut(:,:,gkba%order-1,j))
       call gkba%SetU_newstep(j,Ut(:,:,gkba%order-1,j))
    end do

    do n=0,gkba%order-1
       gkba%CollInt(:,:,n+1) = IC(:,:,gkba%order-1-n)
       gkba%H(:,:,n+1) = H%Ret(:,:,gkba%order-1-n)
       gkba%C(:,:,n+1) = -iu * (matmul(H%Ret(:,:,gkba%order-1-n),gkba%Rho(:,:,gkba%order-1-n)) &
          - matmul(gkba%Rho(:,:,gkba%order-1-n),H%Ret(:,:,gkba%order-1-n))) 
    end do

    ! .. check error ..
    Err = 0.0_dp
    do n=1,gkba%order-1
       Err = Err + sum(abs(gkba%Rho(:,:,n)-Rho_old(:,:,n)))
    end do

    ! .. clean up ..
    deallocate(Rho_old,Ut,IC,Rho_tmp)
  !.................................................
  contains
  !.................................................
    function RhoEOM(nst,t,rho) result(drhodt)
      integer,intent(in)     :: nst
      real(dp),intent(in)    :: t
      complex(dp),intent(in) :: rho(:,:)
      complex(dp)            :: drhodt(nst,nst)
      integer :: n1,n2
      complex(dp) :: Ht(nst,nst),ICt(nst,nst)

      forall(n1=1:nst,n2=1:nst)
         Ht(n1,n2) = Interpol1D(gkba%order-1,gkba%dt,H%Ret(n1,n2,0:),t)
         ICt(n1,n2) = Interpol1D(gkba%order-1,gkba%dt,IC(n1,n2,0:),t)
      end forall

      drhodt = -iu*(matmul(Ht,Rho)-matmul(Rho,Ht)) + ICt

    end function RhoEOM
  !...................................................
    function CollInt_term_start(nst,n,j) result(IC_t)
      integer,intent(in)  :: nst,n,j
      complex(dp),dimension(nst,nst) :: IC_t
      complex(dp),dimension(nst,nst) :: SGMR,Y

      associate(nst => gkba%nst)
        IC_t = zero
        SGMR = SGM%Greater(1:nst,1:nst,n,j)-SGM%Lesser(1:nst,1:nst,n,j)
        Y = SGM%Lesser(1:nst,1:nst,n,j)
        call ZMM(nst,one,one,SGMR,gkba%Rho(:,:,j),Y)
        call ZMM(nst,one,one,Y,Ut(:,:,j,n),IC_t)
        IC_t = -iu * IC_t
      end associate

    end function CollInt_term_start
    !...................................................
  end subroutine GKBA_start
!--------------------------------------------------------------------------------------
  subroutine Gen_GKBA_CollInt(gkba,it,SGM)
    class(gkba_t),intent(inout)          :: gkba
    integer,intent(in)                   :: it
    type(CntrFun_timeslice_t),intent(in) :: SGM

    select case(gkba%nst)
    case(1)
       call Gen_GKBA_CollInt_Scalar(gkba,it,SGM)
    case default
       call Gen_GKBA_CollInt_Matrix(gkba,it,SGM)
    end select

  end subroutine Gen_GKBA_CollInt
!--------------------------------------------------------------------------------------
  subroutine Gen_GKBA_CollInt_Scalar(gkba,tstp,SGM)
    !..................................
    type(gkba_t),intent(inout)           :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: SGM
    !..................................
    integer :: j
    complex(dp) :: IC,SGM_ret,Ucc
    !..................................

    IC = zero
    if(tstp <= Cntrinteg%Greg2) then
       do j=0,tstp
          SGM_ret = SGM%Greater(1,1,j)-SGM%Lesser(1,1,j)
          Ucc = conjg(gkba%Uprop(1,1,j,gkba%stencil2p(1)))
          IC = IC - iu*Cntrinteg%w_retm(tstp,j) * (SGM%Lesser(1,1,j)+SGM_ret*gkba%Rho(1,1,j))*Ucc
       end do
    else
       do j=0,Cntrinteg%GregoryOrder
          SGM_ret = SGM%Greater(1,1,j)-SGM%Lesser(1,1,j)
          Ucc = conjg(gkba%Uprop(1,1,j,gkba%stencil2p(1)))
          IC = IC - iu*Cntrinteg%w_retm(tstp,j) * (SGM%Lesser(1,1,j)+SGM_ret*gkba%Rho(1,1,j))*Ucc
       end do
       do j=tstp-Cntrinteg%GregoryOrder,tstp
          SGM_ret = SGM%Greater(1,1,j)-SGM%Lesser(1,1,j)
          Ucc = conjg(gkba%Uprop(1,1,j,gkba%stencil2p(1)))
          IC = IC - iu*Cntrinteg%w_retm(tstp,j) * (SGM%Lesser(1,1,j)+SGM_ret*gkba%Rho(1,1,j))*Ucc
       end do
       do j=Cntrinteg%GregoryOrder+1,tstp-Cntrinteg%GregoryOrder-1
          SGM_ret = SGM%Greater(1,1,j)-SGM%Lesser(1,1,j)
          Ucc = conjg(gkba%Uprop(1,1,j,gkba%stencil2p(1)))
          IC = IC - iu*Cntrinteg%dt * (SGM%Lesser(1,1,j)+SGM_ret*gkba%Rho(1,1,j))*Ucc
       end do
    end if

    gkba%CollInt(1,1,gkba%stencil(1)) = IC + conjg(IC)

  end subroutine Gen_GKBA_CollInt_Scalar
!--------------------------------------------------------------------------------------
  subroutine Gen_GKBA_CollInt_Matrix(gkba,tstp,SGM)
    !..................................
    type(gkba_t),intent(inout)           :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: SGM
    !..................................
    integer :: nst,j
    complex(dp),dimension(gkba%nst,gkba%nst) :: IC
    complex(dp),dimension(gkba%nst,gkba%nst) :: SGM_ret,Ucc,y
    !..................................

    nst = gkba%nst
    IC = zero
    if(tstp <= Cntrinteg%Greg2) then
       do j=0,tstp
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          call ZMM(nst,-iu*Cntrinteg%w_retm(tstp,j),one,Y,Ucc,IC)
       end do
    else
       do j=0,Cntrinteg%GregoryOrder
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          call ZMM(nst,-iu*Cntrinteg%w_retm(tstp,j),one,Y,Ucc,IC)
       end do
       do j=tstp-Cntrinteg%GregoryOrder,tstp
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          call ZMM(nst,-iu*Cntrinteg%w_retm(tstp,j),one,Y,Ucc,IC)
       end do
       do j=Cntrinteg%GregoryOrder+1,tstp-Cntrinteg%GregoryOrder-1
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          call ZMM(nst,-iu*Cntrinteg%dt,one,Y,Ucc,IC)
       end do
    end if

    call ZHEpart(gkba%nst,IC,gkba%CollInt(:,:,gkba%stencil(1)))

  end subroutine Gen_GKBA_CollInt_Matrix
!--------------------------------------------------------------------------------------
  pure function GKBA_GetU_old(gkba,jt) result(U)
    !..................................
    !    Returns U(t_i,t_j)
    !..................................
    class(gkba_t),intent(in)::gkba
    integer,intent(in)::jt
    complex(dp),dimension(gkba%nst,gkba%nst)::U

    U=gkba%Uprop(:,:,jt,gkba%stencil2p(2))

  end function GKBA_GetU_old
!--------------------------------------------------------------------------------------
  pure function GKBA_GetU_new(gkba,jt) result(U)
    !..................................
    !    Returns U(t_{i+1},t_j)
    !..................................
    class(gkba_t),intent(in)::gkba
    integer,intent(in)::jt
    complex(dp),dimension(gkba%nst,gkba%nst)::U

    U=gkba%Uprop(:,:,jt,gkba%stencil2p(1))

  end function GKBA_GetU_new
!--------------------------------------------------------------------------------------
  pure subroutine GKBA_SetU_old(gkba,jt,U)
    !..................................
    !  sets U(t_i,t_j) to given U
    !..................................
    class(gkba_t),intent(inout)::gkba
    integer,intent(in)::jt
    complex(dp),intent(in)::U(:,:)

    gkba%Uprop(:,:,jt,gkba%stencil2p(2))=U

  end subroutine GKBA_SetU_old
!--------------------------------------------------------------------------------------
  pure subroutine GKBA_SetU_new(gkba,jt,U)
    !..................................
    !  sets U(t_{i+1},t_j) to given U
    !..................................
    class(gkba_t),intent(inout)::gkba
    integer,intent(in)::jt
    complex(dp),intent(in)::U(:,:)

    gkba%Uprop(:,:,jt,gkba%stencil2p(1))=U

  end subroutine GKBA_SetU_new
!--------------------------------------------------------------------------------------
  pure subroutine ExtrapolateH_CF4(gkba,H1,H2)
    !............................................
    !  Extrapolates H(t+c1*dt), H(t+c2*dt) from
    !
    !  H(t), H(t-dt), H(t-2*dt), ...
    !
    !............................................
    type(gkba_t),intent(in) :: gkba
    complex(dp),intent(out) :: H1(:,:),H2(:,:)
    integer  :: k
    real(dp) :: w(0:gkba%order-1,2)

    w = PolyW_extrapol_CF4(gkba%order)

    H1=zero
    H2=zero
    do k=0,gkba%order-1
       H1 = H1 + w(gkba%order-1-k,1)*gkba%H(:,:,gkba%stencil(k+1))
       H2 = H2 + w(gkba%order-1-k,2)*gkba%H(:,:,gkba%stencil(k+1))
    end do

  end subroutine ExtrapolateH_CF4
!--------------------------------------------------------------------------------------
  pure subroutine InterpolateH_CF4(gkba,H1,H2)
    !............................................
    !  Interpolates H(t+c1*dt), H(t+c2*dt) from
    !
    !  H(t+dt), H(t), H(t-dt), ...
    !
    !............................................
    type(gkba_t),intent(in) :: gkba
    complex(dp),intent(out) :: H1(:,:),H2(:,:)
    integer  :: k
    real(dp) :: w(0:gkba%order-1,2)

    w = PolyW_interpol_CF4(gkba%order)

    H1=zero
    H2=zero
    do k=0,gkba%order-1
       H1 = H1 + w(gkba%order-1-k,1)*gkba%H(:,:,gkba%stencil(k+1))
       H2 = H2 + w(gkba%order-1-k,2)*gkba%H(:,:,gkba%stencil(k+1))
    end do

  end subroutine InterpolateH_CF4
!--------------------------------------------------------------------------------------
  subroutine GKBA_Step_AdamsBashforth(gkba,tstp,order,large_size)
    !............................................
    ! Computes the Rho(t+dt) from collision terms
    !
    !  I(t), I(t-dt), I(t-2*dt) , ...
    !
    ! and the time-evolution operator
    !
    ! U(t+dt,t), U(t+dt,t-dt), U(t+dt,t-2*dt) , ...
    !
    !............................................
    type(gkba_t),intent(inout) :: gkba
    integer,intent(in)         :: tstp,order
    logical,intent(in),optional:: large_size
    logical :: use_zgemm
    integer :: k,p,ix
    complex(dp),dimension(gkba%nst,gkba%nst) :: U,IC

    use_zgemm = .false.
    if(present(large_size)) use_zgemm  =large_size

    p = max(1,order)
    p = min(p,5)

    U=gkba%GetU_newstep(tstp)
    call UnitaryStepFBW(gkba%nst,U,gkba%Rho(:,:,tstp),gkba%Rho(:,:,tstp+1),large_size=use_zgemm)

    do k=0,p-1
       ix = gkba%stencil(k+1)
       U = gkba%GetU_newstep(tstp-k)
       call UnitaryStepFBW(gkba%nst,U,gkba%CollInt(:,:,ix),IC,large_size=use_zgemm)
       gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp+1)+gkba%dt*CAdBash(p,k+1)*IC
    end do

  end subroutine GKBA_Step_AdamsBashforth
!--------------------------------------------------------------------------------------
  subroutine GKBA_Step_AdamsBashforth_direct(gkba,tstp,order)
    !............................................
    ! Computes the Rho(t+dt) from collision terms
    !
    !  I(t), I(t-dt), I(t-2*dt) , ...
    !
    ! and the time-evolution operator
    !
    ! U(t+dt,t), U(t+dt,t-dt), U(t+dt,t-2*dt) , ...
    !
    !............................................
    type(gkba_t),intent(inout) :: gkba
    integer,intent(in)         :: tstp,order
    integer :: k,p,ix

    p = max(1,order)
    p = min(p,5)

    gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp)
    do k=0,p-1
       ix = gkba%stencil(k+1)
       gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp+1) + gkba%dt*CAdBash(p,k+1) &
        * (gkba%C(:,:,ix) + gkba%CollInt(:,:,ix))
    end do

  end subroutine GKBA_Step_AdamsBashforth_direct
!--------------------------------------------------------------------------------------
  subroutine GKBA_Step_AdamsMoulton(gkba,tstp,order,large_size)
    !............................................
    ! Computes the Rho(t+dt) from collision terms
    !
    !  I(t+dt), I(t), I(t-dt) , ...
    !
    ! and the time-evolution operator
    !
    ! U(t+dt,t), U(t+dt,t-dt), U(t+dt,t-2*dt) , ...
    !
    !............................................
    type(gkba_t),intent(inout) :: gkba
    integer,intent(in)         :: tstp,order
    logical,intent(in),optional:: large_size
    logical :: use_zgemm
    integer :: k,p,ix
    complex(dp),dimension(gkba%nst,gkba%nst) :: U,IC

    use_zgemm = .false.
    if(present(large_size)) use_zgemm  =large_size

    p = max(1,order)
    p = min(p,6)

    U = gkba%GetU_newstep(tstp)
    call UnitaryStepFBW(gkba%nst,U,gkba%Rho(:,:,tstp),gkba%Rho(:,:,tstp+1),large_size=use_zgemm)

    do k=0,p-1
       ix = gkba%stencil(k+1)
       U = gkba%GetU_newstep(tstp+1-k)
       call UnitaryStepFBW(gkba%nst,U,gkba%CollInt(:,:,ix),IC,large_size=use_zgemm)
       gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp+1)+gkba%dt*CAdMoul(p,k+1)*IC
    end do

  end subroutine GKBA_Step_AdamsMoulton
!--------------------------------------------------------------------------------------
  subroutine GKBA_Step_AdamsMoulton_direct(gkba,tstp,order)
    !............................................
    ! Computes the Rho(t+dt) from collision terms
    !
    !  I(t+dt), I(t), I(t-dt) , ...
    !
    ! and the time-evolution operator
    !
    ! U(t+dt,t), U(t+dt,t-dt), U(t+dt,t-2*dt) , ...
    !
    !............................................
    type(gkba_t),intent(inout) :: gkba
    integer,intent(in)         :: tstp,order
    integer :: k,p,ix

    p = max(1,order)
    p = min(p,6)

    gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp)

    do k=0,p-1
       ix = gkba%stencil(k+1)
       gkba%Rho(:,:,tstp+1) = gkba%Rho(:,:,tstp+1) + gkba%dt*CAdMoul(p,k+1) &
        * (gkba%C(:,:,ix) + gkba%CollInt(:,:,ix))
    end do

  end subroutine GKBA_Step_AdamsMoulton_direct
!--------------------------------------------------------------------------------------
  subroutine GKBA_Predictor(gkba,tstp,large_size,direct_method)
    type(gkba_t),intent(inout)  :: gkba
    integer,intent(in)          :: tstp
    logical,intent(in),optional :: large_size
    logical,intent(in),optional :: direct_method
    logical :: use_zgemm,direct
    integer :: nst,j
    complex(dp),dimension(gkba%nst,gkba%nst) :: H1,H2,Udt,Uold,Unew

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    direct = .false.
    if(present(direct_method)) direct = direct_method

    nst = gkba%nst

    call ExtrapolateH_CF4(gkba,H1,H2)
    call GenU_CF4(gkba%dt,H1,H2,Udt,large_size=use_zgemm)

    ! .. update propagators: U(t_{i+1},t_{j}) = U(t_{i+1},t_i)*U(t_i,t_{j}) ..
    if(use_zgemm) then
      do j=0,tstp-1
         Uold = gkba%GetU_oldstep(j)
         call blas_zgemm('N','N',nst,nst,nst,one,Udt,nst,Uold,nst,zero,Unew,nst)
         call gkba%SetU_newstep(j,Unew)
      end do
    else
      do j=0,tstp-1
         Uold = gkba%GetU_oldstep(j)
         call Zmult(gkba%nst,Udt,Uold,Unew)
         call gkba%SetU_newstep(j,Unew)
      end do
    end if

    call gkba%SetU_newstep(tstp,Udt)
    gkba%Uprop(:,:,tstp+1,gkba%stencil2p(1)) = Eye(gkba%nst)

    ! .. perform time step ..
    if(direct) then
      call GKBA_Step_AdamsBashforth_direct(gkba,tstp,gkba%order)
    else
      call GKBA_Step_AdamsBashforth(gkba,tstp,gkba%order,large_size=use_zgemm)
    end if

    ! .. switch to Corrector ..
    call gkba%SwitchToCorrector()

  end subroutine GKBA_Predictor
!--------------------------------------------------------------------------------------
  subroutine GKBA_Corrector(gkba,tstp,Sigma,last_step,large_size,direct_method)
    type(gkba_t),intent(inout)           :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: Sigma
    logical,intent(in),optional          :: last_step
    logical,intent(in),optional          :: large_size
    logical,intent(in),optional          :: direct_method
    logical :: switch_to_predictor
    logical :: use_zgemm,direct
    integer :: nst,j,ix
    complex(dp),dimension(gkba%nst,gkba%nst) :: H1,H2,Udt,Uold,Unew

    switch_to_predictor = .false.
    if(present(last_step)) switch_to_predictor = last_step

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    direct = .false.
    if(present(direct_method)) direct = direct_method

    nst = gkba%nst

    call InterpolateH_CF4(gkba,H1,H2)
    call GenU_CF4(gkba%dt,H1,H2,Udt,large_size=use_zgemm)

    ! .. update propagators: U(t_{i+1},t_{j}) = U(t_{i+1},t_i)*U(t_i,t_{j}) ..
    if(use_zgemm) then
      do j=0,tstp-1
         Uold = gkba%GetU_oldstep(j)
         call blas_zgemm('N','N',nst,nst,nst,one,Udt,nst,Uold,nst,zero,Unew,nst)
         call gkba%SetU_newstep(j,Unew)
      end do
    else
      do j=0,tstp-1
         Uold = gkba%GetU_oldstep(j)
         call Zmult(gkba%nst,Udt,Uold,Unew)
         call gkba%SetU_newstep(j,Unew)
      end do
    end if

    call gkba%SetU_newstep(tstp,Udt)
    gkba%Uprop(:,:,tstp+1,gkba%stencil2p(1)) = Eye(gkba%nst)

    ! .. update collision integral ..
    call gkba%GenCollInt(tstp+1,Sigma)
    if(gkba%dephasing_mode) then
        ix = gkba%stencil(1)
        call AddDephasing(gkba%nst,gkba%Rho(:,:,tstp+1),gkba%H(:,:,ix),gkba%Tdeph_mat,&
            gkba%CollInt(:,:,ix),large_size=use_zgemm)
    end if

    ! .. perform time step ..
    if(direct) then
      call GKBA_Step_AdamsMoulton_direct(gkba,tstp,gkba%order)
    else
      call GKBA_Step_AdamsMoulton(gkba,tstp,gkba%order,large_size=use_zgemm)
    end if

    ! .. switch to Predictor ..
    if(switch_to_predictor) then
      call gkba%SwitchToNextStep()
    end if

  end subroutine GKBA_Corrector
!--------------------------------------------------------------------------------------
  real(dp) function GKBA_ExpectVal(gkba,tstp,A)
    class(gkba_t),intent(in) :: gkba
    integer,intent(in)       :: tstp
    complex(dp),intent(in)   :: A(:,:)
    integer :: n,m

    GKBA_ExpectVal = 0.0_dp
    do concurrent(n=1:gkba%nst,m=1:gkba%nst)
       GKBA_ExpectVal=GKBA_ExpectVal+real(gkba%Rho(n,m,tstp)*A(m,n),dp)
    end do

  end function GKBA_ExpectVal
!--------------------------------------------------------------------------------------
  function GKBA_GetLesser(gkba,it,jt,large_size) result(Glsr)
    class(gkba_t),intent(in) :: gkba
    integer,intent(in)       :: it,jt
    logical,intent(in),optional :: large_size
    logical :: use_zgemm
    complex(dp) :: Glsr(gkba%nst,gkba%nst)
    complex(dp) :: U(gkba%nst,gkba%nst)

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    if(it>=jt) then
       U=gkba%GetU_newstep(jt)
       if(use_zgemm) then
         call blas_zgemm('N','N',gkba%nst,gkba%nst,gkba%nst,iu,U,gkba%nst,&
          gkba%Rho(:,:,jt),gkba%nst,zero,Glsr,gkba%nst)
       else
         Glsr=zero
         call ZMM(gkba%nst,iu,zero,U,gkba%Rho(:,:,jt),Glsr)
       end if
       return
    else
       U=gkba%GetU_newstep(it)
       if(use_zgemm) then
         call blas_zgemm('N','N',gkba%nst,gkba%nst,gkba%nst,iu,U,gkba%nst,&
          gkba%Rho(:,:,it),gkba%nst,zero,Glsr,gkba%nst)
       else
         Glsr=zero
         call ZMM(gkba%nst,iu,zero,U,gkba%Rho(:,:,it),Glsr)
       end if
       Glsr=-HermConjg(Glsr)
    end if

  end function GKBA_GetLesser
!--------------------------------------------------------------------------------------
  function GKBA_GetGreater(gkba,it,jt,large_size) result(Ggtr)
    class(gkba_t),intent(in) :: gkba
    integer,intent(in)       :: it,jt
    logical,intent(in),optional :: large_size
    logical :: use_zgemm
    complex(dp) :: Ggtr(gkba%nst,gkba%nst)
    integer     :: n
    complex(dp) :: U(gkba%nst,gkba%nst),xrho(gkba%nst,gkba%nst)

    use_zgemm = .false.
    if(present(large_size)) use_zgemm = large_size

    if(it>=jt) then
       xrho=-gkba%Rho(:,:,jt)
       forall(n=1:gkba%nst) xrho(n,n)=one+xrho(n,n)
       U=gkba%GetU_newstep(jt)
       if(use_zgemm) then
        call blas_zgemm('N','N',gkba%nst,gkba%nst,gkba%nst,-iu,U,gkba%nst,&
          xrho,gkba%nst,zero,Ggtr,gkba%nst)
       else
         Ggtr=zero
         call ZMM(gkba%nst,-iu,zero,U,xrho,Ggtr)
       end if
       return
    else
       xrho=-gkba%Rho(:,:,it)
       forall(n=1:gkba%nst) xrho(n,n)=one+xrho(n,n)
       U=gkba%GetU_newstep(it)
       if(use_zgemm) then
         call blas_zgemm('N','N',gkba%nst,gkba%nst,gkba%nst,-iu,U,gkba%nst,&
          xrho,gkba%nst,zero,Ggtr,gkba%nst)
       else
         Ggtr=zero
         call ZMM(gkba%nst,-iu,zero,U,xrho,Ggtr)
       end if
       Ggtr=-HermConjg(Ggtr)
    end if

  end function GKBA_GetGreater
!--------------------------------------------------------------------------------------
  real(dp) function GKBA_Ecorr_Scalar(gkba,tstp,SGM)
    !..................................
    type(gkba_t),intent(in)              :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: SGM
    !..................................
    integer     :: kt
    complex(dp) :: SGMR,SGML,Ucc,IC
    !..................................

    if(tstp == 0) then
       GKBA_Ecorr_Scalar = 0.0_dp
       return
    end if

    IC = zero
    if(tstp <= CntrInteg%Greg2) then
       do kt=0,tstp
          SGMR = SGM%Greater(1,1,kt) - SGM%Lesser(1,1,kt)
          SGML = SGM%Lesser(1,1,kt)
          Ucc=conjg(gkba%Uprop(1,1,kt,gkba%stencil2p(1)))
          IC = IC + Cntrinteg%w_retm(tstp,kt)*(SGML + SGMR*gkba%Rho(1,1,kt))*Ucc
       end do
    else
      do kt=0,CntrInteg%GregoryOrder
         SGMR = SGM%Greater(1,1,kt) - SGM%Lesser(1,1,kt)
         SGML = SGM%Lesser(1,1,kt)
         Ucc=conjg(gkba%Uprop(1,1,kt,gkba%stencil2p(1)))
         IC = IC + Cntrinteg%w_retm(tstp,kt)*(SGML + SGMR*gkba%Rho(1,1,kt))*Ucc
      end do
      do kt=tstp-CntrInteg%GregoryOrder,tstp
         SGMR = SGM%Greater(1,1,kt) - SGM%Lesser(1,1,kt)
         SGML = SGM%Lesser(1,1,kt)
         Ucc=conjg(gkba%Uprop(1,1,kt,gkba%stencil2p(1)))
         IC = IC + Cntrinteg%w_retm(tstp,kt)*(SGML + SGMR*gkba%Rho(1,1,kt))*Ucc
      end do
      do kt=CntrInteg%GregoryOrder+1,tstp-CntrInteg%GregoryOrder-1
         SGMR = SGM%Greater(1,1,kt) - SGM%Lesser(1,1,kt)
         SGML = SGM%Lesser(1,1,kt)
         Ucc=conjg(gkba%Uprop(1,1,kt,gkba%stencil2p(1)))
         IC = IC + Cntrinteg%dt*(SGML + SGMR*gkba%Rho(1,1,kt))*Ucc
      end do
    end if

    GKBA_Ecorr_Scalar = 0.5_dp*dble(IC)

  end function GKBA_Ecorr_Scalar
!--------------------------------------------------------------------------------------
  real(dp) function GKBA_Ecorr_Matrix(gkba,tstp,SGM)
    !..................................
    type(gkba_t),intent(in)              :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: SGM
    !..................................
    integer :: nst,j
    complex(dp) :: al
    complex(dp),dimension(gkba%nst,gkba%nst) :: SGM_ret,Ucc,IC,Y
    !..................................

    if(tstp == 0) then
       GKBA_Ecorr_Matrix = 0.0_dp
       return
    end if

    nst = gkba%nst

    IC = zero
    if(tstp <= Cntrinteg%Greg2) then
       do j=0,tstp
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          al = cmplx(Cntrinteg%w_retm(tstp,j), 0.0_dp, kind=dp)
          call ZMM(nst,al,one,Y,Ucc,IC)
       end do
    else
       do j=0,Cntrinteg%GregoryOrder
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          al = cmplx(Cntrinteg%w_retm(tstp,j), 0.0_dp, kind=dp)
          call ZMM(nst,al,one,Y,Ucc,IC)
       end do
       do j=tstp-Cntrinteg%GregoryOrder,tstp
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          al = cmplx(Cntrinteg%w_retm(tstp,j), 0.0_dp, kind=dp)
          call ZMM(nst,al,one,Y,Ucc,IC)
       end do
       do j=Cntrinteg%GregoryOrder+1,tstp-Cntrinteg%GregoryOrder-1
          SGM_ret = SGM%Greater(1:nst,1:nst,j)-SGM%Lesser(1:nst,1:nst,j)
          Ucc = HermConjg(gkba%Uprop(1:nst,1:nst,j,gkba%stencil2p(1)))
          Y = SGM%Lesser(1:nst,1:nst,j)
          call ZMM(nst,one,one,SGM_ret,gkba%Rho(:,:,j),Y)
          al = cmplx(Cntrinteg%dt, 0.0_dp, kind=dp)
          call ZMM(nst,al,one,Y,Ucc,IC)
       end do
    end if

    GKBA_Ecorr_Matrix = 0.5_dp*dble(Trace(IC))

  end function GKBA_Ecorr_Matrix
!--------------------------------------------------------------------------------------
  real(dp) function GKBA_Ecorr(gkba,tstp,SGM)
    !..................................
    class(gkba_t),intent(in)             :: gkba
    integer,intent(in)                   :: tstp
    type(CntrFun_timeslice_t),intent(in) :: SGM
    !..................................

    select case(gkba%nst)
    case(1)
       GKBA_Ecorr = GKBA_Ecorr_Scalar(gkba,tstp,SGM)
    case default
       GKBA_Ecorr = GKBA_Ecorr_Matrix(gkba,tstp,SGM)
    end select

  end function GKBA_Ecorr
!--------------------------------------------------------------------------------------
  subroutine GKBA_Reconstruct_Green(gkba,h,G)
    integer,parameter :: nstart=4,pIntp=3
    real(dp),parameter :: c1=0.5_dp-sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: c2=0.5_dp+sqrt(3.0_dp)/6.0_dp
    class(gkba_t),intent(in)             :: gkba
    type(CntrFun_1time_t),intent(in)     :: h
    type(CntrFun_2time_t),intent(inout)  :: G
    integer  :: n,j,i1,i2
    real(dp) :: t1,t2
    complex(dp),dimension(gkba%nst,gkba%nst) :: H1,H2,Udt
    type(grid1d_t) :: tgrd

    call assert(gkba%Nt == G%Nt1)
    call assert(gkba%Nt == G%Nt2)
    call assert(gkba%Nst == G%Nstates)

    do n=0,gkba%Nt
       G%Lesser(1:gkba%Nst,1:gkba%Nst,n,n) = gkba%GetLesser(n,n)
       G%Greater(1:gkba%Nst,1:gkba%Nst,n,n) = gkba%GetGreater(n,n)
    end do

    call tgrd%Init(gkba%Nt,0.0_dp,gkba%Nt*gkba%dt)

    do n=0,gkba%Nt-1
       t1 = (n + c1) * gkba%dt
       t2 = (n + c2) * gkba%dt
       call H%Interpolate(tgrd,t1,H1)
       call H%Interpolate(tgrd,t2,H2)

       call GenU_CF4(gkba%dt,H1,H2,Udt)

       do j=0,n
          call UnitaryStepFW(gkba%nst,Udt,G%Lesser(:,:,n,j),G%Lesser(:,:,n+1,j))
          call UnitaryStepFW(gkba%nst,Udt,G%Greater(:,:,n,j),G%Greater(:,:,n+1,j))
          G%Lesser(:,:,j,n+1) = -HermConjg(G%Lesser(:,:,n+1,j))
          G%Greater(:,:,j,n+1) = -HermConjg(G%Greater(:,:,n+1,j))
       end do
    end do

    call tgrd%Clean

  end subroutine GKBA_Reconstruct_Green
!--------------------------------------------------------------------------------------
  subroutine GKBA_Reconstruct_Green_step(gkba,tstp,h,G,AboveDiag)
    !..............................................
    integer,parameter :: nstart=4,pIntp=3
    real(dp),parameter :: c1=0.5_dp-sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: c2=0.5_dp+sqrt(3.0_dp)/6.0_dp
    class(gkba_t),intent(in)                 :: gkba
    integer,intent(in)                       :: tstp
    type(CntrFun_1time_t),intent(in)         :: h
    type(CntrFun_timeslice_t),intent(inout)  :: G
    logical,intent(in),optional              :: AboveDiag
    logical  :: AboveDiag_=.false.
    integer  :: j,k
    real(dp) :: t1,t2
    complex(dp),dimension(gkba%nst,gkba%nst) :: H1,H2,Udt,xrho
    complex(dp),allocatable :: Uprop(:,:,:)
    type(grid1d_t) :: tgrd

    call assert(tstp <= G%Nt)
    call assert(tstp <= gkba%Nt)
    call assert(gkba%Nst == G%Nstates)

    if(present(AboveDiag)) AboveDiag_ = AboveDiag

    G%Lesser(1:gkba%Nst,1:gkba%Nst,tstp) = gkba%GetLesser(tstp,tstp)
    G%Greater(1:gkba%Nst,1:gkba%Nst,tstp) = gkba%GetGreater(tstp,tstp)

    allocate(Uprop(gkba%Nst,gkba%Nst,0:gkba%Nt))
    Uprop(:,:,tstp) = Eye(gkba%Nst)

    call tgrd%Init(gkba%Nt,0.0_dp,gkba%Nt * gkba%dt)

    ! ... G(t_n,t_j) with j <= n ...
    do k=1,tstp
       j = tstp-k
       t1 = (j + c1) * gkba%dt
       t2 = (j + c2) * gkba%dt
       call H%Interpolate(tgrd,t1,H1)
       call H%Interpolate(tgrd,t2,H2)
       call GenU_CF4(gkba%dt,H1,H2,Udt)
       call Zmult(gkba%nst,Uprop(:,:,j+1),Udt,Uprop(:,:,j))
    end do

    do j=0,tstp-1
       call Zmult(gkba%nst,Uprop(:,:,j),gkba%rho(:,:,j),G%Lesser(:,:,j))
       xrho = Eye(gkba%nst) - gkba%Rho(:,:,j)
       call Zmult(gkba%nst,Uprop(:,:,j),xrho,G%Greater(:,:,j))
    end do

    G%Lesser(:,:,0:tstp) = iu * G%Lesser(:,:,0:tstp)
    G%Greater(:,:,0:tstp) = -iu * G%Greater(:,:,0:tstp)

    if(AboveDiag_) then
       ! ... G(t_n,t_j) , j = n+1,..,Nt ...
       do j=tstp,gkba%Nt-1
          t1 = (j + c1) * gkba%dt
          t2 = (j + c2) * gkba%dt
          call H%Interpolate(tgrd,t1,H1)
          call H%Interpolate(tgrd,t2,H2)
          call GenU_CF4(gkba%dt,H1,H2,Udt)
          call Zmult(gkba%nst,Uprop(:,:,j),Udt,Uprop(:,:,j+1))
       end do

       xrho = Eye(gkba%nst) - gkba%Rho(:,:,tstp)
       do j=tstp+1,gkba%Nt
          call Zmult(gkba%nst,Uprop(:,:,j),gkba%rho(:,:,tstp),G%Lesser(:,:,j))
          call Zmult(gkba%nst,Uprop(:,:,j),xrho,G%Greater(:,:,j))
       end do

       G%Lesser(:,:,tstp+1:gkba%Nt) = iu * G%Lesser(:,:,tstp+1:gkba%Nt)
       G%Greater(:,:,tstp+1:gkba%Nt) = -iu * G%Greater(:,:,tstp+1:gkba%Nt)

       do j=tstp+1,gkba%Nt
          G%Lesser(:,:,j) = -hermconjg(G%Lesser(:,:,j))
          G%Greater(:,:,j) = -hermconjg(G%Greater(:,:,j))
       end do

    end if

    call tgrd%Clean()
    deallocate(Uprop)

  end subroutine GKBA_Reconstruct_Green_Step
!--------------------------------------------------------------------------------------
  pure complex(dp) function Interpol1D(n,h,u,x)
    integer,intent(in) :: n
    real(dp), intent(in) :: h
    complex(dp), intent(in) :: u(0:)
    real(dp), intent(in) :: x
    integer :: i,k
    real(dp) :: Lx

    Interpol1D=0.0_dp
    do i=0,n
       Lx=1.0_dp
       do k=0,n
          if(k==i) cycle
          Lx=Lx*(x-k*h)/(i*h-k*h)
       end do
       Interpol1D=Interpol1D+u(i)*Lx
    end do
  end function Interpol1D
!--------------------------------------------------------------------------------------
  pure function PolyW(p,x) result(w)
    integer,intent(in)   :: p
    real(dp),intent(in)  :: x
    real(dp)             :: w(0:p-1)
    integer :: i,k

    do i=0,p-1
       w(i) = 1.0_dp
       do k=0,p-1
          if(k==i) cycle
          w(i) = w(i)*(x-k)/dble(i-k)
       end do
    end do
  end function PolyW
!--------------------------------------------------------------------------------------
  pure function PolyW_extrapol_CF4(p) result(w)
    real(dp),parameter :: c1 = 0.5_dp - sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: c2 = 0.5_dp + sqrt(3.0_dp)/6.0_dp
    integer,intent(in)   :: p
    real(dp)             :: w(0:p-1,2)

    w(0:,1) = PolyW(p,p-1+c1)
    w(0:,2) = PolyW(p,p-1+c2)

  end function PolyW_extrapol_CF4
!--------------------------------------------------------------------------------------
  pure function PolyW_interpol_CF4(p) result(w)
    real(dp),parameter :: c1 = 0.5_dp - sqrt(3.0_dp)/6.0_dp
    real(dp),parameter :: c2 = 0.5_dp + sqrt(3.0_dp)/6.0_dp
    integer,intent(in)   :: p
    real(dp)             :: w(0:p-1,2)

    w(0:,1) = PolyW(p,p-2+c1)
    w(0:,2) = PolyW(p,p-2+c2)

  end function PolyW_interpol_CF4
!--------------------------------------------------------------------------------------
  subroutine Init_GKBA_Array(gkba_arr,N,nst,tgrid,order)
    type(gkba_t),allocatable,intent(out) :: gkba_arr(:)
    integer,intent(in)               :: N
    integer,intent(in)               :: nst
    type(grid1d_t),intent(in)        :: tgrid
    integer,intent(in),optional      :: order
    integer :: i,order_

    order_ = 5
    if(present(order)) order_ = min(order,6)

    if(allocated(gkba_arr)) deallocate(gkba_arr)
    allocate(gkba_arr(1:N))

    do i=1,N
      gkba_arr(i)%order = order_
      gkba_arr(i)%nst = nst
      gkba_arr(i)%Nt = tgrid%Npts
      gkba_arr(i)%dt = tgrid%dX

      allocate(gkba_arr(i)%H(nst,nst,order_)) ; gkba_arr(i)%H=zero
      allocate(gkba_arr(i)%C(nst,nst,order_)) ; gkba_arr(i)%C=zero
      allocate(gkba_arr(i)%Rho(nst,nst,0:tgrid%Npts)) ; gkba_arr(i)%Rho=zero
      allocate(gkba_arr(i)%Uprop(nst,nst,0:tgrid%Npts,2)) ; gkba_arr(i)%Uprop=zero
      allocate(gkba_arr(i)%CollInt(nst,nst,order_)) ; gkba_arr(i)%CollInt=zero
      allocate(gkba_arr(i)%stencil(order_))
      gkba_arr(i)%stencil = [(i,i=1,order_)]
    end do

  end subroutine Init_GKBA_Array
!--------------------------------------------------------------------------------------
  subroutine Clean_GKBA_array(gkba_arr)
    type(gkba_t),allocatable,intent(inout) :: gkba_arr(:)
    integer :: i

    do i=1,size(gkba_arr)
       call gkba_arr(i)%Clean()
    end do
    deallocate(gkba_arr)
    
  end subroutine Clean_GKBA_array
!--------------------------------------------------------------------------------------

!======================================================================================
end module Kbkit_Gkba
