module kbkit_integration
!====================================================================================
!! author: Michael Schueler
!!
!! Contains tools for computing one-dimensional grid-based integrals
!! using Gregory's rule and cubic spline quadrature.
!====================================================================================
   use Mdef,only:&
      dp,zero
   use Mgrid1d,only:&
      grid1d_t
   use Mcoefficients,only:&
      StartPolyW,&
      VolterraW
   use Mgregory,only:&
      GenGridWeights
   implicit none
!------------------------------------------------------------------------------------
   private
   public &
      CntrIntegrals_t,&
      GregoryIntegral
!------------------------------------------------------------------------------------
   interface GregoryIntegral
      module procedure GregoryIntegral_scalar,GregoryIntegral_matrix,&
         GregoryIntegral_matrix_work
   end interface GregoryIntegral

   abstract interface

   function integrand_scalar(n,j) result(f)
#ifdef OMPSIMD
      !$OMP DECLARE SIMD(integrand_scalar) UNIFORM(n)
#endif
      import :: dp
      integer,intent(in) :: n,j
      complex(dp) :: f
   end function integrand_scalar

   function integrand_matrix(size,n,j) result(f)
      import :: dp
      integer,intent(in) :: size,n,j
      complex(dp) :: f(size,size)
   end function integrand_matrix

   function integrand_matrix_work(size,n,j,work) result(f)
      import :: dp
      integer,intent(in) :: size,n,j
      complex(dp),intent(inout) :: work(:,:)
      complex(dp) :: f(size,size)
   end function integrand_matrix_work

   end interface
!------------------------------------------------------------------------------------
   type CntrIntegrals_t
      !! contour integrator class
      integer, public    :: GregoryOrder,Greg2                      !! order of Gregory integration
      integer, public    :: Nt                                      !! number of real time steps
      real(dp), private  :: Tmx                                     !! max. propagation time
      real(dp), public   :: dt                                      !! time step size
      real(dp), allocatable, dimension(:,:),private :: w_viestart_  !! Volterra starting weights for real axis
      real(dp), allocatable, dimension(:),private :: w_vieomeg_     !! Volterra omega weights for real axis
      real(dp), allocatable, dimension(:,:),private :: w_retm_      !! quadrature weights for real axis
      real(dp), allocatable, dimension(:,:),private :: w_poly_      !! polynomial quadrature weights for real axis
   contains
      procedure,public :: Init => Init_Integ
      procedure,public :: Clean => Clean_Integ
      procedure,public :: w_viestart
      procedure,public :: w_vieomeg
      procedure,public :: w_retm
      procedure,public :: w_poly
   end type CntrIntegrals_t
!------------------------------------------------------------------------------------
contains
!------------------------------------------------------------------------------------
  subroutine Init_Integ(me,GregorgyOrder,t_grid)
    class(CntrIntegrals_t),intent(inout) :: me
    integer,intent(in)                   :: GregorgyOrder
    type(grid1d_t),intent(in)            :: t_grid
    integer  :: n,k,nstart
    type(grid1d_t) :: tstart

    me%GregoryOrder = GregorgyOrder
    nstart = me%GregoryOrder - 1
    me%Greg2 = 2*(me%GregoryOrder+1)
    me%Nt=t_grid%Npts
    me%Tmx=t_grid%Xmax
    me%dt=me%Tmx/dble(me%Nt)

    call tstart%Init(me%Greg2,0.0_dp,me%Greg2*me%dt)
    call GenGridWeights(me%GregoryOrder,tstart,me%w_retm_,transp=.true.)
    !me%w_retm_ = transpose(me%w_retm_)
    call tstart%Clean()
       
    allocate(me%w_poly_(0:nstart,0:nstart))
    me%w_poly_ = me%dt * StartPolyW(me%GregoryOrder)

    ! --- Volterra weights ---
    allocate(me%w_viestart_(0:2*GregorgyOrder-2,0:GregorgyOrder-1))
    allocate(me%w_vieomeg_(0:GregorgyOrder-1))

    call VolterraW(GregorgyOrder,me%w_viestart_,me%w_vieomeg_)
    
  end subroutine Init_Integ
!------------------------------------------------------------------------------------
  subroutine Clean_Integ(me)
    class(CntrIntegrals_t)::me

    if(allocated(me%w_retm_)) deallocate(me%w_retm_)
    if(allocated(me%w_poly_)) deallocate(me%w_poly_)
    if(allocated(me%w_viestart_)) deallocate(me%w_viestart_)
    if(allocated(me%w_vieomeg_)) deallocate(me%w_vieomeg_)

  end subroutine Clean_Integ
!------------------------------------------------------------------------------------
  pure real(dp) function w_viestart(me,n,j)
#ifdef OMPSIMD
    !$OMP DECLARE SIMD(w_viestart) UNIFORM(me)
#endif
    class(CntrIntegrals_t),intent(in) :: me
    integer,intent(in) :: n,j

    w_viestart = me%w_viestart_(n,j)

  end function w_viestart
!------------------------------------------------------------------------------------
  pure real(dp) function w_vieomeg(me,j)
#ifdef OMPSIMD
    !$OMP DECLARE SIMD(w_vieomeg) UNIFORM(me)
#endif
    class(CntrIntegrals_t),intent(in) :: me
    integer,intent(in) :: j

    w_vieomeg = me%w_vieomeg_(j)

  end function w_vieomeg
!------------------------------------------------------------------------------------
  pure real(dp) function w_retm(me,n,j)
#ifdef OMPSIMD
    !$OMP DECLARE SIMD(w_retm) UNIFORM(me,n)
#endif
    class(CntrIntegrals_t),intent(in) :: me
    integer,intent(in) :: n,j
    integer :: l

    if(n < me%Greg2) then
       w_retm = me%w_retm_(j,n)
       return
    end if

    if(j <= me%GregoryOrder) then
       w_retm = me%w_retm_(j,me%Greg2)
       return
    end if

    if(j >= n-me%GregoryOrder) then
       l = j - (n-me%GregoryOrder) + me%GregoryOrder + 2
       w_retm = me%w_retm_(l,me%Greg2)
       return
    end if

    w_retm = me%dt

  end function w_retm
!------------------------------------------------------------------------------------
  pure real(dp) function w_poly(me,n,j)
    class(CntrIntegrals_t),intent(in) :: me
    integer,intent(in) :: n,j

    w_poly = me%w_poly_(n,j)

  end function w_poly
!------------------------------------------------------------------------------------
   function GregoryIntegral_scalar(n,integrand,CInteg,skiplast) result(fint)
      integer,intent(in) :: n
      procedure(integrand_scalar) :: integrand
      type(CntrIntegrals_t),intent(in) :: CInteg
      logical,intent(in),optional :: skiplast
      integer :: nmax,j
      complex(dp) :: f,fint

      nmax = n
      if(present(skiplast)) then
         if(skiplast) nmax = n - 1
      end if

      fint = zero

      if(n < 2*CInteg%GregoryOrder-2) then
         do j=0,CInteg%GregoryOrder-1
            f = integrand(n,j)
            fint = fint + CInteg%w_viestart(n,j)*f
         end do
      else
         do j=0,CInteg%GregoryOrder-1
            f = integrand(n,j)
            fint = fint + CInteg%w_viestart(2*CInteg%GregoryOrder-2,j)*f
         end do
      end if

      if(n >= CInteg%GregoryOrder) then
         do j=max(n-CInteg%GregoryOrder+1,CInteg%GregoryOrder),nmax
            f = integrand(n,j)
            fint = fint + CInteg%w_vieomeg(n-j)*f
         end do

#ifdef OMPSIMD
         !$OMP SIMD REDUCTION(+:fint) PRIVATE(f)
#endif
         do j=CInteg%GregoryOrder,n-CInteg%GregoryOrder
            f = integrand(n,j)
            fint = fint + f
         end do
      end if

      fint = fint * CInteg%dt

   end function GregoryIntegral_scalar
!------------------------------------------------------------------------------------
   function GregoryIntegral_matrix(size,n,integrand,CInteg,skiplast) result(fint)
      integer,intent(in) :: size,n
      procedure(integrand_matrix) :: integrand
      type(CntrIntegrals_t),intent(in) :: CInteg
      logical,intent(in),optional :: skiplast
      integer :: nmax,j
      complex(dp),dimension(size,size) :: f,fint

      nmax = n
      if(present(skiplast)) then
         if(skiplast) nmax = n - 1
      end if

      fint = zero

      if(n < 2*CInteg%GregoryOrder-2) then
         do j=0,CInteg%GregoryOrder-1
            f = integrand(size,n,j)
            fint = fint + CInteg%w_viestart(n,j)*f
         end do
      else
         do j=0,CInteg%GregoryOrder-1
            f = integrand(size,n,j)
            fint = fint + CInteg%w_viestart(2*CInteg%GregoryOrder-2,j)*f
         end do
      end if

      if(n >= CInteg%GregoryOrder) then
         do j=max(n-CInteg%GregoryOrder+1,CInteg%GregoryOrder),nmax
            f = integrand(size,n,j)
            fint = fint + CInteg%w_vieomeg(n-j)*f
         end do

         do j=CInteg%GregoryOrder,n-CInteg%GregoryOrder
            f = integrand(size,n,j)
            fint = fint + f
         end do
      end if

      fint = fint * CInteg%dt

   end function GregoryIntegral_matrix
!------------------------------------------------------------------------------------
  function GregoryIntegral_matrix_work(size,n,integrand,work,CInteg,skiplast) result(fint)
    integer,intent(in) :: size,n
    procedure(integrand_matrix_work) :: integrand
    complex(dp),intent(inout)        :: work(:,:)
    type(CntrIntegrals_t),intent(in) :: CInteg
    logical,intent(in),optional :: skiplast
    integer :: nmax,j
    complex(dp),dimension(size,size) :: f,fint

    nmax = n
    if(present(skiplast)) then
       if(skiplast) nmax = n - 1
    end if

    fint = zero

    if(n < 2*CInteg%GregoryOrder-2) then
       do j=0,CInteg%GregoryOrder-1
          f = integrand(size,n,j,work)
          fint = fint + CInteg%w_viestart(n,j)*f
       end do
    else
       do j=0,CInteg%GregoryOrder-1
          f = integrand(size,n,j,work)
          fint = fint + CInteg%w_viestart(2*CInteg%GregoryOrder-2,j)*f
       end do
    end if

    if(n >= CInteg%GregoryOrder) then
       do j=max(n-CInteg%GregoryOrder+1,CInteg%GregoryOrder),nmax
          f = integrand(size,n,j,work)
          fint = fint + CInteg%w_vieomeg(n-j)*f
       end do

       do j=CInteg%GregoryOrder,n-CInteg%GregoryOrder
          f = integrand(size,n,j,work)
          fint = fint + f
       end do
    end if

    fint = fint * CInteg%dt

  end function GregoryIntegral_matrix_work
!------------------------------------------------------------------------------------

!====================================================================================
end module Kbkit_Integration
