module kbkit_spectral
!======================================================================================
! This module provides tools for constructing free Green's functions from a give
! density of states
!======================================================================================
   use omp_lib
   use Mdebug
   use Mdef,only:&
      dp,iu,one,zero,nfermi,nnfermi,nbose
   use Mgrid1d,only:&
      grid1d_t
   use kbkit_greenfun,only:&
      CntrFun_2time_t,&
      CntrFun_timeslice_t
   implicit none
!------------------------------------------------------------------------------------
   private
   public &
      Green_from_spectral

   interface Green_from_spectral
      module procedure Green_from_spectral_2time,&
         Green_from_spectral_timeslice_keldysh, Green_from_spectral_timediff
   end interface Green_from_spectral      
!------------------------------------------------------------------------------------
   integer,parameter :: m_max = 10,n_max = 500, ntab_max = 800
   real(dp),parameter :: epsabs=1.0e-6_dp,epsrel=0.0_dp   
   real(dp),parameter :: pi = 3.1415926535897932384626433832795_dp
   ! .. global private variables ..
   procedure(spectralfunction),pointer :: spect => null()
!------------------------------------------------------------------------------------
   interface 
     pure function spectralfunction(w) result(s)
       import :: dp
       implicit none
       real(dp),intent(in) :: w
       real(dp) :: s
     end function spectralfunction

     pure function wfunc(w,tid) result(s)
       import :: dp
       implicit none
       real(dp),intent(in) :: w
       integer,intent(in)  :: tid
       real(dp) :: s
     end function wfunc
   end interface    

   type params_t
      integer  :: tid
      integer  :: sig
      real(dp) :: Beta,Mu
   end type params_t
!------------------------------------------------------------------------------------
   type(params_t) :: wrapper_params
!------------------------------------------------------------------------------------
contains
!------------------------------------------------------------------------------------
   pure real(dp) function wrap_gtr_fermi(w,tid)
      real(dp),intent(in) :: w
      integer,intent(in)  :: tid
      wrap_gtr_fermi = spect(w) * nnfermi(wrapper_params%Beta,w-wrapper_params%mu)
   end function wrap_gtr_fermi
!------------------------------------------------------------------------------------
   pure real(dp) function wrap_gtr_bose(w,tid)
      real(dp),intent(in) :: w
      integer,intent(in)  :: tid
      wrap_gtr_bose = -spect(w) * nbose(wrapper_params%Beta,-(w-wrapper_params%mu))
   end function wrap_gtr_bose
!------------------------------------------------------------------------------------
   pure real(dp) function wrap_lsr_fermi(w,tid)
      real(dp),intent(in) :: w
      integer,intent(in)  :: tid
      wrap_lsr_fermi = spect(w) * nfermi(wrapper_params%Beta,w-wrapper_params%mu)
   end function wrap_lsr_fermi
!------------------------------------------------------------------------------------
   pure real(dp) function wrap_lsr_bose(w,tid)
      real(dp),intent(in) :: w
      integer,intent(in)  :: tid
      wrap_lsr_bose = spect(w) * nbose(wrapper_params%Beta,(w-wrapper_params%mu))
   end function wrap_lsr_bose
!------------------------------------------------------------------------------------
   subroutine Green_lesser_fermi(Dw,wmin,wmax,mu,Beta,tgrd,Gles)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      real(dp),intent(in)       :: Beta
      type(grid1d_t),intent(in) :: tgrd
      complex(dp),intent(inout) :: Gles(0:)
      integer :: n
      real(dp) :: fcos,fsin,wmax_
      ! .. quadpack ..
      integer  :: neval,ier
      real(dp) :: abserr

      wmax_ = min(wmax,mu+20.0_dp/Beta)

      spect => Dw
      wrapper_params%Beta = Beta
      wrapper_params%Mu = Mu
      wrapper_params%sig = -1

      !$OMP PARALLEL DO PRIVATE(fcos,fsin,abserr,neval)
      do n=0,tgrd%Npts
         call Filon_quad(wrap_lsr_fermi,wmin,wmax_,tgrd%Xpts(n),1,ntab_max,epsabs,fcos,abserr,neval)
         call Filon_quad(wrap_lsr_fermi,wmin,wmax_,tgrd%Xpts(n),2,ntab_max,epsabs,fsin,abserr,neval)
         Gles(n) = iu*fcos + fsin
      end do
      
   end subroutine Green_lesser_fermi
!------------------------------------------------------------------------------------
   subroutine Green_lesser_bose(Dw,wmin,wmax,mu,Beta,tgrd,Gles)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      real(dp),intent(in)       :: Beta
      type(grid1d_t),intent(in) :: tgrd
      complex(dp),intent(inout) :: Gles(0:)
      integer :: n
      real(dp) :: fcos,fsin,wmax_
      ! .. quadpack ..
      integer  :: neval,ier
      real(dp) :: abserr

      wmax_ = min(wmax,mu+20.0_dp/Beta)

      spect => Dw
      wrapper_params%Beta = Beta
      wrapper_params%Mu = Mu
      wrapper_params%sig = 1
      
      !$OMP PARALLEL DO PRIVATE(fcos,fsin,abserr,neval)
      do n=0,tgrd%Npts
         call Filon_quad(wrap_lsr_bose,wmin,wmax_,tgrd%Xpts(n),1,ntab_max,epsabs,fcos,abserr,neval)
         call Filon_quad(wrap_lsr_bose,wmin,wmax_,tgrd%Xpts(n),2,ntab_max,epsabs,fsin,abserr,neval)
         Gles(n) = -iu*fcos - fsin
      end do

   end subroutine Green_lesser_bose
!------------------------------------------------------------------------------------
   subroutine Green_greater_fermi(Dw,wmin,wmax,mu,Beta,tgrd,Ggtr)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      real(dp),intent(in)       :: Beta
      type(grid1d_t),intent(in) :: tgrd
      complex(dp),intent(inout) :: Ggtr(0:)
      integer :: n
      real(dp) :: fcos,fsin,wmin_
      ! .. quadpack ..
      integer  :: neval,ier
      real(dp) :: abserr

      wmin_ = max(wmin,mu-20.0_dp/Beta)

      spect => Dw
      wrapper_params%Beta = Beta
      wrapper_params%Mu = Mu
      wrapper_params%sig = -1

      !$OMP PARALLEL DO PRIVATE(fcos,fsin,abserr,neval)
      do n=0,tgrd%Npts
         call Filon_quad(wrap_gtr_fermi,wmin_,wmax,tgrd%Xpts(n),1,ntab_max,epsabs,fcos,abserr,neval)
         call Filon_quad(wrap_gtr_fermi,wmin_,wmax,tgrd%Xpts(n),2,ntab_max,epsabs,fsin,abserr,neval)
         Ggtr(n) = -iu*fcos - fsin
      end do

   end subroutine Green_greater_fermi
!------------------------------------------------------------------------------------
   subroutine Green_greater_bose(Dw,wmin,wmax,mu,Beta,tgrd,Ggtr)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      real(dp),intent(in)       :: Beta
      type(grid1d_t),intent(in) :: tgrd
      complex(dp),intent(inout) :: Ggtr(0:)
      integer :: n
      real(dp) :: fcos,fsin,wmin_
      ! .. quadpack ..
      integer  :: neval,ier
      real(dp) :: abserr

      wmin_ = max(wmin,mu-20.0_dp/Beta)

      spect => Dw
      wrapper_params%Beta = Beta
      wrapper_params%Mu = Mu
      wrapper_params%sig = 1

      !$OMP PARALLEL DO PRIVATE(fcos,fsin,abserr,neval)
      do n=0,tgrd%Npts
         call Filon_quad(wrap_gtr_bose,wmin_,wmax,tgrd%Xpts(n),1,ntab_max,epsabs,fcos,abserr,neval)
         call Filon_quad(wrap_gtr_bose,wmin_,wmax,tgrd%Xpts(n),2,ntab_max,epsabs,fsin,abserr,neval)
         Ggtr(n) = -iu*fcos - fsin
      end do

   end subroutine Green_greater_bose
!------------------------------------------------------------------------------------
   subroutine Green_from_spectral_timediff(Dw,wmin,wmax,mu,Beta,FermBos,tgrd,Ggtr,Glsr)   
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu,Beta
      integer,intent(in)        :: FermBos
      type(grid1d_t),intent(in) :: tgrd
      complex(dp),intent(inout) :: Ggtr(0:)
      complex(dp),intent(inout) :: Glsr(0:)
      type(grid1d_t) :: tg

      call tg%Init(2*tgrd%Npts,-tgrd%Xmax,tgrd%Xmax)
      if(FermBos == -1) then
         call Green_greater_fermi(Dw,wmin,wmax,mu,Beta,tg,Ggtr)
         call Green_lesser_fermi(Dw,wmin,wmax,mu,Beta,tg,Glsr)         
      else
         call Green_greater_bose(Dw,wmin,wmax,mu,Beta,tg,Ggtr)
         call Green_lesser_bose(Dw,wmin,wmax,mu,Beta,tg,Glsr) 
      end if

      call tg%Clean()

   end subroutine Green_from_spectral_timediff
!------------------------------------------------------------------------------------
   subroutine Green_from_spectral_2time(Dw,wmin,wmax,mu,tgrd,G,realtime)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      type(grid1d_t),intent(in) :: tgrd
      type(CntrFun_2time_t),intent(inout) :: G
      logical,intent(in),optional :: realtime
      integer :: n,j
      complex(dp),allocatable :: G_les(:),G_gtr(:)
      type(grid1d_t) :: tg

      call assert(tgrd%Npts == G%Nt1)
      call assert(tgrd%Npts == G%Nt2)

      call tg%Init(2*tgrd%Npts,-tgrd%Xmax,tgrd%Xmax)
      allocate(G_les(0:tg%Npts),G_gtr(0:tg%Npts))

      if(G%FermBos == -1) then
         call Green_greater_fermi(Dw,wmin,wmax,mu,G%Beta,tg,G_gtr)
         call Green_lesser_fermi(Dw,wmin,wmax,mu,G%Beta,tg,G_les)         
      else
         call Green_greater_bose(Dw,wmin,wmax,mu,G%Beta,tg,G_gtr)
         call Green_lesser_bose(Dw,wmin,wmax,mu,G%Beta,tg,G_les) 
      end if

      do concurrent(n=0:tgrd%Npts,j=0:tgrd%Npts)
         G%Lesser(1,1,n,j) = G_les(n-j+tgrd%Npts)
         G%Greater(1,1,n,j) = G_gtr(n-j+tgrd%Npts)
      end do

      deallocate(G_les,G_gtr)
      call tg%Clean()

   end subroutine Green_from_spectral_2time
!------------------------------------------------------------------------------------
   subroutine Green_from_spectral_timeslice_keldysh(Dw,wmin,wmax,mu,tgrd,G)
      procedure(spectralfunction) :: Dw
      real(dp),intent(in)       :: wmin
      real(dp),intent(in)       :: wmax
      real(dp),intent(in)       :: mu
      type(grid1d_t),intent(in) :: tgrd
      type(CntrFun_timeslice_t),intent(inout) :: G
      integer :: n
      complex(dp),allocatable :: G_les(:),G_gtr(:)
      type(grid1d_t) :: tg

      call assert(tgrd%Npts == G%Nt)

      allocate(G_les(0:tgrd%Npts),G_gtr(0:tgrd%Npts))

      if(G%FermBos == -1) then
         call Green_greater_fermi(Dw,wmin,wmax,mu,G%Beta,tgrd,G_gtr)
         call Green_lesser_fermi(Dw,wmin,wmax,mu,G%Beta,tgrd,G_les)         
      else
         call Green_greater_bose(Dw,wmin,wmax,mu,G%Beta,tgrd,G_gtr)
         call Green_lesser_bose(Dw,wmin,wmax,mu,G%Beta,tgrd,G_les) 
      end if

      G%Lesser(1,1,0:) = G_les(0:)
      G%Greater(1,1,0:) = G_gtr(0:)

      deallocate(G_les,G_gtr)
 
   end subroutine Green_from_spectral_timeslice_keldysh
!------------------------------------------------------------------------------------
   subroutine Filon_quad(f,a,b,omega,sc_flag,nmax,epsabs,res,abserr,neval,threadid)
      procedure(wfunc) :: f
      real(dp),intent(in) :: a,b,omega
      integer,intent(in) :: sc_flag
      integer,intent(in) :: nmax
      real(dp),intent(in) :: epsabs
      real(dp),intent(out) :: res
      real(dp),intent(out) :: abserr
      integer,intent(out) :: neval
      integer,intent(in),optional :: threadid
      integer :: tid_
      integer :: n,i
      real(dp) :: x,res_old,err
      real(dp),allocatable :: ftab(:)

      tid_ = 0
      if(present(threadid)) tid_ = threadid

      allocate(ftab(nmax))

      res_old = HUGE(1.0_dp)
      neval = 0

      NTABLOOP:do n=51,nmax,50
         do i=1,n
            x = a + (b-a)*(i-1)/dble(n-1)
            ftab(i) = f(x,tid_)
            neval = neval + 1
         end do
         if(sc_flag == 1) then
            call filon_cos(n,ftab,a,b,omega,res)
         else 
            call filon_sin(n,ftab,a,b,omega,res)
         end if
         err = abs(res_old - res)
         if(err < epsabs) then
            exit NTABLOOP
         else
            res_old = res
         end if
      end do NTABLOOP

      abserr = abs(res_old - res)

      deallocate(ftab)


   end subroutine Filon_quad
!------------------------------------------------------------------------------------
   subroutine filon_cos ( ntab, ftab, a, b, t, result )
!*****************************************************************************80
   !
   !! FILON_COS uses Filon's method on integrals with a cosine factor.
   !
   !  Discussion:
   !
   !    The integral to be approximated has the form:
   !
   !      Integral ( A <= X <= B ) F(X) * COS(T*X) dX
   !
   !    where T is user specified.
   !
   !    The function is interpolated over each subinterval by
   !    a parabolic arc.
   !
   !  Modified:
   !
   !    10 February 2006
   !
   !  Author:
   !
   !    John Burkardt
   !
   !  Reference:
   !
   !    Milton Abramowitz, Irene Stegun,
   !    Handbook of Mathematical Functions,
   !    National Bureau of Standards, 1964,
   !    ISBN: 0-486-61272-4,
   !    LC: QA47.A34.
   !
   !    Stephen Chase, Lloyd Fosdick,
   !    An Algorithm for Filon Quadrature,
   !    Communications of the Association for Computing Machinery,
   !    Volume 12, Number 8, August 1969, pages 453-457.
   !
   !    Stephen Chase, Lloyd Fosdick,
   !    Algorithm 353:
   !    Filon Quadrature,
   !    Communications of the Association for Computing Machinery,
   !    Volume 12, Number 8, August 1969, pages 457-458.
   !
   !    Philip Davis, Philip Rabinowitz,
   !    Methods of Numerical Integration,
   !    Second Edition,
   !    Dover, 2007,
   !    ISBN: 0486453391,
   !    LC: QA299.3.D28.
   !
   !  Parameters:
   !
   !    Input, integer ( kind = 4 ) NTAB, the number of data points.
   !    NTAB must be odd, and greater than 1.
   !
   !    Input, real ( kind = 8 ) FTAB(NTAB), contains the value of the function
   !    at A, A+H, A+2*H, ... , B-H, B, where H = (B-A)/(NTAB-1).
   !
   !    Input, real ( kind = 8 ) A, B, the limits of integration.
   !
   !    Input, real ( kind = 8 ) T, the multiplier of the X argument of the cosine.
   !
   !    Output, real ( kind = 8 ) RESULT, the approximate value of the integral.
   !
     implicit none

     integer ( kind = 4 ) ntab

     real ( kind = 8 ) a
     real ( kind = 8 ) alpha
     real ( kind = 8 ) b
     real ( kind = 8 ) beta
     real ( kind = 8 ) c2n
     real ( kind = 8 ) c2nm1
     real ( kind = 8 ) cost
     real ( kind = 8 ) ftab(ntab)
     real ( kind = 8 ) gamma
     real ( kind = 8 ) h
     real ( kind = 8 ) result
     real ( kind = 8 ) sint
     real ( kind = 8 ) t
     real ( kind = 8 ) theta
     real ( kind = 8 ) xtab(ntab)

     if ( a == b ) then
       result = 0.0D+00
       return
     end if
    
     if ( ntab <= 1 ) then
       write ( *, '(a)' ) ' '
       write ( *, '(a)' ) 'FILON_COS - Fatal error!'
       write ( *, '(a)' ) '  NTAB < 2'
       write ( *, '(a,i8)' ) '  NTAB = ', ntab
       stop 1
     end if
    
     if ( mod ( ntab, 2 ) /= 1 ) then
       write ( *, '(a)' ) ' '
       write ( *, '(a)' ) 'FILON_COS - Fatal error!'
       write ( *, '(a)' ) '  NTAB must be odd.'
       write ( *, '(a,i8)' ) '  NTAB = ', ntab
       stop 1
     end if
   !
   !  Set up a vector of the NTAB X values.
   ! 
     call r8vec_even ( ntab, a, b, xtab )

     h = ( b - a ) / real ( ntab - 1, kind = 8 )

     theta = t * h
     sint = sin ( theta )
     cost = cos ( theta )

     if ( 6.0D+00 * abs ( theta ) <= 1.0D+00 ) then

       alpha = 2.0D+00 * theta**3 /   45.0D+00 &
             - 2.0D+00 * theta**5 /  315.0D+00 &
             + 2.0D+00 * theta**7 / 4725.0D+00
     
       beta =  2.0D+00            /     3.0D+00 &
             + 2.0D+00 * theta**2 /    15.0D+00 &
             - 4.0D+00 * theta**4 /   105.0D+00 &
             + 2.0D+00 * theta**6 /   567.0D+00 &
             - 4.0D+00 * theta**8 / 22275.0D+00

       gamma = 4.0D+00            /      3.0D+00 &
             - 2.0D+00 * theta**2 /     15.0D+00 &
             +           theta**4 /    210.0D+00 &
             -           theta**6 /  11340.0D+00

     else

       alpha = ( theta**2 + theta * sint * cost &
         - 2.0D+00 * sint**2 ) / theta**3

       beta = ( 2.0D+00 * theta + 2.0D+00 * theta * cost**2 &
         - 4.0D+00 * sint * cost ) / theta**3

       gamma = 4.0D+00 * ( sint - theta * cost ) / theta**3
     
     end if

     c2n = sum ( ftab(1:ntab:2) * cos ( t * xtab(1:ntab:2) ) ) &
       - 0.5D+00 * ( ftab(ntab) * cos ( t * xtab(ntab) ) &
                   + ftab(1) * cos ( t * xtab(1) ) )

     c2nm1 = sum ( ftab(2:ntab-1:2) * cos ( t * xtab(2:ntab-1:2) ) )
    
     result = h * ( &
         alpha * ( ftab(ntab) * sin ( t * xtab(ntab) ) & 
                 - ftab(1)    * sin ( t * xtab(1) ) ) &
       + beta * c2n &
       + gamma * c2nm1 )

   end subroutine filon_cos



   subroutine filon_sin ( ntab, ftab, a, b, t, result )

!*****************************************************************************80
   !
   !! FILON_SIN uses Filon's method on integrals with a sine factor.
   !
   !  Discussion:
   !
   !    The integral to be approximated has the form
   !
   !      Integral ( A <= X <= B ) F(X) * SIN(T*X) dX
   !
   !    where T is user specified.
   !
   !    The function is interpolated over each subinterval by
   !    a parabolic arc.
   !
   !  Modified:
   !
   !    10 February 2006
   !
   !  Author:
   !
   !    John Burkardt
   !
   !  Reference:
   !
   !    Milton Abramowitz, Irene Stegun,
   !    Handbook of Mathematical Functions,
   !    National Bureau of Standards, 1964,
   !    ISBN: 0-486-61272-4,
   !    LC: QA47.A34.
   !
   !    Stephen Chase, Lloyd Fosdick,
   !    An Algorithm for Filon Quadrature,
   !    Communications of the Association for Computing Machinery,
   !    Volume 12, Number 8, August 1969, pages 453-457.
   !
   !    Stephen Chase, Lloyd Fosdick,
   !    Algorithm 353:
   !    Filon Quadrature,
   !    Communications of the Association for Computing Machinery,
   !    Volume 12, Number 8, August 1969, pages 457-458.
   !
   !    Philip Davis, Philip Rabinowitz,
   !    Methods of Numerical Integration,
   !    Second Edition,
   !    Dover, 2007,
   !    ISBN: 0486453391,
   !    LC: QA299.3.D28.
   !
   !  Parameters:
   !
   !    Input, integer ( kind = 4 ) NTAB, the number of data points, 
   !    including the endpoints.  NTAB must be odd, and greater than 1.
   !
   !    Input, real ( kind = 8 ) FTAB(NTAB), contains the value of the function
   !    at A, A+H, A+2*H, ... , B-H, B, where H = (B-A)/(NTAB-1).
   !
   !    Input, real ( kind = 8 ) A, B, the limits of integration.
   !
   !    Input, real ( kind = 8 ) T, multiplier of the X argument of the sine.
   !
   !    Output, real ( kind = 8 ) RESULT, the approximate value of the integral.
   !
     implicit none

     integer ( kind = 4 ) ntab

     real ( kind = 8 ) a
     real ( kind = 8 ) alpha
     real ( kind = 8 ) b
     real ( kind = 8 ) beta
     real ( kind = 8 ) cost
     real ( kind = 8 ) ftab(ntab)
     real ( kind = 8 ) gamma
     real ( kind = 8 ) h
     real ( kind = 8 ) result
     real ( kind = 8 ) s2n
     real ( kind = 8 ) s2nm1
     real ( kind = 8 ) sint
     real ( kind = 8 ) t
     real ( kind = 8 ) theta
     real ( kind = 8 ) xtab(ntab)

     if ( a == b ) then
       result = 0.0D+00
       return
     end if
    
     if ( ntab <= 1 ) then
       write ( *, '(a)' ) ' '
       write ( *, '(a)' ) 'FILON_SIN - Fatal error!'
       write ( *, '(a)' ) '  NTAB < 2'
       write ( *, '(a,i8)' ) '  NTAB = ',ntab
       stop 1
     end if
    
     if ( mod ( ntab, 2 ) /= 1 ) then
       write ( *, '(a)' ) ' '
       write ( *, '(a)' ) 'FILON_SIN - Fatal error!'
       write ( *, '(a)' ) '  NTAB must be odd.'
       write ( *, '(a,i8)' ) '  NTAB = ',ntab
       stop 1
     end if
   !
   !  Set up a vector of the NTAB X values.
   ! 
     call r8vec_even ( ntab, a, b, xtab )

     h = ( b - a ) / real ( ntab - 1, kind = 8 )
     theta = t * h

     sint = sin ( theta )
     cost = cos ( theta )

     if ( 6.0D+00 * abs ( theta ) <= 1.0D+00 ) then

       alpha = 2.0D+00 * theta**3 /   45.0D+00 &
             - 2.0D+00 * theta**5 /  315.0D+00 &
             + 2.0D+00 * theta**7 / 4725.0D+00
     
       beta =  2.0D+00            /     3.0D+00 &
             + 2.0D+00 * theta**2 /    15.0D+00 &
             - 4.0D+00 * theta**4 /   105.0D+00 &
             + 2.0D+00 * theta**6 /   567.0D+00 &
             - 4.0D+00 * theta**8 / 22275.0D+00

       gamma = 4.0D+00            /      3.0D+00 &
             - 2.0D+00 * theta**2 /     15.0D+00 &
             +           theta**4 /    210.0D+00 &
             -           theta**6 /  11340.0D+00

     else
    
       alpha = ( theta**2 + theta * sint * cost &
         - 2.0D+00 * sint**2 ) / theta**3

       beta = ( 2.0D+00 * theta + 2.0D+00 * theta * cost**2 &
         - 4.0D+00 * sint * cost ) / theta**3

       gamma = 4.0D+00 * ( sint - theta * cost ) / theta**3
    
     end if
     
     s2n = sum ( ftab(1:ntab:2) * sin ( t * xtab(1:ntab:2) ) ) &
       - 0.5D+00 * ( ftab(ntab) * sin ( t * xtab(ntab) ) &
                   + ftab(1) * sin ( t * xtab(1) ) )

     s2nm1 = sum ( ftab(2:ntab-1:2) * sin ( t * xtab(2:ntab-1:2) ) )

     result = h * ( &
         alpha * ( ftab(1) * cos ( t * xtab(1) ) &
                 - ftab(ntab) * cos ( t * xtab(ntab) ) ) &
       + beta * s2n &
       + gamma * s2nm1 )

   end subroutine filon_sin


   subroutine r8vec_even ( n, alo, ahi, a )

   !*****************************************************************************80
   !
   !! R8VEC_EVEN returns N values, evenly spaced between ALO and AHI.
   !
   !  Modified:
   !
   !    17 February 2004
   !
   !  Author:
   !
   !    John Burkardt
   !
   !  Parameters:
   !
   !    Input, integer ( kind = 4 ) N, the number of values.
   !
   !    Input, real ( kind = 8 ) ALO, AHI, the low and high values.
   !
   !    Output, real ( kind = 8 ) A(N), N evenly spaced values.
   !    Normally, A(1) = ALO and A(N) = AHI.
   !    However, if N = 1, then A(1) = 0.5*(ALO+AHI).
   !
     implicit none

     integer ( kind = 4 ) n

     real ( kind = 8 ) a(n)
     real ( kind = 8 ) ahi
     real ( kind = 8 ) alo
     integer ( kind = 4 ) i

     if ( n == 1 ) then

       a(1) = 0.5D+00 * ( alo + ahi )

     else

       do i = 1, n
         a(i) = ( real ( n - i,     kind = 8 ) * alo   &
                + real (     i - 1, kind = 8 ) * ahi ) &
                / real ( n     - 1, kind = 8 )
       end do

     end if

   end subroutine r8vec_even

!======================================================================================
end module kbkit_spectral
