import os
import numpy as np
import scipy.linalg as la
from scipy.interpolate import UnivariateSpline
import h5py
import surface as surf
import surface_kp as surf_kp
import surface_dirac as surf_dirac
import surface_5ql as surf_5ql
import surface_parabolic as surf_par
import MatrixElement as matel
#--------------------------------------------------------------------
def GetPESMatel_kp(kpt,pars):
    b0 = pars['b0'] 
    dk = pars['dk']

    ka = np.sqrt((kpt[0])**2 + kpt[1]**2)
    mel = b0 * np.exp(-0.5*(ka/dk)**2) * np.ones(2)   

    return mel
#--------------------------------------------------------------------
def GetPESMatel_5QL(kpt,pars,band_basis=True):
    a0 = pars['a0']
    b0 = pars['b0']
    dk = pars['dk']

    ka = np.sqrt((kpt[0])**2 + kpt[1]**2)

    if band_basis:
        mel = (a0 + b0 * np.exp(-0.5*(ka/dk)**2)) * np.ones(6)
    else:
        mel_orb = (a0 + b0 * np.exp(-0.5*(ka/dk)**2)) * np.ones(6)
               
        bite = surf_5ql.BiSe_Surf_5QL()
        Hk = bite.get_ham(kpt,band_basis=False)
        eps, vect = la.eigh(Hk)

        mel = np.einsum('ja,j->a', vect, mel_orb)
    
    return mel
#--------------------------------------------------------------------
def GetPESMatel_3bnd(kpt,pars,band_basis=True):
    a0 = pars['a0']
    b0 = pars['b0']
    dk = pars['dk']

    ka = np.sqrt((kpt[0])**2 + kpt[1]**2)

    if band_basis:
        mel = (a0 + b0 * np.exp(-0.5*(ka/dk)**2)) * np.ones(3)
    else:
        mel_orb = (a0 + b0 * np.exp(-0.5*(ka/dk)**2)) * np.ones(3)
               
        bite = surf_5ql.BiSe_Surf_5QL()
        Hk = bite.get_ham(kpt,band_basis=False)
        eps, vect = la.eigh(Hk)

        mel_5ql = np.einsum('ja,j->a', vect, mel_orb)
        mel = np.array([mel_5ql[0], mel_5ql[1], mel_5ql[-1]])
    
    return mel
#--------------------------------------------------------------------
def GetHam(kpt,dk=1.0e-4):
    nst = 2
    Ry = 27.211386
    aB = 0.52917721

    alat = 4.14 / aB
    A0 = -0.0255 / Ry
    A11 = 0.1937 / Ry
    A12 = 0.224 / Ry
    A14 = 0.0551 / Ry
    m11 = -1.6978 / Ry
    B0 = 0.0164 / Ry
    B11 = 0.1203 / Ry
    B12 = 0.3263 / Ry
    B14 = 0.0

    Apar = A0, A11, A12, A14
    Bpar = B0, B11, B12, B14 

    bise = surf.BiSe_Surf(alat, Apar, Bpar, m11)

    Hk = bise.get_ham(kpt)
    eps, vect = la.eigh(Hk)

    velo = bise.get_velocity(kpt,dk=dk)

    return eps, velo
#--------------------------------------------------------------------
def GenHam(prefix,kpt,dk=1.0e-4):
    nst = 2

    eps, velo = GetHam(kpt,dk=dk)
    
    H0 = np.zeros([nst, nst], dtype=np.complex_)
    Pmom = np.zeros([nst, nst, 3], dtype=np.complex_)

    H0[0,0] = eps[0]; H0[1,1] = eps[1]

    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]

    file_ham = prefix + "_ham.h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------
def GenGamma(prefix,gam):
    nst = 2
    gmm = np.zeros([nst, nst], dtype=np.complex_)

    gmm = gmm + gam * np.eye(nst)
    file_gmm = prefix + "_gamma{}.h5".format(np.round(gam,decimals=6))
    matel.store_ham(file_gmm, gmm, rewrite=False)

    return file_gmm
#--------------------------------------------------------------------
def GetHam_kp(kpt,kscale=1.0,inter_band=True):
    bise = surf_kp.BiSe_Surf_kp(scale=kscale)

    Hk = bise.get_ham(kpt)
    velo = bise.get_velocity_gk(kpt,inter_band=inter_band)

    return Hk, velo
#--------------------------------------------------------------------
def GetHam_kp_bulk(kpt,vsc,nbulk=2):
    Ry = 27.211386
    Eshift = 0.176 / Ry
    nbnd = nbulk + 2

    Hk = np.zeros([nbnd,nbnd], dtype=np.complex_)
    velo = np.zeros([nbnd,nbnd,2], dtype=np.complex_)

    bite = surf_kp.BiSe_Surf_kp()
    Hk_surf = bite.get_ham(kpt)
    vk_surf = bite.get_velocity_gk(kpt)

    epsk_surf, vectk = la.eigh(Hk_surf)

    Hk[0:2,0:2] = np.diag(epsk_surf)
    velo[0:2,0:2,:] = np.einsum('ia,ijr,jb->abr', np.conj(vectk),vk_surf,vectk)

    data = np.loadtxt("data/bands_dft.dat")
    kxs = data[:,0]
    Ecb1 = data[:,5] + Eshift
    Ecb2 = data[:,6] + Eshift

    spl1 = UnivariateSpline(kxs, Ecb1, s=0.0)
    spl1_der = spl1.derivative()
    spl2 = UnivariateSpline(kxs, Ecb2, s=0.0)
    spl2_der = spl2.derivative()

    als = np.linspace(0.0, 1.0, nbulk)
    for i in range(nbulk):
        Hk[i+2,i+2] = (1.0 - als[i]) * spl1(kpt[0]) + als[i] * spl2(kpt[0])
        velo[i+2,i+2,0] = (1.0 - als[i]) * spl1_der(kpt[0]) + als[i] * spl2_der(kpt[0])
        velo[i+2,i+2,1] = (1.0 - als[i]) * spl1_der(kpt[1]) + als[i] * spl2_der(kpt[1])
        velo[1,i+2,0] = vsc
        velo[i+2,1,0] = vsc
        velo[1,i+2,1] = vsc
        velo[i+2,1,1] = vsc

    return Hk, velo
#--------------------------------------------------------------------
def GenHam_kp(prefix,kpt,kscale=1.0,vz=0.0,band_basis=False,inter_band=True):
    nst = 2

    H0, velo = GetHam_kp(kpt,kscale=kscale,inter_band=inter_band)
    if band_basis:
        epsk, vectk = la.eigh(H0)
        H0 = np.zeros([nst,nst])
        H0 = H0 + np.diag(epsk)
        velo = np.einsum('ia,ijr,jb->abr', np.conj(vectk),velo,vectk)

    Pmom = np.zeros([nst, nst, 3], dtype=np.complex_)
    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]
    Pmom[:,:,2] = vz * np.eye(nst)

    file_ham = prefix + "_kp_ham.h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_kp_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------
def GenHam_kp_bulk(prefix,kpt,vsc,nbulk=2):
    nbnd = nbulk + 2

    H0, velo = GetHam_kp_bulk(kpt,vsc,nbulk=nbulk)

    Pmom = np.zeros([nbnd, nbnd, 3], dtype=np.complex_)
    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]

    file_ham = prefix + "_kp_ham.h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_kp_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------
def GenHam_5QL(prefix,kpt,vsb1,vsb2):
    nbnd = 6

    bite = surf_5ql.BiSe_Surf_5QL(vsb=vsb1,vsb2=vsb2)
    H0 = bite.get_ham(kpt)
    velo = bite.get_velocity(kpt)

    Pmom = np.zeros([nbnd, nbnd, 3], dtype=np.complex_)
    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]

    file_ham = prefix + "_5ql_ham.h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_5ql_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------
def GenPESMatel_kp(prefix,kpt,pars):
    mel = GetPESMatel_kp(kpt,pars)

    file_mel = prefix + "_kp_matel_b{}_dk{}.h5".\
        format(pars['b0'],pars['dk'])

    if not os.path.isfile(file_mel):
        f = h5py.File(file_mel, "w")
        f.attrs['nst'] = 2
        f.create_dataset('real', data=np.real(mel))
        f.create_dataset('imag', data=np.imag(mel))
        f.close()

    return file_mel
#--------------------------------------------------------------------
def GenPESMatel_5QL(prefix,kpts,pars,band_basis=True):
    nk = kpts.shape[0]
    mels = np.zeros([6,nk], dtype=np.complex_)

    for ik, kpt in enumerate(kpts):
        mels[:,ik] = GetPESMatel_5QL(kpt,pars,band_basis=band_basis)

    file_mel = prefix + "_5ql_matel_a{}_b{}_dk{}_nk{}.h5".\
        format(pars['a0'],pars['b0'],pars['dk'],nk)

    #if not os.path.isfile(file_mel):
    f = h5py.File(file_mel, "w")
    f.attrs['nst'] = 6
    f.attrs['nk'] = nk
    f.create_dataset('real', data=np.real(mels.T))
    f.create_dataset('imag', data=np.imag(mels.T))
    f.close()

    return file_mel
#--------------------------------------------------------------------
def GenPESMatel_3bnd(prefix,kpts,pars,band_basis=True):
    nk = kpts.shape[0]
    mels = np.zeros([3,nk], dtype=np.complex_)

    for ik, kpt in enumerate(kpts):
        mels[:,ik] = GetPESMatel_3bnd(kpt,pars,band_basis=band_basis)

    file_mel = prefix + "_3bnd_matel_a{}_b{}_dk{}_nk{}.h5".\
        format(pars['a0'],pars['b0'],pars['dk'],nk)

    #if not os.path.isfile(file_mel):
    f = h5py.File(file_mel, "w")
    f.attrs['nst'] = 3
    f.attrs['nk'] = nk
    f.create_dataset('real', data=np.real(mels.T))
    f.create_dataset('imag', data=np.imag(mels.T))
    f.close()

    return file_mel
#--------------------------------------------------------------------
def GenTmat_kp_bulk(prefix,nbulk,T2_ss,T2_sb,T1_bb,T1_ss=None):
    nbnd = nbulk + 2

    T2_bb = 2 * T1_bb

    Tmat = 1.0e8 * np.ones([nbnd, nbnd])
    Tmat[0,1] = T2_ss
    Tmat[1,0] = T2_ss
    if not T1_ss == None:
        Tmat[0,0] = T1_ss
        Tmat[1,1] = T1_ss
    for i in range(nbulk):
        Tmat[0,i+2] = T2_sb
        Tmat[i+2,0] = T2_sb
        Tmat[1,i+2] = T2_sb
        Tmat[i+2,1] = T2_sb
        Tmat[i+2,i+2] = T1_bb
        for j in range(nbulk):
            if not i == j:
                Tmat[i+2,j+2] = T2_bb

    file_tmat = prefix + "_kp_bulk_tmat.h5"
    matel.store_tmat(file_tmat, Tmat, rewrite=False)

    return file_tmat
#--------------------------------------------------------------------
def GenTmat_5ql(prefix,T2_ss,T2_sb,T1_bb,T1_ss=None):
    nbulk = 4
    nbnd = nbulk + 2

    T2_bb = 2 * T1_bb

    Tmat = 1.0e8 * np.ones([nbnd, nbnd])
    Tmat[0,1] = T2_ss
    Tmat[1,0] = T2_ss
    if not T1_ss == None:
        Tmat[0,0] = T1_ss
        Tmat[1,1] = T1_ss
    
    for i in range(nbulk):
        Tmat[0,i+2] = T2_sb
        Tmat[i+2,0] = T2_sb
        Tmat[1,i+2] = T2_sb
        Tmat[i+2,1] = T2_sb

    Tmat[2,2] = T1_bb
    Tmat[3,3] = T1_bb
    Tmat[4,4] = T1_ss
    Tmat[5,5] = T1_ss

    for j in range(nbulk):
        if not i == j:
            Tmat[i+2,j+2] = T2_bb

    file_tmat = prefix + "_5ql_tmat.h5"
    matel.store_tmat(file_tmat, Tmat, rewrite=False)

    return file_tmat
#--------------------------------------------------------------------
def GetHam_Dirac(kpt,Eshift=0.0):
    Ry = 27.211386
    aB = 0.52917721
    mu = 0.32 / Ry + Eshift
    kF = 0.083 * aB
    eF = 0.32 / Ry
    vF = eF / kF

    bise = surf_dirac.BiSe_Surf(vF, Eshift=Eshift)

    Hk = bise.get_ham(kpt)
    velo = bise.get_velocity()

    return Hk, velo
#--------------------------------------------------------------------
def GenHam_Dirac(prefix,kpt,Eshift=0.0):
    nst = 2

    H0, velo = GetHam_Dirac(kpt,Eshift=Eshift)

    Pmom = np.zeros([nst, nst, 3], dtype=np.complex_)
    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]

    file_ham = prefix + "_dirac_ham"
    if np.abs(Eshift) > 1.0e-10:
        file_ham += "_Eshift{}".format(np.round(Eshift,decimals=5))
    file_ham += ".h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_dirac_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------     
def GenHam_Parabolic(prefix,kpt,E0=0.0,a0=1.0):
    nst = 1

    model = surf_par.Surf_Parabolic(E0,a0)
    H0 = model.get_ham(kpt)
    velo = model.get_velocity(kpt)
    Pmom = np.zeros([nst, nst, 3], dtype=np.complex_)
    Pmom[:,:,0] = velo[:,:,0]
    Pmom[:,:,1] = velo[:,:,1]

    file_ham = prefix + "_parabolic_ham.h5"
    matel.store_ham(file_ham, H0, rewrite=False)

    file_pmom = prefix + "_parabolic_pmom.h5"
    matel.store_pmom(file_pmom, Pmom, rewrite=False)

    return file_ham, file_pmom
#--------------------------------------------------------------------  
