import sys
import os
import numpy as np
import scipy.linalg as la
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
from matplotlib import cm
from ColormapARPES import cmap_experiment
import surface_kp as surf_kp
from constants import tfs, aB, Ry
#--------------------------------------------------------------
def PlotOcck(ts,ks,occk,kmax,tmin,tmax,pmax=1.0,tshift=0.0,asp=1.0,fout=""):
    mycmap = cmap_experiment()

    asp_ = asp * tfs*(ts[-1] - ts[0])/(ks[-1]/aB - ks[0]/aB)
    
    fig, ax = plt.subplots()
    ax.set_xlabel(r"$t\, $ (fs)")
    ax.set_ylabel(r"$k_y\, (\AA^{-1})$")

    t1 = tfs*ts[0] - tshift
    t2 = tfs*ts[-1] - tshift
    im = ax.imshow(occk.T, origin="lower", extent=(t1, t2, ks[0]/aB, ks[-1]/aB),\
                   vmin=0.0, vmax=pmax, cmap=mycmap, interpolation='bilinear', aspect=asp_)

    ax.set_xlim(tmin,tmax)
    ax.set_ylim(-kmax,kmax)
    
    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig)
#--------------------------------------------------------------
def PlotKcut(ws,spectk,Emin,Emax,Eshift=0.0,iprobe=0,fout=""):
    Ry = 27.21138602
    fig, ax = plt.subplots()

    cols = cm.coolwarm(np.linspace(0.0,1.0,spectk.shape[-1]))

    '''
    for n in range(spectk.shape[-1]):
        ax.plot(Ry*ws-Eshift, spectk[:,n], c=cols[n])
    '''

    ax.plot(Ry*ws-Eshift, spectk[:,iprobe], c=cols[iprobe])

    ax.set_xlim(Emin,Emax)
    
    ax.set_xlabel(r"$\epsilon - \epsilon_\mathrm{F}\,$ (eV)")

    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig) 
#--------------------------------------------------------------
def PlotSpect(ws,ks,probe_t0,spectk,kmax,Emin,Emax,epsk,\
    fac=1.0,probe_indx=0,Eshift=0.0,Avec=0.0,w0=0.0,der2=False,fout=""):
    Ry = 27.21138602
    aB = 0.52917721

    nk = spectk.shape[0]
    nprobe = spectk.shape[-1]

    if der2:
        for ik in range(nk):
            for n in range(nprobe):
                spl = UnivariateSpline(ws, spectk[ik,:,n], k=4, s=0)
                der_spl = spl.derivative(n=2)
                spectk[ik,:,n] = -der_spl(ws)

    smax = fac * np.amax(spectk)
    # smax = 2.0e5
    #mycmap = cmap_arpes3()
    mycmap = cmap_experiment()
    # mycmap = cm.bone_r

    fig, ax = plt.subplots()
    # ax.axvline(x=0.0,c='k',ls='--')
    # ax.axhline(y=0.0,c='dimgray',ls='--')
    ax.axhline(y=0.0,c='red',ls='--')
    ax.set_ylabel(r"$\epsilon - \epsilon_\mathrm{F}\,$ (eV)")
    ax.set_xlabel(r"$k_y \, (\AA^{-1})$")

    im = ax.imshow(spectk[:,:,probe_indx].T, origin="lower", extent=(ks[0]/aB,ks[-1]/aB,Ry*ws[0]-Eshift,Ry*ws[-1]-Eshift),\
                  vmin=0.0, vmax=smax, cmap=mycmap, interpolation='none', aspect=0.8)


    if w0 > 0.0:
        for i in range(epsk.shape[1]):
            for j in range(-1,3):
                if j == 0:
                    l = 2
                else:
                    l = 1
                ax.plot(ks/aB, Ry * epsk[:,i] - Eshift + j * w0, c='grey', lw=l)
    else:
        for i in range(epsk.shape[1]):
            ax.plot(ks/aB, Ry * epsk[:,i] - Eshift, c='grey')
                

    ax.set_xlim(-kmax, kmax)
    ax.set_ylim(Emin, Emax)

    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig) 
#--------------------------------------------------------------
def PlotSpectAll(ws,ks,probe_t0,spectk,kmax,Emin,Emax,epsk=[],\
	         w0=0.0,fac=1.0,Eshift=0.0,probe_range=[],title="",fout=""):
    mycmap = cmap_experiment()

    nk = spectk.shape[0]
    nprobe = spectk.shape[-1]

    pr_range = np.arange(0,nprobe,step=1)
    if len(probe_range) > 0:
        pr_range = probe_range

    npr = len(pr_range)
    
    smax = fac * np.amax(spectk)

    fig, ax = plt.subplots(1,npr,sharey=True,figsize=(2.5*npr,3.5))

    ax[0].set_ylabel(r"$\epsilon - \epsilon_\mathrm{F}\,$ (eV)")

    for n in range(npr):
        ax[n].axhline(y=0.0,c='dimgrey',ls='--')
        ax[n].set_xlabel(r"$k_y\, (\AA^{-1})$")

        im = ax[n].imshow(spectk[:,:,n].T, origin="lower", extent=(ks[0]/aB,ks[-1]/aB,Ry*ws[0]-Eshift,Ry*ws[-1]-Eshift),\
                          vmin=0.0, vmax=smax, cmap=mycmap, interpolation='bicubic', aspect=0.7)

        if w0 > 0.0:
            for i in range(epsk.shape[1]):
                for j in range(-1,3):
                    if j == 0:
                        l = 2
                    else:
                        l = 1
                        
                    ax[n].plot(ks/aB, Ry * epsk[:,i] - Eshift + j * w0, c='grey', lw=l)
        else:
            for i in range(epsk.shape[1]):
                ax[n].plot(ks/aB, Ry * epsk[:,i] - Eshift, c='grey', lw=2)

        ax[n].set_xlim(-kmax, kmax)
        ax[n].set_title(r"$t = {} \, $ fs".format(probe_t0[n]))

    ax[0].set_ylim(Emin, Emax)
    if len(title) > 0:
        fig.suptitle(title, fontsize="x-large")

    plt.tight_layout()

    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig)

#--------------------------------------------------------------
def PlotSpectMatrix(ws,ks,probe_t0,spectks,kmax,Emin,Emax,epsk,\
	            w0=0.0,fac=1.0,Eshift=0.0,probe_range=[],\
                    title="",fout=""):

    mycmap = cmap_experiment()
  
    nsp = spectks.shape[0]
    nk = spectks.shape[1]
    nprobe = spectks.shape[-1]

    pr_range = np.arange(0,nprobe,step=1)
    if len(probe_range) > 0:
        pr_range = probe_range

    npr = len(pr_range)
    
    smin = 0.0
    smax = fac * np.amax(spectks)

    fig, ax = plt.subplots(nsp,npr,sharex=True,sharey=True,figsize=(2*npr,3.5*nsp))

    for isp in range(nsp):
        ax[isp,0].set_ylabel(r"$\epsilon - \epsilon_\mathrm{F}\,$ (eV)")

    
    for n in range(npr):
        ax[-1,n].set_xlabel(r"$k_y\, (\AA^{-1})$")

        kr = np.array([ks[0], ks[-1]]) / aB
        
        for isp in range(nsp):
            ax[isp,n].axhline(y=0.0,c='dimgray',ls='--')

            im = ax[isp,n].imshow(spectks[isp,:,:,n].T, origin="lower", extent=(kr[0],kr[1],Ry*ws[0]-Eshift,Ry*ws[-1]-Eshift),\
                                  vmin=0.0, vmax=smax, cmap=mycmap, interpolation='bicubic', aspect=0.7)
                
            if w0 > 0.0:
                for i in range(epsk.shape[1]):
                    for j in range(-1,3):
                        if j == 0:
                            l = 2
                        else:
                            l = 1
                        ax[isp,n].plot(ks/aB, Ry * epsk[:,i] - Eshift + j * w0, c='grey', lw=l)
            else:
                for i in range(epsk.shape[1]):
                    ax[isp,n].plot(ks/aB, Ry * epsk[:,i] - Eshift, c='grey', lw=1)

        ax[0,n].set_title(r"$t = {} \, $ fs".format(probe_t0[n]))
        #ax[0,n].text(0.0, 0.8*Emax, r"$t = {} \, $ fs".format(probe_t0[n]), horizontalalignment='center')
        
        ax[-1,n].set_xlim(-kmax, kmax)
        if len(title) > 0:
            fig.suptitle(title, fontsize="x-large")

    for isp in range(nsp):
        ax[isp,0].set_ylim(Emin, Emax)

    plt.tight_layout()

    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig)
#--------------------------------------------------------------
def PlotSpectMatrix_SingleDelay(ws,ks,spectks,kmax,Emin,Emax,epsk,\
	            w0=0.0,fac=1.0,Eshift=0.0,kshift=[],fout=""):
    Ry = 27.21138602
    aB = 0.52917721

    # smax = 2.0e5
    #mycmap = cmap_arpes3()
    mycmap = cmap_experiment()
    # mycmap = GetCmap()
    # mycmap = cm.bone_r

    nsp = spectks.shape[0]
    nk = spectks.shape[1]
    npr = spectks.shape[-1]

    smin = 0.0
    smax = fac * np.amax(spectks)

    fig, ax = plt.subplots(nsp,npr,sharex=True,sharey=True,figsize=(2.8*npr,3.5*nsp))

    for isp in range(nsp):
        ax[isp,0].set_ylabel(r"$\epsilon - \epsilon_\mathrm{F}\,$ (eV)")

    
    for n in range(npr):
        ax[-1,n].set_xlabel(r"$k_y\, (\AA^{-1})$")

        kr = np.array([ks[0], ks[-1]]) / aB
        if len(kshift) > 0:
            kr += -kshift[n]
        
        for isp in range(nsp):
            ax[isp,n].axhline(y=0.0,c='red',ls='--')

            im = ax[isp,n].imshow(spectks[isp,:,:,n].T, origin="lower", extent=(kr[0],kr[1],Ry*ws[0]-Eshift,Ry*ws[-1]-Eshift),\
                                  vmin=0.0, vmax=smax, cmap=mycmap, interpolation='bicubic', aspect=0.7)
                
            if w0 > 0.0:
                for i in range(epsk.shape[1]):
                    for j in range(-1,3):
                        if j == 0:
                            l = 2
                        else:
                            l = 1
                        ax[isp,n].plot(ks/aB, Ry * epsk[:,i] - Eshift + j * w0, c='grey', lw=l)
            else:
                for i in range(epsk.shape[1]):
                    ax[isp,n].plot(ks/aB, Ry * epsk[:,i] - Eshift, c='grey', lw=1)
                        

        #ax[0,n].set_title(r"$t = {} \, $ fs".format(probe_t0[n]))
        ax[-1,n].set_xlim(-kmax, kmax)

    for isp in range(nsp):
        ax[isp,0].set_ylim(Emin, Emax)

    plt.tight_layout()

    if len(fout) > 0:
        plt.savefig(fout, bbox_inches="tight")
    else:
        plt.show()

    plt.close(fig)
#--------------------------------------------------------------
