import sys
import os
import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
#--------------------------------------------------------------
# t (fs), frequency (THz), duration (fs), amplitude (kV/cm), phase (/pi) offset (fs), fchirp (THz)
def waveform(t_, frequency_, duration_, amplitude_, phase_, offset_, fchirp_):
    return amplitude_ * np.cos(2 * np.pi * (frequency_ * 1e-3) * (fchirp_ * 1e-3 * (t_ - offset_) + 1) * (t_ - offset_)
                               - phase_ * np.pi) \
           * np.exp(-(((t_ - offset_) * 2 * np.sqrt(np.log(2))) / duration_) ** 2)
#--------------------------------------------------------------
def box_win(t1,t2,ts):
    tx = ts - t1
    tq = t2 - t1
    phi = np.pi * tx/tq

    f = np.zeros(len(ts))
    It, = np.where((tx > 0.0) & (tx < tq))
    f[It] = 0.5 - 0.75 * np.cos(phi[It]) + 0.25 * np.cos(phi[It])**3
    It, = np.where(tx >= tq)
    f[It] = 1.0
    return f
#--------------------------------------------------------------
def gauss_mod(Tdur,t):
    tm = 0.5 * Tdur
    r1 = box_win(0.0, Tdur/6.0, t)
    r2 = 1.0 - box_win(5./6. * Tdur, Tdur, t)
    return r1 * r2 * np.exp(-0.5*(6*(t-tm)/Tdur)**2)
#--------------------------------------------------------------
def trapz_pulse(F0,w0,Tdur,tramp,t):
    r1 = box_win(0, tramp, t)
    r2 = 1.0 - box_win(Tdur - tramp, Tdur, t)
    fp = F0 * r1 * r2 * np.sin(w0 * t)
    return fp
#--------------------------------------------------------------
def GetProbeDur(fwhm,pulse_shape=1):
    if pulse_shape == 1:
        return 3.0/np.sqrt(2.0*np.log(2.0)) * fwhm
    elif pulse_shape == 2:
        return 8*0.567296 * fwhm
    elif pulse_shape == 3:
        return 5 * fwhm
    else:
        return 2*fwhm
#--------------------------------------------------------------
def deriv(ts,f):
    spl = UnivariateSpline(ts, f, s=0)
    dspl = spl.derivative()
    return dspl(ts)
#---------------------------------------------------------------------------
def quad_rk4(t,y):
    dt = t[1]-t[0]
    integ = np.zeros(len(t))
    for j in range(0,len(t)-1):
            y12 = y(t[j] + 0.5*dt)
            integ[j+1] = integ[j] + dt/6.0*(y(t[j]) + 4.*y12 + y(t[j+1]))
    return integ
#---------------------------------------------------------------------------
def gauss_pulse(F0,w0,t0,Tdur,t,phi=np.pi/2.,asym=0.0,asym_t0=0.0,asym_dur=100):
    f = F0 * np.cos(w0*(t-t0)-phi) * np.exp(-0.5*((t-t0)/Tdur)*((t-t0)/Tdur))
    if asym > 0.0:
        b = box_win(asym_t0 - asym_dur, asym_t0, t)
        fac = (1.0 + asym) - asym * b
        f *= fac

    return f
#---------------------------------------------------------------------------
def sin2_pulse(F0,w0,t0,Tdur,ts):
    It, = np.where((ts > t0) & (ts < t0 + Tdur))
    f = np.zeros(len(ts))
    f[It] = F0 * np.sin(np.pi*(ts[It] - t0)/Tdur)**2 * np.sin(w0*(ts[It]-t0))
    return f
#---------------------------------------------------------------------------
def Ncycle_pulse(F0,w0,t0,nc,ts):
    Tdur = 2*np.pi * nc / w0
    return sin2_pulse(F0,w0,t0,Tdur,ts)
#---------------------------------------------------------------------------                                                                        
def Box_pulse(F0,w0,t0,nc,ts):
    Tdur = 2*np.pi * nc / w0
    It, = np.where((ts > t0) & (ts < t0 + Tdur))
    f = np.zeros(len(ts))
    f[It] = F0 * np.sin(w0*(ts[It]-t0))
    return f
#---------------------------------------------------------------------------
def switchon(F0,w0,tswitch,ts):
    f = np.ones(len(ts))
    It, = np.where(ts < tswitch)
    f[It] = np.sin(0.5*np.pi*ts[It]/tswitch)**2

    return F0 * f * np.sin(w0*ts)
#---------------------------------------------------------------------------
def Calc_Afield(tgrd,F0,w0,t0,Tdur,phi):

    def fun(t):
        return -gauss_pulse(F0,w0,t0,Tdur,t,phi)

    Afield = quad_rk4(tgrd,fun)

    return Afield
#--------------------------------------------------------------
def WritePump(file_out,F0,w0,t0,Tdur,pol,phi=0.0,asym=0.0,asym_t0=0.0,asym_dur=100,nt=800,rewrite=True):
    ts = np.linspace(t0 - 5*Tdur, t0 + 5*Tdur, nt)
    fpump = gauss_pulse(F0,w0,t0,Tdur,ts,phi=phi,asym=asym,asym_t0=asym_t0,asym_dur=asym_dur)
    Efield_x = pol[0] * fpump
    Efield_y = pol[1] * fpump
    Efield_z = pol[2] * fpump
    
    if (not os.path.isfile(file_out)) or rewrite:
        np.savetxt(file_out,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')
#--------------------------------------------------------------
def WritePump_2Gauss(file_out,F0,shg,w0,t0,Tdur,pol,phi=0.0,dlt=0.0,nt=800,rewrite=True):
    ts = np.linspace(t0 - 5*Tdur, t0 + 5*Tdur, nt)
    fpump1 = gauss_pulse(F0,w0,t0,Tdur,ts,phi=phi)
    fpump2 = gauss_pulse(F0 * shg, 2*w0,t0,Tdur,ts,phi=phi+dlt)
    Efield_x = pol[0] * (fpump1 + fpump2)
    Efield_y = pol[1] * (fpump1 + fpump2)
    Efield_z = pol[2] * (fpump1 + fpump2)
    
    if (not os.path.isfile(file_out)) or rewrite:
            np.savetxt(file_out,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')            
#--------------------------------------------------------------                                                                                     
def WriteBoxPump(file_out,F0,w0,t0,nc,pol,nt=800,rewrite=True):
    Tdur = 2*np.pi * nc / w0
    ts = np.linspace(t0, t0 + Tdur, nt)
    fpump = Box_pulse(F0,w0,t0,nc,ts)
    Efield_x = pol[0] * fpump
    Efield_y = pol[1] * fpump
    Efield_z = pol[2] * fpump

    if (not os.path.isfile(file_out)) or rewrite:
        np.savetxt(file_out,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')
#--------------------------------------------------------------                                                                                     
def WriteTrapzPump(file_out,F0,w0,Tdur,tramp,pol,nt=800,rewrite=True):
    ts = np.linspace(0.0, 1.1*Tdur, nt)
    fpump = trapz_pulse(F0,w0,Tdur,tramp,ts)
    Efield_x = pol[0] * fpump
    Efield_y = pol[1] * fpump
    Efield_z = pol[2] * fpump

    if (not os.path.isfile(file_out)) or rewrite:
        np.savetxt(file_out,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')
#--------------------------------------------------------------
def GetField(ts, pump):
    F0, w0, Tdur, phi, s0 = pump['F0'], pump['w0'], pump['Tdur'], pump['phi'], pump['s0']
    pol = np.array(pump['pol'])
    t0 = s0 * Tdur
    Efs = gauss_pulse(F0,w0,t0,Tdur,ts,phi=phi)
    E_spl = UnivariateSpline(ts, Efs, s=0)
    A_spl = E_spl.antiderivative()
    Avs = - A_spl(ts)

    return Avs, Efs
#--------------------------------------------------------------
def GenField(pump,path="data/field/"):
    Ry = 27.21138602
    tfs = 0.02418884254
    f_fac = 1.945e-7
    F0, w0, Tdur, phi, s0 = pump['F0'], pump['w0'], pump['Tdur'], pump['phi'], pump['s0']
    asym = pump['asym']
    t0 = s0 * Tdur / tfs
    pol = [1.0, 0.0, 0.0]

    file_field = path + 'pulse_F{}_w{}_Tdur{}_phi{}_s{}'.format(F0,w0,Tdur,phi,s0)
    if asym['fac'] > 0.0:
        file_field += '_asym{}_dur{}.dat'.format(asym['fac'],asym['dur'])
        WritePump(file_field,F0 * f_fac,w0/Ry,t0,Tdur/tfs,pol,phi=phi*np.pi,\
                  asym=asym['fac'],asym_t0=t0,asym_dur=asym['dur']/tfs,
                  nt=1000,rewrite=False)
    else:
        file_field += '.dat'
        WritePump(file_field,F0 * f_fac,w0/Ry,t0,Tdur/tfs,pol,phi=phi*np.pi,\
                  nt=1000,rewrite=False)        
        
    return file_field
#--------------------------------------------------------------
def GenField_PumpProbe(pump,probe,path="data/field/",write_file=True):
    Ry = 27.21138602
    tfs = 0.02418884254
    f_fac = 1.945e-7

    Tpump = np.round(pump['fwhm'] / 2.355, decimals=5) / tfs
    Tprobe = np.round(probe['fwhm'] / 2.355, decimals=5) / tfs
    t0 = pump['s0'] * Tpump 
    
    dt = 0.01 * 2*np.pi/(pump['w0']/Ry + probe['w0']/Ry)
    nt = 4 * int(Tpump / dt)
    
    ts = np.linspace(t0 - 5*Tpump, t0 + 5*Tpump, nt)
    fpump = gauss_pulse(f_fac * pump['F0'],pump['w0']/Ry,t0,Tpump,ts,phi=pump['phi']*np.pi)
    fprobe = gauss_pulse(f_fac * probe['F0'],probe['w0']/Ry,t0 + probe['delay']/tfs,Tprobe,ts)

    file_field = path + 'pulse_pump_probe_F{}_w{}_Tdur{}_phi{}_s{}_Fpr{}_wpr{}_tpr{}_Tpr{}.dat'.\
        format(pump['F0'],pump['w0'],pump['fwhm'],pump['phi'],pump['s0'],probe['F0'],probe['w0'],probe['delay'],probe['fwhm'])

    if write_file:
        Efield_x = pump['pol'][0] * fpump + probe['pol'][0] * fprobe
        Efield_y = pump['pol'][1] * fpump + probe['pol'][1] * fprobe
        Efield_z = pump['pol'][2] * fpump + probe['pol'][2] * fprobe
        np.savetxt(file_field,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')
        
    return file_field
#--------------------------------------------------------------
def GenWaveform(p,path="data/field/"):
    Ry = 27.21138602
    tfs = 0.02418884254
    f_fac = 1.945e-7
    F0, w0, Tdur, cep, toff, s0, fchirp = p['F0'], p['w0'], p['Tdur'], p['cep'], p['toff'], p['s0'], p['fchirp']

    pol = [1.0, 0.0, 0.0]

    file_field = path + 'waveform_F{}_w{}_Tdur{}_cep{}_chirp{}_s{}.dat'.format(F0,w0,Tdur,cep,fchirp,s0)

    F0 *= f_fac
    # w0 *= 1.0 / Ry
    # Tdur *= 1.0 / tfs
    # cep *= np.pi
    # toff *= 1.0 / tfs
    # fchirp *= 1.0 / Ry
    t0 = s0 * Tdur

    if not os.path.isfile(file_field):
        ts = np.linspace(t0 - 4*Tdur, t0 + 4*Tdur, 800) / tfs
        fpump =  - waveform(ts * tfs, w0, Tdur, F0, cep, toff + t0, fchirp)    
        Efield_x = pol[0] * fpump
        Efield_y = pol[1] * fpump
        Efield_z = pol[2] * fpump
        np.savetxt(file_field,np.c_[ts,Efield_x,Efield_y,Efield_z],fmt='%20.10f')

    return file_field
#--------------------------------------------------------------
