import numpy as np
import scipy.linalg as la
import matplotlib.pyplot as plt
#--------------------------------------------------------------------
class BiSe_Surf():
   """docstring for BiSe_Surf"""
   #==============================================================
   def __init__(self, alat, Apar, Bpar, m11):
      # Apar -> [A0, A11, A12, A14]
      # Bpar -> [B0, B11, B12, B14]
      sq3 = np.sqrt(3.0)

      self.a0 = alat
      self.a1 = self.a0 * np.array([1.0, 0.0])
      self.a2 = self.a0 * np.array([-0.5, 0.5*sq3])
      self.a3 = self.a0 * np.array([-0.5, -0.5*sq3])
      self.b1 = self.a0 * np.array([0.0, sq3/3.0])
      self.b2 = self.a0 * np.array([-0.5, -sq3/6.0])
      self.b3 = self.a0 * np.array([0.5, -sq3/6.0])
      self.Apar = Apar
      self.Bpar = Bpar
      self.m11 = m11
   #==============================================================
   def get_h0(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar
      h0 = 2*A0 * (np.cos(np.dot(k,self.a1)) + np.cos(np.dot(k,self.a2)) + np.cos(np.dot(k,self.a3)) )
      h0 += 2*B0 * (np.cos(np.dot(k,self.b1)) + np.cos(np.dot(k,self.b2)) + np.cos(np.dot(k,self.b3)) )
      return h0
   #==============================================================
   def get_h1(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar
      w = -2*np.pi/3.0
      sw = np.sin(w)
      cw = np.cos(w)

      h1 = -2*A14*sw * (np.sin(np.dot(k,self.a2)) - np.sin(np.dot(k,self.a3)))
      h1 += 2*B14 * (np.sin(np.dot(k,self.b1)) + cw * (np.sin(np.dot(k,self.b2)) + np.sin(np.dot(k,self.b3))))
      return h1
   #==============================================================
   def get_h2(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar
      w = -2*np.pi/3.0
      sw = np.sin(w)
      cw = np.cos(w)

      h2 = -2*B14*sw * (np.sin(np.dot(k,self.b2)) - np.sin(np.dot(k,self.b3)))
      h2 += -2*A14 * (np.sin(np.dot(k,self.a1)) + cw * (np.sin(np.dot(k,self.a2)) + np.sin(np.dot(k,self.a3))))
      return h2
   #==============================================================
   def get_h3(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar

      h3 = 2 * A12 * (np.sin(np.dot(k,self.a1)) + np.sin(np.dot(k,self.a2)) + np.sin(np.dot(k,self.a3)))
      return h3
   #==============================================================
   def get_h4(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar

      h4 = -2 * B12 * (np.sin(np.dot(k,self.b1)) + np.sin(np.dot(k,self.b2)) + np.sin(np.dot(k,self.b3)))
      return h4
   #==============================================================
   def get_h5(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar

      h5 = 2*A11*(np.cos(np.dot(k,self.a1)) + np.cos(np.dot(k,self.a2)) + np.cos(np.dot(k,self.a3)))
      h5 += 2*B11*(np.cos(np.dot(k,self.b1)) + np.cos(np.dot(k,self.b2)) + np.cos(np.dot(k,self.b3)))
      h5 += self.m11

      return h5
   #==============================================================
   def get_ham(self,k):
      A0, A11, A12, A14 = self.Apar
      B0, B11, B12, B14 = self.Bpar
      h0 = self.get_h0(k)
      h1 = self.get_h1(k)
      h2 = self.get_h2(k)
      h3 = self.get_h3(k)
      h5 = self.get_h5(k)
      h0_k0 = self.get_h0(np.array([0.0,0.0]))
      h5_k0 = self.get_h5(np.array([0.0,0.0]))

      cb = np.sqrt(1.0 - (B0/B11)**2)

      Hk = np.zeros([2,2], dtype=np.complex_)
      Hk[0,0] = h0 - h0_k0 + cb * h3 + B0/B11 * (h5_k0 - h5)
      Hk[1,1] = h0 - h0_k0 - cb * h3 + B0/B11 * (h5_k0 - h5)
      Hk[0,1] = cb * (h1 - 1j * h2)
      Hk[1,0] = cb * (h1 + 1j * h2)

      return Hk
   #==============================================================
   def get_spin(self,k):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)

      sz0 = np.abs(vect[0,0])**2 - np.abs(vect[1,0])**2
      sz1 = np.abs(vect[0,1])**2 - np.abs(vect[1,1])**2
      sx0 = 2 * np.real(np.conj(vect[0,0]) * vect[1,0])
      sx1 = 2 * np.real(np.conj(vect[0,1]) * vect[1,1])
      sy0 = 2 * np.imag(np.conj(vect[0,0]) * vect[1,0])
      sy1 = 2 * np.imag(np.conj(vect[0,1]) * vect[1,1])

      return sx0,sx1,sy0,sy1,sz0,sz1
   #==============================================================
   def berryconnection(self,k,dk=1.0e-3):
      k1 = k + np.array([dk,0.0])

      Hk = self.get_ham(k)
      Hk1 = self.get_ham(k1)
      eps0, vect0 = la.eigh(Hk)
      eps1, vect1 = la.eigh(Hk1)

      vgroup = (eps1 - eps0)/dk
      a01 = 1j*(vect0[:,0] - vect1[:,1])/dk
      a10 = 1j*(vect0[:,1] - vect1[:,0])/dk

      return eps0,vgroup,a01,a10
   #==============================================================
   def get_velocity(self,k,dk=1.0e-3):
      velo = np.zeros([2,2], dtype=np.complex_)

      eps,vgroup,a01,a10 = self.berryconnection(k,dk=dk)

      velo[0,0] = vgroup[0]
      velo[1,1] = vgroup[1]
      velo[0,1] = 1j * (eps[0] - eps[1]) * a01
      velo[1,0] = 1j * (eps[1] - eps[0]) * a10

      velo = 0.5 * (velo + np.conj(velo.T))

      return velo







