import numpy as np
import scipy.linalg as la
from scipy.interpolate import UnivariateSpline
from scipy.optimize import brentq
import matplotlib.pyplot as plt
import surface_kp as surf_kp
#----------------------------------------------------------------------
def nfermi(beta,w):
   return 1.0 / (1.0 + np.exp(beta * w))
#--------------------------------------------------------------
def GetBands_5QL(ks,vsb,nbulk,Avec=0.0):
    bite = BiSe_Surf_5QL(vsb=vsb,vsb2=vsb)
    epsk = np.zeros([len(ks),nbulk+2])
    for ik, kx in enumerate(ks):
        kpt = [kx, 0.0]
        Hk = bite.get_ham(kpt)
        vk = bite.get_velocity(kpt)
        HkA = Hk - Avec * vk[:,:,0]
        epsk[ik,:] = la.eigh(HkA, eigvals_only=True)

    return epsk
#--------------------------------------------------------------

#--------------------------------------------------------------------
class BiSe_Surf_3bnd():
   """docstring for BiSe_Surf_3bnd"""
   #==============================================================
   def __init__(self,vsb=0.0,vsb2=0.0):
      Ry = 27.211386
      #afac = 1.0/ 0.52917721
      afac = 0.52917721

      self.c2 = 0.32 / Ry # eV
      self.ksym = 0.12 * afac # A-1
      self.r2 = 1.57 / Ry / afac # eV A
      self.kasym = 0.215 * afac # A-1
      #self.C0 = 0.176 / Ry # eV
      self.C0 = 0.0
      self.A = 0.609 / Ry / afac # eV A

      self.Eb = 0.285 / Ry
      self.ab = 1.0

      self.vsb = vsb
   #==============================================================
   def get_vF(self,mu,kmax=0.1,nk=100):
      ks = np.linspace(0.0,kmax,nk)
      epsk = np.zeros(nk)
      for ik, kx in enumerate(ks):
         eps = self.get_eps([kx, 0.0])
         epsk[ik] = eps[1]

      spl = UnivariateSpline(ks, epsk)
      spl_der = spl.derivative()

      def root_func(k):
         return spl(k) - mu

      kF = brentq(root_func,0.0,kmax)

      return spl_der(kF)
   #==============================================================
   def get_Ck(self,k):
      kr = np.sqrt(k[0]**2 + k[1]**2)
      Ck = self.c2/self.ksym**2 * 1./(1. + (kr/self.ksym)**2)
      return Ck
   #==============================================================
   def get_Rk(self,k):
      kr = np.sqrt(k[0]**2 + k[1]**2)
      Rk = self.r2/self.kasym**2 * 1./(1. + (kr/self.kasym)**4)
      return Rk
   #==============================================================
   def get_ham_surf(self,k):
      kx, ky = k
      Ck = self.get_Ck(k)
      Rk = self.get_Rk(k)
      
      H2x2 = np.zeros([2,2], dtype=np.complex_)
      H2x2[0,0] = self.C0 + Ck * (kx**2 + ky**2) + 2*Rk*(kx**3 - 3*ky**2*kx)
      H2x2[1,1] = self.C0 + Ck * (kx**2 + ky**2) - 2*Rk*(kx**3 - 3*ky**2*kx)
      H2x2[0,1] = self.A*(ky + 1j*kx)
      H2x2[1,0] = np.conj(H2x2[0,1])  

      return H2x2
   #==============================================================
   def get_ham(self,k,band_basis=False):
      kx, ky = k
      
      H2x2 = self.get_ham_surf(k)
      eps2 = la.eigh(H2x2, eigvals_only=True)

      epsk = np.zeros(3)
      epsk[0:2] = eps2
      epsk[2] = self.Eb + self.ab * (kx**2 + ky**2)

      Hk = np.diag(epsk)

      return Hk
   #==============================================================
   def get_rot(self,k):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)
      return vect
   #==============================================================
   def get_occ(self,k,beta,mu):
      Hk = self.get_ham(k)
      eps = la.eigh(Hk, eigvals_only=True)
      return nfermi(beta,eps - mu)
   #==============================================================
   def get_rho(self,k,beta,mu,band_basis=False):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)
      rho_diag = np.diag(nfermi(beta,eps - mu))
      if band_basis:
         return rho_diag

      rho = vect @ rho_diag @ np.conj(vect.T)
      return rho
   #==============================================================
   def get_eps(self,k):
      Hk = self.get_ham(k)
      eps = la.eigh(Hk, eigvals_only=True)
      return eps
   #==============================================================
   def get_velocity(self,k):
      kx, ky = k

      kp = surf_kp.BiSe_Surf_kp()
      v2x2 = kp.get_velocity_gk(k)
      H2x2 = kp.get_ham(k)
      
      eps, vect = la.eigh(H2x2)
      v2x2 = np.einsum('ia,ijr,jb->abr', np.conj(vect),v2x2,vect)

      velo = np.zeros([3,3,2], dtype=np.complex_)
      velo[0:2,0:2,:] = v2x2
      velo[2,2,0] = 2* self.ab * kx
      velo[2,2,1] = 2* self.ab * ky

      velo[1,2,:] = -1j * self.vsb
      velo[2,1,:] = 1j * self.vsb

      return velo







