import numpy as np
import scipy.linalg as la
from scipy.interpolate import UnivariateSpline
from scipy.optimize import brentq
import matplotlib.pyplot as plt
#----------------------------------------------------------------------
def nfermi(beta,w):
   return 1.0 / (1.0 + np.exp(beta * w))
#--------------------------------------------------------------------
def GetBands_kp(ks,Avec=0.0):
    bite = BiSe_Surf_kp()
    epsk = np.zeros([len(ks),2])
    for ik, kx in enumerate(ks):
        kpt = [kx, 0.0]
        Hk = bite.get_ham(kpt)
        vk = bite.get_velocity(kpt)
        HkA = Hk - Avec * vk[:,:,0]
        epsk[ik,:] = la.eigh(HkA, eigvals_only=True)

    return epsk
#--------------------------------------------------------------------
class BiSe_Surf_kp():
   """docstring for BiSe_Surf_kp"""
   #==============================================================
   def __init__(self, scale=1.0):
      Ry = 27.211386
      #afac = 1.0/ 0.52917721
      afac = 0.52917721

      self.ks = scale
      self.c2 = 0.32 / Ry # eV
      self.ksym = 0.12 * afac # A-1
      self.r2 = 1.57 / Ry / afac # eV A
      self.kasym = 0.215 * afac # A-1
      #self.C0 = 0.176 / Ry # eV
      self.C0 = 0.0
      self.A = 0.609 / Ry / afac # eV A
   #==============================================================
   def get_vF(self,mu,kmax=0.1,nk=100):
      ks = np.linspace(0.0,kmax,nk)
      epsk = np.zeros(nk)
      for ik, kx in enumerate(ks):
         eps = self.get_eps([kx, 0.0])
         epsk[ik] = eps[1]

      spl = UnivariateSpline(ks, epsk)
      spl_der = spl.derivative()

      def root_func(k):
         return spl(k) - mu

      kF = brentq(root_func,0.0,kmax)

      return spl_der(kF)
   #==============================================================
   def get_Ck(self,k):
      kr = np.sqrt(k[0]**2 + k[1]**2)
      Ck = self.c2/self.ksym**2 * 1./(1. + (kr/self.ksym)**2)
      return Ck
   #==============================================================
   def get_Rk(self,k):
      kr = np.sqrt(k[0]**2 + k[1]**2)
      Rk = self.r2/self.kasym**2 * 1./(1. + (kr/self.kasym)**4)
      return Rk
   #==============================================================
   def get_ham(self,k,band_basis=False):
      kx, ky = k
      kx *= 1.0 / self.ks
      kx *= 1.0 / self.ks
      Ck = self.get_Ck(k)
      Rk = self.get_Rk(k)
      
      Hk = np.zeros([2,2], dtype=np.complex_)
      Hk[0,0] = self.C0 + Ck * (kx**2 + ky**2) + 2*Rk*(kx**3 - 3*ky**2*kx)
      Hk[1,1] = self.C0 + Ck * (kx**2 + ky**2) - 2*Rk*(kx**3 - 3*ky**2*kx)
      Hk[0,1] = self.A*(ky + 1j*kx)
      Hk[1,0] = np.conj(Hk[0,1])

      if band_basis:
         eps, vect = la.eigh(Hk)
         Hk = np.conj(vect.T) @ Hk @ vect

      return Hk
   #==============================================================
   def get_rot(self,k):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)
      return vect
   #==============================================================
   def get_occ(self,k,beta,mu):
      Hk = self.get_ham(k)
      eps = la.eigh(Hk, eigvals_only=True)
      return nfermi(beta,eps - mu)
   #==============================================================
   def get_rho(self,k,beta,mu,band_basis=False):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)
      rho_diag = np.diag(nfermi(beta,eps - mu))
      if band_basis:
         return rho_diag

      rho = vect @ rho_diag @ np.conj(vect.T)
      return rho
   #==============================================================
   def get_eps(self,k):
      Hk = self.get_ham(k)
      eps = la.eigh(Hk, eigvals_only=True)
      return eps
   #==============================================================
   def get_ham_td(self,k,Ef,Av,dk=1.0e-4):
      ka = k - Av
      # Hk = self.get_ham(ka,band_basis=True)
      eps = self.get_eps(ka)
      Ab_x, Ab_y = self.dipole_matrix(ka,dk=dk)

      Hk = np.zeros([2,2], dtype=np.complex_)
      Hk += np.diag(eps)

      Hk += -Ef[0] * Ab_x - Ef[1] * Ab_y
      Hk = 0.5 * (Hk + np.conj(Hk.T))

      return Hk
   #==============================================================
   def get_ham_pei(self,k,Av):
      ka = k - Av
      Hk = self.get_ham(ka)
      return Hk
   #==============================================================
   def get_current_k(self,rho,k,Av,dk=1.0e-4):
      ka = k - Av
      vk = self.get_velocity(ka,dk=1.0e-4)

      Jk = np.real(np.einsum('ij, jid -> d', rho, vk))
      return Jk
   #==============================================================
   def get_current_gk(self,rho,k,Av):
      ka = k - Av
      vk = self.get_velocity_gk(ka)

      Jk = np.real(np.einsum('ij, jid -> d', rho, vk))
      return Jk
   #==============================================================
   def get_spin(self,k):
      Hk = self.get_ham(k)
      eps, vect = la.eigh(Hk)

      sz0 = np.abs(vect[0,0])**2 - np.abs(vect[1,0])**2
      sz1 = np.abs(vect[0,1])**2 - np.abs(vect[1,1])**2
      sx0 = 2 * np.real(np.conj(vect[0,0]) * vect[1,0])
      sx1 = 2 * np.real(np.conj(vect[0,1]) * vect[1,1])
      sy0 = 2 * np.imag(np.conj(vect[0,0]) * vect[1,0])
      sy1 = 2 * np.imag(np.conj(vect[0,1]) * vect[1,1])

      return sx0,sx1,sy0,sy1,sz0,sz1
   #==============================================================
   def dipole_matrix(self,k,dk=1.0e-4):
      A_x = np.zeros([2,2], dtype=np.complex_)
      A_y = np.zeros([2,2], dtype=np.complex_)

      k1 = k + np.array([dk,0.0])

      Hk = self.get_ham(k)
      Hk1 = self.get_ham(k1)
      eps0, vect0 = la.eigh(Hk)
      eps1, vect1 = la.eigh(Hk1)

      for i in range(2):
         for j in range(2):
            A_x[i,j] = 1j/dk * np.vdot(vect0[:,i], vect1[:,j] - vect0[:,j])

      k1 = k + np.array([0.0,dk])

      Hk = self.get_ham(k)
      Hk1 = self.get_ham(k1)
      eps0, vect0 = la.eigh(Hk)
      eps1, vect1 = la.eigh(Hk1)

      for i in range(2):
         for j in range(2):
            A_y[i,j] = 1j/dk * np.vdot(vect0[:,i], vect1[:,j] - vect0[:,j])

      return A_x, A_y
   #==============================================================
   def berryconnection(self,k,dk=1.0e-3):
      k1 = k + np.array([dk,0.0])

      Hk = self.get_ham(k)
      Hk1 = self.get_ham(k1)
      eps0, vect0 = la.eigh(Hk)
      eps1, vect1 = la.eigh(Hk1)

      vgroup_x = (eps1 - eps0)/dk
      a01_x = 1j/dk * np.vdot(vect0[:,0], vect1[:,1] - vect0[:,1])
      a10_x = 1j/dk * np.vdot(vect0[:,1], vect1[:,0] - vect0[:,0])


      k1 = k + np.array([0.0,dk])

      Hk = self.get_ham(k)
      Hk1 = self.get_ham(k1)
      eps0, vect0 = la.eigh(Hk)
      eps1, vect1 = la.eigh(Hk1)

      vgroup_y = (eps1 - eps0)/dk
      a01_y = 1j/dk * np.vdot(vect0[:,0], vect1[:,1] - vect0[:,1])
      a10_y = 1j/dk * np.vdot(vect0[:,1], vect1[:,0] - vect0[:,0])

      return eps0,vgroup_x,a01_x,a10_x,vgroup_y,a01_y,a10_y
   #==============================================================
   def get_velocity_gk(self,k,dk=1e-3,inter_band=True):
      kx, ky = k
      Ck = self.get_Ck(k)
      Rk = self.get_Rk(k)
      velo = np.zeros([2,2,2], dtype=np.complex_)

      kp = k + np.array([0.5*dk,0.0])
      km = k + np.array([-0.5*dk,0.0])

      Hp = self.get_ham(kp,band_basis=False)
      Hm = self.get_ham(km,band_basis=False)
      vk = (Hp - Hm) / dk
      velo[:,:,0] = 0.5 * (vk + np.conj(vk.T))

      kp = k + np.array([0.0, 0.5*dk])
      km = k + np.array([0.0, -0.5*dk])
      Hp = self.get_ham(kp,band_basis=False)
      Hm = self.get_ham(km,band_basis=False)      
      vk = (Hp - Hm) / dk
      velo[:,:,1] = 0.5 * (vk + np.conj(vk.T))

      if not inter_band:
         Hk = self.get_ham(k,band_basis=False)
         eps, vect = la.eigh(Hk)
         velo_diag = np.einsum('ia,ijr,jb->abr', np.conj(vect), velo, vect)
         velo_diag[0,1,:] = 0.0
         velo_diag[1,0,:] = 0.0
         velo = np.einsum('ia,abr,jb->ijr', vect, velo_diag, np.conj(vect) )
      
      return velo
   #==============================================================
   def get_velocity(self,k,dk=1.0e-3):
      velo = np.zeros([2,2,2], dtype=np.complex_)

      eps,vgroup_x,a01_x,a10_x,vgroup_y,a01_y,a10_y = self.berryconnection(k,dk=dk)

      velo[0,0,0] = vgroup_x[0]
      velo[1,1,0] = vgroup_x[1]
      velo[0,1,0] = 1j * (eps[0] - eps[1]) * a01_x
      velo[1,0,0] = 1j * (eps[1] - eps[0]) * a10_x

      velo[0,0,1] = vgroup_y[0]
      velo[1,1,1] = vgroup_y[1]
      velo[0,1,1] = 1j * (eps[0] - eps[1]) * a01_y
      velo[1,0,1] = 1j * (eps[1] - eps[0]) * a10_y

      for i in range(1,2):
         velo[:,:,i] = 0.5 * (velo[:,:,i] + np.conj(velo[:,:,i].T))

      return velo








