import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

# load homemade python moduels
import absplot_v03 as absplot
import zapes_utilities_v03 as zutil

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_curvature_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# global figure parameters
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.33
emax_plot = 0.32
colormap1 = 'inferno'
ticks_color1 = 'white'
colormap2 = 'gnu_v2_r'
ticks_color2 = 'black'
figs_clim_200nm = [1000, 4000]
figs_clim_400nm_raw = [300, 2000]
figs_clim_400nm_2deriv = [0.002, 0.03]

lw_dft_thin = 1.0
color_dft = 'gray'
color_dft_projection1 = 'white'
color_dft_projection2 = 'gray'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=66)

#######################################################################################################################
# Plot structure 1 (raw ARPES images)
#######################################################################################################################

# create empty frame for capsulation
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=0, width=100, height=65)
# eframe1.visualize_frame()

# create multiple plotting frame
img_left = 11.5
img_bottom = 9.5
img_width = 37
img_hoffset = 46.5
img_height = 48 * (37.0/35.0)  # the same aspect ratio in Fig. 2
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color1,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color1,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1_ = './image_processed_2ppe5424_raw_v07_6eV.png'
image_processed_200nm = zutil.ImageProcessed(img_filepath1_, use_pixel_coord=False, ef_shift=0.0)
pframe11.assign_image(image_processed_200nm.image, extent=image_processed_200nm.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap1, clim=figs_clim_200nm, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2_ = './image_processed_2ppe7044_0_sc_raw_shift2d_n40_v10.png'
image_processed_400nm_raw_pixel = zutil.ImageProcessed(img_filepath2_, use_pixel_coord=False, ef_shift=0.0)
pframe12.assign_image(image_processed_400nm_raw_pixel.image, extent=image_processed_400nm_raw_pixel.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap1, clim=figs_clim_400nm_raw, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load dft calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT data
for pframe in [pframe11, pframe12]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                           facecolor=color_dft_projection1, alpha=alpha_dft_projection)
    pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                           facecolor=color_dft_projection1, alpha=alpha_dft_projection)
    pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)

# set other attributes
pframe11.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe11, pframe12]:
    pframe.ax.hlines(0, image_processed_400nm_raw_pixel.kmin, image_processed_400nm_raw_pixel.kmax,
                     color=ticks_color1, lw=0.5, ls=(0, (8, 5)))
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.text(0.5, 1.03, r'200 nm, raw', fontsize=fontsize_main,
                 horizontalalignment='center', transform=pframe11.ax.transAxes)
pframe12.ax.text(0.5, 1.03, r'400 nm, raw', fontsize=fontsize_main,
                 horizontalalignment='center', transform=pframe12.ax.transAxes)

# add colorbar
pframe12.add_colorbar(cleft=39.5, cbottom=12, cwidth=2, cheight=25)
pframe12.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color1, show_ticks=False)
pframe12.cax.text(0.5, 1.05, 'max', transform=pframe12.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe12.cax.text(0.5, -0.13, 'min', transform=pframe12.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (curvature ARPES images)
#######################################################################################################################

# create empty frame for capsulation
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=103, bottom=0, width=100, height=65)
# eframe2.visualize_frame()

# create multiple plotting frame
img_left = 11.5
img_bottom = 9.5
img_width = 37
img_hoffset = 46.5
img_height = 48 * (37.0/35.0)  # the same aspect ratio in Fig. 2
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe22 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe22.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign images
img_filepath3_ = './image_processed_2ppe7044_0_sc_2deriv_narrow2_shift2d_n40_v10.png'
image_processed_400nm_2deriv = zutil.ImageProcessed(img_filepath3_, ef_shift=0.0)
pframe21.assign_image(image_processed_400nm_2deriv.image, extent=image_processed_400nm_2deriv.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_400nm_2deriv, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4_ = './image_processed_2ppe7044_0_sc_2deriv_wide2_shift2d_n40_v07.png'
image_processed_400nm_2deriv_wide = zutil.ImageProcessed(img_filepath4_, ef_shift=0.0)
pframe22.assign_image(image_processed_400nm_2deriv_wide.image, extent=image_processed_400nm_2deriv_wide.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_400nm_2deriv, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load dft calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT data
for pframe in [pframe21, pframe22]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                           facecolor=color_dft_projection2, alpha=alpha_dft_projection)
    pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                           facecolor=color_dft_projection2, alpha=alpha_dft_projection)
    pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)

# set other attributes
pframe21.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe21, pframe22]:
    pframe.ax.hlines(0, image_processed_400nm_2deriv.kmin, image_processed_400nm_2deriv.kmax,
                     color=ticks_color2, lw=0.5, ls=(0, (8, 5)))
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe21.ax.text(0.5, 1.03, r'400 nm, curvature 1', fontsize=fontsize_main,
                 horizontalalignment='center', transform=pframe21.ax.transAxes)
pframe22.ax.text(0.5, 1.03, r'400 nm, curvature 2', fontsize=fontsize_main,
                 horizontalalignment='center', transform=pframe22.ax.transAxes)

# add colorbar
pframe22.add_colorbar(cleft=39.5, cbottom=12, cwidth=2, cheight=25)
pframe22.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color2, show_ticks=False)
pframe22.cax.text(0.5, 1.05, 'max', transform=pframe22.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe22.cax.text(0.5, -0.13, 'min', transform=pframe22.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(0.5, 62, 'a', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(50.5, 62, 'b', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(103.5, 62, 'c', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(153.5, 62, 'd', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
