import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

# load homemade python moduels
import absplot_v03 as absplot
import zapes_utilities_v03 as zutil

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_streaking_current_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
t_offset = -2.0
tmin_plot = -130
tmax_plot = 130
colormap = 'inferno'
figs_clim_kstreaking = [0.25e+6, 1.35e+6]
ticks_color = 'white'
colormap2 = 'gnu_v2_r'
figs_clim_kstreaking_2deriv = [0.003, 0.02]
ticks_color2 = 'black'
figs_clim_second_derivative = [0.00, 0.03]

lw_dft_thick = 2.0
lw_dft_thin = 1.0
color_dft = 'gray'
color_dft_projection = 'gray'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=120)

#######################################################################################################################
# Plot structure 1 (raw streaking images)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=1, bottom=68, width=55, height=50)
# eframe1.visualize_frame()

# create plot frames
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=11, bottom=10, width=38, height=38)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_kstreaking_2ppe7044_shift2d_n40_v19.png'
image_processed1 = zutil.ImageProcessedTemporal(img_filepath1, temporal_axis='x', other_axis='momentum')
extent = np.array(image_processed1.get_extent()) + np.array([t_offset, t_offset, 0.0, 0.0])
pframe11.assign_image(image_processed1.image, extent=extent,
                      xmin=tmin_plot, xmax=tmax_plot, ymin=-0.2, ymax=0.2,
                      colormap=colormap, clim=figs_clim_kstreaking, transpose=False, inversex=False,
                      func_colormap=zutil.util_cmap)

# set attributes
pframe11.ax.set_ylabel(r'$\Delta k_{e}$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=2)
for pframe in [pframe11]:
    pframe.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)

# add colorbar
pframe11.add_colorbar(cleft=40.5, cbottom=6, cwidth=1.5, cheight=20)
pframe11.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color, show_ticks=False)
pframe11.cax.text(0.5, 1.05, 'max', transform=pframe11.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe11.cax.text(0.5, -0.13, 'min', transform=pframe11.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (2deriv streaking images)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=58, bottom=68, width=55, height=50)
# eframe2.visualize_frame()

# create plot frames
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=11, bottom=10, width=38, height=38)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_kstreaking_2ppe7044_shift2d_n40_v10_streaking.png'
image_processed1 = zutil.ImageProcessedTemporal(img_filepath1, temporal_axis='x', other_axis='momentum')
extent = np.array(image_processed1.get_extent()) + np.array([t_offset, t_offset, 0.0, 0.0])
pframe21.assign_image(image_processed1.image, extent=extent,
                      xmin=tmin_plot, xmax=tmax_plot, ymin=-0.2, ymax=0.2,
                      colormap=colormap2, clim=figs_clim_kstreaking_2deriv, transpose=False, inversex=False,
                      func_colormap=zutil.util_cmap)

# set attributes
pframe21.ax.set_ylabel(r'$\Delta k_{e}$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=2)
for pframe in [pframe21]:
    pframe.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)

# add colorbar
pframe21.add_colorbar(cleft=40.5, cbottom=6, cwidth=1.5, cheight=20)
pframe21.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color, show_ticks=False)
pframe21.cax.text(0.5, 1.05, 'max', transform=pframe21.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe21.cax.text(0.5, -0.13, 'min', transform=pframe21.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (extracted waveforms)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=116, bottom=68,
                             width=88, height=50)
# eframe3.visualize_frame()

# create plot frames
img_left = 12.5
img_bottom = 10
img_width = 65
img_height = 38
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad,
                      reflect_top=False, reflect_right=False)
pframe31.add_right_axis(ticks_fontsize=fontsize_main, ticks_label_ypad=ticks_ypad)

inset_width = 17
inset_height = 12
pframe31_inset1 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                                    left=img_left + img_width - inset_width,
                                    bottom=img_bottom + img_height - inset_height,
                                    width=inset_width, height=inset_height)
pframe31_inset1.modify_ticks(ticks_fontsize=fontsize_main-1, ticks_color='black',
                             ticks_label_xpad=3, ticks_label_ypad=2,
                             reflect_top=False, reflect_right=False)

# load Eparallel waveform files
tmp_filepath = './waveform_extraction_2ppe7044_shift2d_n40_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset
k_streaking = np.real(data_waveform[:, 1])
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel * 1e-3
freq1 = np.real(data_waveform[:, 21])
E_parallel_deconvolved_fft = data_waveform[:, 23] / factor_uv_deconvolve_parallel * 1e-3

# set plot range
pframe31.ax.set_xlim(tmin_plot, tmax_plot)
pframe31.ax.set_ylim(-np.max(np.abs(k_streaking)) * 1.4, np.max(np.abs(k_streaking)) * 1.8)
pframe31.ax_right.set_ylim(-np.max(np.abs(E_parallel_deconvolved)) * 1.4, np.max(np.abs(E_parallel_deconvolved)) * 1.8)
pframe31_inset1.ax.set_xlim(10, 60)
pframe31_inset1.ax.set_ylim(0, 1.2)

# plot data
pframe31.ax.plot(delay_time, k_streaking, color='blue', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe31.ax_right.plot(delay_time, E_parallel_deconvolved, color='black', lw=0, marker='o', ms=4,
                       markerfacecolor='none')
pframe31.ax_right.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=2, markerfacecolor='none')
pframe31_inset1.ax.plot(freq1, np.abs(E_parallel_deconvolved_fft) / np.max(np.abs(E_parallel_deconvolved_fft)),
                        color='black', lw=1)
pframe31_inset1.ax.fill_between(freq1, np.abs(E_parallel_deconvolved_fft) / np.max(np.abs(E_parallel_deconvolved_fft)),
                                0, color='black', alpha=0.5)

# set other attributes
pframe31.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=2)
pframe31.ax.set_ylabel(r'${k}_{\mathrm{streak}}$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=2)
pframe31.ax_right.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=0)
pframe31.ax.tick_params(axis='y', colors='blue')
pframe31.ax.yaxis.label.set_color('blue')
pframe31_inset1.ax.set_xlabel(r'$\nu$ (THz)', fontsize=fontsize_main-1, labelpad=0)
pframe31_inset1.ax.set_ylabel(r'$E_y(\nu)$', fontsize=fontsize_main-1, labelpad=-1)
pframe31.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe31.fix_yaxis_interval(0.05)

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Plot structure 4 (experimental ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe4 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=0, width=137, height=65)
# eframe4.visualize_frame()

# create multiple plotting frame
img_left = 12
img_bottom = 10
img_width = 38
img_hoffset = 40
img_height = 38 * 48/35
pframe41 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe41.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe42 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe42.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe43 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe43.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
kmin_plot = -0.18
kmax_plot = 0.18
emin_plot = -0.35 * 0.18/0.15 + 0.02
emax_plot = 0.3 * 0.18/0.15 + 0.02
img_filepath1 = './image_processed_2ppe7044_30_sc_2deriv_wide2_shift2d_n40_v07.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe41.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe7044_80_sc_2deriv_wide2_shift2d_n40_v07.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe42.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe7044_90_sc_2deriv_wide2_shift2d_n40_v07.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe43.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load dft calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT data
for pframe in [pframe41, pframe42, pframe43]:
    if pframe == pframe41:
        pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
        pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
        pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)
    else:
        pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
        pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)
    pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)

# load current waveform files
tmp_filepath = './current_waveform_extraction_2ppe7044_shift2d_n40_v07.dat'
data_current_waveform = np.loadtxt(tmp_filepath,
                                   dtype=complex,
                                   converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                                   comments=';')
delay_time_current = np.real(data_current_waveform[:, 0]) + t_offset
ef_pixel_right = np.int16(np.real(data_current_waveform[:, 1]))
ef_pixel_left = np.real(np.real(data_current_waveform[:, 2]))
# current analysis parameters
k_origin_pixel = 475  # pixel unit
dEk_dk_pixel = 2.2  # pixel/pixel = no unit
box_right_pixel = [k_origin_pixel-10, 700, 980, 120, dEk_dk_pixel]
box_left_pixel = [k_origin_pixel+5, 700, 980, 120, -dEk_dk_pixel]
ek = 1.8  # eV unit
ef = 1.856  # eV unit
ep = 10.0  # eV unit
ek_pixel = 765  # pixel unit
efactor = 0.151 * ep / 1000.0  # eV/pixel unit

# set attributes
list_idx_data = [30, 80, 90]
for idx, pframe in enumerate([pframe41, pframe42, pframe43]):
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax, color=ticks_color2, lw=0.5, ls=(0, (8, 5)))
    if pframe == pframe41:
        pframe.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
    else:
        pframe.ax.tick_params(labelleft=False)
    zutil.draw_integration_box(box_right_pixel, 'red', pframe.ax, ek, efactor, ek_pixel, k_origin_pixel, ef)
    zutil.draw_integration_box(box_left_pixel, 'blue', pframe.ax, ek, efactor, ek_pixel, k_origin_pixel, ef)
    pframe.fix_yaxis_interval(0.1)

pframe41.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(delay_time_current[30]),
                 transform=pframe41.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe42.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(delay_time_current[80]),
                 transform=pframe42.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe43.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(delay_time_current[90]),
                 transform=pframe43.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

# add colorbar
pframe43.add_colorbar(cleft=40.5, cbottom=4, cwidth=2, cheight=25)
pframe43.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='white', show_ticks=False)
pframe43.cax.text(0.5, 1.05, 'max', transform=pframe43.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe43.cax.text(0.5, -0.13, 'min', transform=pframe43.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 4
#######################################################################################################################


#######################################################################################################################
# Plot structure 5 (waveforms)
#######################################################################################################################
# create empty frame for this structure
eframe5 = absplot.EmptyFrame(mainframe=mainframe, left=137.5, bottom=0, width=60, height=65)
# eframe5.visualize_frame()

# create single plotting frame
pframe51 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5, left=15, bottom=44, width=41, height=18)
pframe51.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe52 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5, left=15, bottom=10, width=41, height=34)
pframe52.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load Eparallel waveform files
tmp_filepath = './waveform_extraction_2ppe7044_shift2d_n40_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel * 1e-3

# load current waveform files
tmp_filepath = './current_waveform_extraction_2ppe7044_shift2d_n40_v07.dat'
data_current_waveform = np.loadtxt(tmp_filepath,
                                   dtype=complex,
                                   converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                                   comments=';')
factor_uv_deconvolve_current = 0.8
delay_time_current = np.real(data_current_waveform[:, 0]) + t_offset
current_deconvolved = np.real(data_current_waveform[:, 14]) / factor_uv_deconvolve_current
current_Boltzmann = np.real(data_current_waveform[:, 19]) / factor_uv_deconvolve_current

# set plot range
pframe51.ax.set_xlim(tmin_plot, tmax_plot)
scale_efield = np.max(np.abs(E_parallel_fit_rescaled_deconvolved)) * 1.25
pframe51.ax.set_ylim(-scale_efield, scale_efield)
pframe52.ax.set_xlim(tmin_plot, tmax_plot)
scale_current = np.max(np.abs(current_deconvolved)) * 1.25 * 1.1
pframe52.ax.set_ylim(-scale_current, scale_current)

# plot data
pframe51.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1.5)
color_current = (255 / 255, 64 / 255, 128 / 255)
pframe52.ax.plot(delay_time_current, current_deconvolved, color=color_current, lw=0, marker='o', ms=3,
                 markerfacecolor='none')
pframe52.ax.plot(delay_time_current, current_Boltzmann, color=color_current, lw=1)
for idx_data in [30, 80, 90]:
    pframe51.ax.vlines(delay_time[idx_data], -scale_efield, scale_efield, color='blue', lw=1.5, alpha=0.15)
    pframe52.ax.vlines(delay_time_current[idx_data], -scale_current, scale_current, color='blue', lw=1.5, alpha=0.15)

# set attributes
pframe51.ax.tick_params(labelbottom=False)
pframe51.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=2)
pframe51.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe52.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe52.ax.set_ylabel(r'$j_y$ (A cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe52.ax.hlines(0.0, delay_time_current[0], delay_time_current[-1], color='black', lw=0.25)
pframe52.fix_yaxis_interval(10)

#######################################################################################################################
# End of Plot structure 5
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(1, 115, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(58, 115, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(116, 115, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(1, 65, 'd', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(137, 65, 'e', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
