import os
import re
import cv2
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
from matplotlib import patches
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from PIL import Image
from PIL.PngImagePlugin import PngImageFile, PngInfo


class ImageProcessed:
    def __init__(self, filepath_, use_pixel_coord=False, use_kinetic_energy=False, ef_shift=0.0):
        if not os.path.exists(filepath_):
            raise OSError(f'{filepath_} not found')
        # PIL cannot read 16-bit PNG correctly -> use OpenCV
        img_ = np.array(cv2.imread(filepath_, cv2.IMREAD_UNCHANGED), dtype=np.uint16)
        # read metadata using PIL
        img2_ = PngImageFile(filepath_)
        list_tmp_string = img2_.text['#02 png_max/png_offset (HEX)'].split(" ")
        png_offset = float.fromhex(list_tmp_string[1])
        list_tmp_string = img2_.text['#04 image min/max (HEX)'].split(" ")
        img_min = float.fromhex(list_tmp_string[0])
        img_max = float.fromhex(list_tmp_string[1])
        img_offset = (img_min + img_max) / 2.0
        scaling_factor = float.fromhex(img2_.text['#06 image scaling factor (HEX)'])
        self.image = (img_ - png_offset) / scaling_factor + img_offset
        self.n_image = len(filepath_)
        self.use_kinetic_energy = use_kinetic_energy
        self.use_pixel_coord = use_pixel_coord
        self.ef_shift = ef_shift

        # read metadata
        target_image_ = PngImageFile(filepath_)
        list_tmp_string_ = target_image_.text['#08 kmin_pixel/kmax_pixel/row (pixel/pixel/-, HEX)'].split(" ")
        self.kmin_pixel = float.fromhex(list_tmp_string_[0])
        self.kmax_pixel = float.fromhex(list_tmp_string_[1])
        self.row = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#10 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-, HEX)'].split(" ")
        self.ekmin_pixel = float.fromhex(list_tmp_string_[0])
        self.ekmax_pixel = float.fromhex(list_tmp_string_[1])
        self.column = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#12 kmin/kmax/row (A^-1/A^-1/-, HEX)'].split(" ")
        self.kmin = float.fromhex(list_tmp_string_[0])
        self.kmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#14 ekmin/ekmax/column (eV/eV/-, HEX)'].split(" ")
        self.ekmin = float.fromhex(list_tmp_string_[0])
        self.ekmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#16 emin/emax/column (eV/eV/-, HEX)'].split(" ")
        self.emin = float.fromhex(list_tmp_string_[0])
        self.emax = float.fromhex(list_tmp_string_[1])

        # store the original Fermi level infromation and fine adjust EF for plotting
        self.ef = self.ekmin - self.emin
        self.emin += self.ef_shift
        self.emax += self.ef_shift

    def get_extent(self):
        if self.use_pixel_coord:
            return [self.kmin_pixel, self.kmax_pixel, self.ekmin_pixel, self.ekmax_pixel]
        elif self.use_kinetic_energy:
            return [self.kmin, self.kmax, self.ekmin, self.ekmax]
        else:
            return [self.kmin, self.kmax, self.emin, self.emax]


class ImageProcessedMulti:
    def __init__(self, filepaths_, use_pixel_coord=False, use_kinetic_energy=False, ef_shift=0.0):
        self.images = []
        for filepath_ in filepaths_:
            # PIL cannot read 16-bit PNG correctly -> use OpenCV
            img_ = np.array(cv2.imread(filepath_, cv2.IMREAD_UNCHANGED), dtype=np.uint16)
            # read metadata using PIL
            img2_ = PngImageFile(filepath_)
            list_tmp_string = img2_.text['#02 png_max/png_offset (HEX)'].split(" ")
            png_offset = float.fromhex(list_tmp_string[1])
            list_tmp_string = img2_.text['#04 image min/max (HEX)'].split(" ")
            img_min = float.fromhex(list_tmp_string[0])
            img_max = float.fromhex(list_tmp_string[1])
            img_offset = (img_min + img_max) / 2.0
            scaling_factor = float.fromhex(img2_.text['#06 image scaling factor (HEX)'])
            self.images.append((img_ - png_offset) / scaling_factor + img_offset)
        self.n_image = len(filepaths_)
        self.use_kinetic_energy = use_kinetic_energy
        self.use_pixel_coord = use_pixel_coord
        self.ef_shift = ef_shift

        # read metadata
        target_image_ = PngImageFile(filepaths_[0])
        list_tmp_string_ = target_image_.text['#08 kmin_pixel/kmax_pixel/row (pixel/pixel/-, HEX)'].split(" ")
        self.kmin_pixel = float.fromhex(list_tmp_string_[0])
        self.kmax_pixel = float.fromhex(list_tmp_string_[1])
        self.row = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#10 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-, HEX)'].split(" ")
        self.ekmin_pixel = float.fromhex(list_tmp_string_[0])
        self.ekmax_pixel = float.fromhex(list_tmp_string_[1])
        self.column = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#12 kmin/kmax/row (A^-1/A^-1/-, HEX)'].split(" ")
        self.kmin = float.fromhex(list_tmp_string_[0])
        self.kmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#14 ekmin/ekmax/column (eV/eV/-, HEX)'].split(" ")
        self.ekmin = float.fromhex(list_tmp_string_[0])
        self.ekmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#16 emin/emax/column (eV/eV/-, HEX)'].split(" ")
        self.emin = float.fromhex(list_tmp_string_[0])
        self.emax = float.fromhex(list_tmp_string_[1])

        # store the original Fermi level infromation and fine adjust EF for plotting
        self.ef = self.ekmin - self.emin
        self.emin += self.ef_shift
        self.emax += self.ef_shift

    def get_extent(self):
        if self.use_pixel_coord:
            return [self.kmin_pixel, self.kmax_pixel, self.ekmin_pixel, self.ekmax_pixel]
        elif self.use_kinetic_energy:
            return [self.kmin, self.kmax, self.ekmin, self.ekmax]
        else:
            return [self.kmin, self.kmax, self.emin, self.emax]


class ImageProcessedTemporal:
    def __init__(self, filepath_, temporal_axis='x', other_axis='energy',
                 use_pixel_coord=False, use_kinetic_energy=False, ef_shift=0.0):
        # PIL cannot read 16-bit PNG correctly -> use OpenCV
        img_ = np.array(cv2.imread(filepath_, cv2.IMREAD_UNCHANGED), dtype=np.uint16)
        # read metadata using PIL
        img2_ = PngImageFile(filepath_)
        list_tmp_string = img2_.text['#02 png_max/png_offset (HEX)'].split(" ")
        png_offset = float.fromhex(list_tmp_string[1])
        list_tmp_string = img2_.text['#04 image min/max (HEX)'].split(" ")
        img_min = float.fromhex(list_tmp_string[0])
        img_max = float.fromhex(list_tmp_string[1])
        img_offset = (img_min + img_max) / 2.0
        scaling_factor = float.fromhex(img2_.text['#06 image scaling factor (HEX)'])
        self.image = (img_ - png_offset) / scaling_factor + img_offset
        self.n_image = len(filepath_)
        self.temporal_axis = temporal_axis
        self.other_axis = other_axis
        self.use_kinetic_energy = use_kinetic_energy
        self.use_pixel_coord = use_pixel_coord
        self.ef_shift = ef_shift

        # read metadata
        target_image_ = PngImageFile(filepath_)
        list_tmp_string_ = target_image_.text['#08 kmin_pixel/kmax_pixel/row (pixel/pixel/-, HEX)'].split(" ")
        self.kmin_pixel = float.fromhex(list_tmp_string_[0])
        self.kmax_pixel = float.fromhex(list_tmp_string_[1])
        self.row = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#10 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-, HEX)'].split(" ")
        self.ekmin_pixel = float.fromhex(list_tmp_string_[0])
        self.ekmax_pixel = float.fromhex(list_tmp_string_[1])
        self.column = int(float.fromhex(list_tmp_string_[2]))
        list_tmp_string_ = target_image_.text['#12 kmin/kmax/row (A^-1/A^-1/-, HEX)'].split(" ")
        self.kmin = float.fromhex(list_tmp_string_[0])
        self.kmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#14 ekmin/ekmax/column (eV/eV/-, HEX)'].split(" ")
        self.ekmin = float.fromhex(list_tmp_string_[0])
        self.ekmax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#16 emin/emax/column (eV/eV/-, HEX)'].split(" ")
        self.emin = float.fromhex(list_tmp_string_[0])
        self.emax = float.fromhex(list_tmp_string_[1])
        list_tmp_string_ = target_image_.text['#20 tmin/tmax/layer (fs/fs/-, HEX)'].split(" ")
        self.tmin = float.fromhex(list_tmp_string_[0])
        self.tmax = float.fromhex(list_tmp_string_[1])
        self.layer = int(float.fromhex(list_tmp_string_[2]))

        # store the original Fermi level infromation and fine adjust EF for plotting
        self.ef = self.ekmin - self.emin
        self.emin += self.ef_shift
        self.emax += self.ef_shift

    def get_extent(self):
        if self.other_axis == 'momentum':
            extent1 = self.kmin_pixel if self.use_pixel_coord else self.kmin
            extent2 = self.kmax_pixel if self.use_pixel_coord else self.kmax
        else:  # when self.other_axis == 'energy'
            if self.use_kinetic_energy:
                extent1 = self.ekmin_pixel if self.use_pixel_coord else self.ekmin
                extent2 = self.ekmax_pixel if self.use_pixel_coord else self.ekmax
            else:
                extent1 = self.emin
                extent2 = self.emax

        if self.temporal_axis == 'y':
            return [extent1, extent2, self.tmin, self.tmax]
        else:  # when self.temporal_axis == 'x'
            return [self.tmin, self.tmax, extent1, extent2]


def save_image_processed(filepath, img, row, column,
                         kmin_pixel, kmax_pixel, ekmin_pixel, ekmax_pixel,
                         kmin, kmax, ekmin, ekmax, emin, emax, figs_clim):
    version_zapes_image_processed = 'v01'
    # NOTE: need to guarantee all the variables have consistent format: float in python (= double precision in C)
    png_max_ = 65535  # 16 bit png
    png_offset_ = float(0.5 * png_max_)
    img_min_ = float(np.min(img))
    img_max_ = float(np.max(img))
    # clamp too small values to zero
    if np.abs(img_min_) < 1e-100:
        img_min_ = 0.0
    if np.abs(img_max_) < 1e-100:
        img_max_ = 0.0
    img_offset_ = 0.5 * (img_min_ + img_max_)
    if img_min_ != img_max_:
        scaling_factor_ = png_max_ / (img_max_ - img_min_)
    else:
        scaling_factor_ = 1.0
    png_img_ = np.array(png_offset_ + (img - img_offset_) * scaling_factor_, dtype=np.uint16)
    im_ = Image.fromarray(png_img_)
    im_.save(filepath)

    # write png metadata
    target_image_ = PngImageFile(filepath)
    png_metadata_ = PngInfo()
    png_metadata_.add_text('#00 zapes processed image PNG version', version_zapes_image_processed)
    png_metadata_.add_text('#01 png_max/png_offset', '{:d} {:.1f}'.format(png_max_, png_offset_))
    png_metadata_.add_text('#02 png_max/png_offset (HEX)', '{} {}'.format(float(png_max_).hex(), png_offset_.hex()))
    png_metadata_.add_text('#03 image min/max', '{:.6f} {:.6f}'.format(img_min_, img_max_))
    png_metadata_.add_text('#04 image min/max (HEX)', '{} {}'.format(img_min_.hex(), img_max_.hex()))
    png_metadata_.add_text('#05 image scaling factor', '{:.8f}'.format(scaling_factor_))
    png_metadata_.add_text('#06 image scaling factor (HEX)', '{}'.format(scaling_factor_.hex()))
    png_metadata_.add_text('#07 kmin_pixel/kmax_pixel/row (pixel/pixel/-)',
                           '{:d} {:d} {:d}'.format(kmin_pixel, kmax_pixel, row))
    png_metadata_.add_text('#08 kmin_pixel/kmax_pixel/row (pixel/pixel/-, HEX)',
                           '{} {} {}'.format(float(kmin_pixel).hex(), float(kmax_pixel).hex(), float(row).hex()))
    png_metadata_.add_text('#09 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-)',
                           '{:d} {:d} {:d}'.format(ekmin_pixel, ekmax_pixel, column))
    png_metadata_.add_text('#10 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-, HEX)',
                           '{} {} {}'.format(float(ekmin_pixel).hex(), float(ekmax_pixel).hex(), float(column).hex()))
    png_metadata_.add_text('#11 kmin/kmax/row (A^-1/A^-1/-)',
                           '{:.3f} {:.3f} {:d}'.format(kmin, kmax, row))
    png_metadata_.add_text('#12 kmin/kmax/row (A^-1/A^-1/-, HEX)',
                           '{} {} {}'.format(float(kmin).hex(), float(kmax).hex(), float(row).hex()))
    png_metadata_.add_text('#13 ekmin/ekmax/column (eV/eV/-)',
                           '{:.3f} {:.3f} {:d}'.format(ekmin, ekmax, column))
    png_metadata_.add_text('#14 ekmin/ekmax/column (eV/eV/-, HEX)',
                           '{} {} {}'.format(float(ekmin).hex(), float(ekmax).hex(), float(column).hex()))
    png_metadata_.add_text('#15 emin/emax/column (eV/eV/-)',
                           '{:.3f} {:.3f} {:d}'.format(emin, emax, column))
    png_metadata_.add_text('#16 emin/emax/column (eV/eV/-, HEX)',
                           '{} {} {}'.format(float(emin).hex(), float(emax).hex(), float(column).hex()))
    png_metadata_.add_text('#17 clim_min/clim_max',
                           '{:.6f} {:.6f}'.format(figs_clim[0], figs_clim[1]))
    png_metadata_.add_text('#18 clim_min/clim_max (HEX)',
                           '{} {}'.format(float(figs_clim[0]).hex(), float(figs_clim[1]).hex()))
    target_image_.save(filepath, pnginfo=png_metadata_)


def save_image_processed_temporal(filepath, img, row, column, layer,
                                  kmin_pixel, kmax_pixel, ekmin_pixel, ekmax_pixel,
                                  kmin, kmax, ekmin, ekmax, emin, emax, figs_clim, delay_time):
    version_zapes_image_processed = 'v01'
    # NOTE: need to guarantee all the variables have consistent format: float in python (= double precision in C)
    png_max_ = 65535  # 16 bit png
    png_offset_ = float(0.5 * png_max_)
    img_min_ = float(np.min(img))
    img_max_ = float(np.max(img))
    # clamp too small values to zero
    if np.abs(img_min_) < 1e-100:
        img_min_ = 0.0
    if np.abs(img_max_) < 1e-100:
        img_max_ = 0.0
    img_offset_ = 0.5 * (img_min_ + img_max_)
    if img_min_ != img_max_:
        scaling_factor_ = png_max_ / (img_max_ - img_min_)
    else:
        scaling_factor_ = 1.0
    png_img_ = np.array(png_offset_ + (img - img_offset_) * scaling_factor_, dtype=np.uint16)
    im_ = Image.fromarray(png_img_)
    im_.save(filepath)

    # write png metadata
    target_image_ = PngImageFile(filepath)
    png_metadata_ = PngInfo()
    png_metadata_.add_text('#00 zapes processed image PNG version', version_zapes_image_processed)
    png_metadata_.add_text('#01 png_max/png_offset', '{:d} {:.1f}'.format(png_max_, png_offset_))
    png_metadata_.add_text('#02 png_max/png_offset (HEX)', '{} {}'.format(float(png_max_).hex(), png_offset_.hex()))
    png_metadata_.add_text('#03 image min/max', '{:.6f} {:.6f}'.format(img_min_, img_max_))
    png_metadata_.add_text('#04 image min/max (HEX)', '{} {}'.format(img_min_.hex(), img_max_.hex()))
    png_metadata_.add_text('#05 image scaling factor', '{:.8f}'.format(scaling_factor_))
    png_metadata_.add_text('#06 image scaling factor (HEX)', '{}'.format(scaling_factor_.hex()))
    png_metadata_.add_text('#07 kmin_pixel/kmax_pixel/row (pixel/pixel/-)',
                           '{:d} {:d} {:d}'.format(kmin_pixel, kmax_pixel, row))
    png_metadata_.add_text('#08 kmin_pixel/kmax_pixel/row (pixel/pixel/-, HEX)',
                           '{} {} {}'.format(float(kmin_pixel).hex(), float(kmax_pixel).hex(), float(row).hex()))
    png_metadata_.add_text('#09 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-)',
                           '{:d} {:d} {:d}'.format(ekmin_pixel, ekmax_pixel, column))
    png_metadata_.add_text('#10 ekmin_pixel/ekmax_pixel/column (pixel/pixel/-, HEX)',
                           '{} {} {}'.format(float(ekmin_pixel).hex(), float(ekmax_pixel).hex(), float(column).hex()))
    png_metadata_.add_text('#11 kmin/kmax/row (A^-1/A^-1/-)',
                           '{:.3f} {:.3f} {:d}'.format(kmin, kmax, row))
    png_metadata_.add_text('#12 kmin/kmax/row (A^-1/A^-1/-, HEX)',
                           '{} {} {}'.format(float(kmin).hex(), float(kmax).hex(), float(row).hex()))
    png_metadata_.add_text('#13 ekmin/ekmax/column (eV/eV/-)',
                           '{:.3f} {:.3f} {:d}'.format(ekmin, ekmax, column))
    png_metadata_.add_text('#14 ekmin/ekmax/column (eV/eV/-, HEX)',
                           '{} {} {}'.format(float(ekmin).hex(), float(ekmax).hex(), float(column).hex()))
    png_metadata_.add_text('#15 emin/emax/column (eV/eV/-)',
                           '{:.3f} {:.3f} {:d}'.format(emin, emax, column))
    png_metadata_.add_text('#16 emin/emax/column (eV/eV/-, HEX)',
                           '{} {} {}'.format(float(emin).hex(), float(emax).hex(), float(column).hex()))
    png_metadata_.add_text('#17 clim_min/clim_max',
                           '{:.6f} {:.6f}'.format(figs_clim[0], figs_clim[1]))
    png_metadata_.add_text('#18 clim_min/clim_max (HEX)',
                           '{} {}'.format(float(figs_clim[0]).hex(), float(figs_clim[1]).hex()))
    png_metadata_.add_text('#19 tmin/tmax/layer (fs/fs/-)',
                           '{:.3f} {:.3f} {:d}'.format(delay_time[0], delay_time[-1], layer))
    png_metadata_.add_text('#20 tmin/tmax/layer (fs/fs/-, HEX)',
                           '{} {} {}'.format(float(delay_time[0]).hex(), float(delay_time[-1]).hex(),
                                             float(layer).hex()))
    target_image_.save(filepath, pnginfo=png_metadata_)


def util_cmap(cmap_name):
    if cmap_name in ['gnu_v2', 'gnu_v2_r', 'whitejet',
                     'schueler1', 'schueler1_r', 'schueler2', 'schueler2_r',
                     'schueler3', 'schueler3_r', 'schueler4', 'schueler4_r']:
        return cmap_original(cmap_name)
    else:
        return plt.get_cmap(cmap_name)


def util_cmap_tuple(tuple_cmap_name):
    tuple_ = ()
    for cmap_name_ in tuple_cmap_name:
        tuple_ += (util_cmap(cmap_name_),)
    return tuple_


def cmap_original(cmap_name):
    if cmap_name == 'gnu_v2':
        return LinearSegmentedColormap.from_list('gnu_v2', [(0.0, [1.0, 1.0, 0.0]),
                                                            (1.0 - 0.75, [1.0, 0.5, 0.0]),
                                                            (0.5, [0.8, 0.2, 0.0]),
                                                            (1.0 - 0.22, [0.5, 0.0, 1.0]),
                                                            (1.0, [1.0, 1.0, 1.0])])
    elif cmap_name == 'gnu_v2_r':
        # defined by Johannes as colormap_gnu_v2 in vossplot
        return LinearSegmentedColormap.from_list('gnu_v2_r', [(0.0, [1.0, 1.0, 1.0]),
                                                              (0.22, [0.5, 0.0, 1.0]),
                                                              (0.5, [0.8, 0.2, 0.0]),
                                                              (0.75, [1.0, 0.5, 0.0]),
                                                              (1.0, [1.0, 1.0, 0.0])])
    elif cmap_name == 'whitejet':
        list_colormap = np.loadtxt(os.path.dirname(__file__) + '/colormaps/colormap_whitejet.dat', delimiter=',')
        return ListedColormap(list_colormap, name='whitejet')
    elif cmap_name == 'schueler1':
        # defined by Dr. Schueler as 'GetCmap'
        return LinearSegmentedColormap.from_list('schueler1', [(0.0, 'black'),
                                                               (0.25, 'navy'),
                                                               (0.5, 'blue'),
                                                               (0.75, 'cyan'),
                                                               (1.0, 'white')])
    elif cmap_name == 'schueler1_r':
        return LinearSegmentedColormap.from_list('schueler1_r', [(0.0, 'white'),
                                                                 (0.25, 'cyan'),
                                                                 (0.5, 'blue'),
                                                                 (0.75, 'navy'),
                                                                 (1.0, 'black')])
    elif cmap_name in ['schueler2', 'schueler2_r']:
        # defined by Dr. Schueler as 'cmap_arpes'
        gr1 = [224. / 255, 224. / 255, 244. / 255]
        b0 = [0.2, 0.2, 1.0]
        g1 = [80. / 255, 185. / 255, 90. / 255]
        y1 = [1.0, 1.0, 81. / 255]
        br1 = [120. / 255, 81. / 255, 0]
        br2 = [222. / 255, 170. / 255, 135. / 255]
        if cmap_name == 'schueler2':
            return LinearSegmentedColormap.from_list('schueler2', [(0.0, br2),
                                                                   (1. / 6., br1),
                                                                   (2. / 6., y1),
                                                                   (3. / 6., g1),
                                                                   (4. / 6., b0),
                                                                   (5. / 6., gr1),
                                                                   (1.0, 'white')])
        else:
            return LinearSegmentedColormap.from_list('schueler2_r', [(0.0, 'white'),
                                                                     (1. / 6., gr1),
                                                                     (2. / 6., b0),
                                                                     (3. / 6., g1),
                                                                     (4. / 6., y1),
                                                                     (5. / 6., br1),
                                                                     (1.0, br2)])
    elif cmap_name == 'schueler3':
        return LinearSegmentedColormap.from_list('schueler3', [(0.0, 'red'),
                                                               (1. / 6., 'yellow'),
                                                               (2. / 6., 'green'),
                                                               (3. / 6., 'cyan'),
                                                               (4. / 6., 'blue'),
                                                               (5. / 6., 'indigo'),
                                                               (1.0, 'white')])
    elif cmap_name == 'schueler3_r':
        # defined by Dr. Schueler as 'cmap_arpes2'
        return LinearSegmentedColormap.from_list('schueler3_r', [(0.0, 'white'),
                                                                 (1. / 6., 'indigo'),
                                                                 (2. / 6., 'blue'),
                                                                 (3. / 6., 'cyan'),
                                                                 (4. / 6., 'green'),
                                                                 (5. / 6., 'yellow'),
                                                                 (1.0, 'red')])
    elif cmap_name == 'schueler4':
        # defined by Dr. Schueler as 'cmap_arpes2'
        return LinearSegmentedColormap.from_list('schueler4', [(0.0, 'red'),
                                                               (1.0 - 3. / 4., 'yellow'),
                                                               (1.0 - 0.3, 'green'),
                                                               (1.0 - 1. / 5., 'royalblue'),
                                                               (1.0, 'whitesmoke')])
    elif cmap_name == 'schueler4_r':
        # defined by Dr. Schueler as 'cmap_arpes2'
        return LinearSegmentedColormap.from_list('schueler4_r', [(0.0, 'whitesmoke'),
                                                                 (1. / 5., 'royalblue'),
                                                                 (0.3, 'green'),
                                                                 (3. / 4., 'yellow'),
                                                                 (1.0, 'red')])


def sort_2ppe_filenames(filenames_, rex_2ppe_=r'(2ppe\d+_)(\d+)(.png)'):
    irex_tmp = re.compile(rex_2ppe_)
    ilist_tmp = []
    for fn_ in filenames_:
        fn_base, fn_body = os.path.split(fn_)
        m_ = irex_tmp.match(fn_body)
        g_ = m_.groups()
        if g_ is not None:
            ilist_tmp.append(int(g_[1]))
    return np.array(filenames_)[np.argsort(np.array(ilist_tmp))]


def rm_dead(image_):
    dead_pixels = [[497, 490], [346, 574], [452, 219], [415, 165], [325, 782], [317, 65], [27, 710]]
    for dp_ in dead_pixels:
        image_[tuple(dp_)] = (image_[dp_[0] - 1, dp_[1]]
                              + image_[dp_[0] + 1, dp_[1]]
                              + image_[dp_[0], dp_[1] - 1]
                              + image_[dp_[0], dp_[1] + 1]) / 4.
    return image_


def bleach_prep(bg_file_path_=None, n_convolve_=40):
    if bg_file_path_ is not None:
        img_bg_tmp = load_zapes2d_16bit(bg_file_path_, None, False, None)
        img_bg_convolve_tmp = ndimage.convolve(img_bg_tmp,
                                               np.ones((n_convolve_, n_convolve_)) / float(n_convolve_ ** 2),
                                               mode='reflect')
        img_bg_convolve_norm_tmp = img_bg_convolve_tmp / np.max(img_bg_convolve_tmp)
        return img_bg_convolve_norm_tmp
    else:
        return None


def load_zapes2d_16bit(filename_, list_roi_bg_, flag_bg_correction_, image_bg_bleach_):
    # A zapes data file is written as a 16-bit png file
    # Nevertheless, for compatibility with OpenCV modules, load as 32-bit float
    image_tmp = np.array(cv2.imread(filename_, cv2.IMREAD_UNCHANGED), dtype=np.float32)
    image_tmp = rm_dead(image_tmp)
    if list_roi_bg_ is not None:
        bg_avg = np.average(image_tmp[list_roi_bg_[1]:list_roi_bg_[3], list_roi_bg_[0]:list_roi_bg_[2]])
        image_tmp -= bg_avg
    if flag_bg_correction_:
        image_tmp /= image_bg_bleach_
    return image_tmp


def zapes_reflect_filter(image_, filter_data_path_, version_trim_filter_):
    # reflect across contour lines surrounding a signal area

    # load contour data
    # defined by the OpenCV contour format: [[[y, x]], ...]
    # -> ex. contour_lower_tmp[:, 0, 1] gives x coordinates of the lower contour and ~[:, 0, 0] gives y coordinates
    # defined by the pixel coordinate (not roi-pixel coordinate!)
    # defined in the clockwise direction (lower -> right -> upper -> left)
    contour_lower_tmp = np.load(filter_data_path_ + '/trim_filter_contour_lower_' + version_trim_filter_ + '.npy')
    contour_right_tmp = np.load(filter_data_path_ + '/trim_filter_contour_right_' + version_trim_filter_ + '.npy')
    contour_upper_tmp = np.load(filter_data_path_ + '/trim_filter_contour_upper_' + version_trim_filter_ + '.npy')
    contour_left_tmp = np.load(filter_data_path_ + '/trim_filter_contour_left_' + version_trim_filter_ + '.npy')

    img_reflected_tmp = image_.copy()
    row_tmp, column_tmp = img_reflected_tmp.shape
    # reflect across each contour line
    for i_ in range(contour_lower_tmp[0, 0, 1], contour_lower_tmp[-1, 0, 1]):
        edge_tmp = contour_lower_tmp[i_ - contour_lower_tmp[0, 0, 1], 0, 0]
        for j_ in range(0, edge_tmp + 1):
            img_reflected_tmp[i_, j_] = image_[i_, 2 * edge_tmp - j_]  # reflect
    for j_ in range(contour_right_tmp[0, 0, 0], contour_right_tmp[-1, 0, 0] + 1):
        edge_tmp = contour_right_tmp[j_ - contour_right_tmp[0, 0, 0], 0, 1]
        for i_ in range(edge_tmp, row_tmp):
            img_reflected_tmp[i_, j_] = image_[2 * edge_tmp - i_, j_]  # reflect
    for i_ in range(contour_upper_tmp[-1, 0, 1] + 1, contour_upper_tmp[0, 0, 1]):
        edge_tmp = contour_upper_tmp[contour_upper_tmp[0, 0, 1] - i_, 0, 0]
        for j_ in range(edge_tmp, column_tmp):
            img_reflected_tmp[i_, j_] = image_[i_, 2 * edge_tmp - j_]  # reflect
    for j_ in range(contour_left_tmp[-1, 0, 0] + 1, contour_left_tmp[0, 0, 0]):
        edge_tmp = contour_left_tmp[contour_left_tmp[0, 0, 0] - j_, 0, 1]
        for i_ in range(0, edge_tmp + 1):
            img_reflected_tmp[i_, j_] = image_[2 * edge_tmp - i_, j_]  # reflect

    img_reflected_tmp2 = img_reflected_tmp.copy()
    # process remaining areas
    for i_ in range(0, contour_lower_tmp[0, 0, 1]):
        edge_tmp = contour_left_tmp[-1, 0, 0]
        for j_ in range(0, edge_tmp):
            img_reflected_tmp2[i_, j_] = img_reflected_tmp[i_, 2 * edge_tmp - j_]  # reflect
    for i_ in range(contour_lower_tmp[-1, 0, 1], row_tmp):
        edge_tmp = contour_right_tmp[0, 0, 0]
        for j_ in range(0, edge_tmp):
            img_reflected_tmp2[i_, j_] = img_reflected_tmp[i_, 2 * edge_tmp - j_]  # reflect
    for i_ in range(contour_upper_tmp[0, 0, 1], row_tmp):
        edge_tmp = contour_right_tmp[-1, 0, 0]
        for j_ in range(edge_tmp, column_tmp):
            img_reflected_tmp2[i_, j_] = img_reflected_tmp[i_, 2 * edge_tmp - j_]  # reflect
    for i_ in range(0, contour_upper_tmp[-1, 0, 1]):
        edge_tmp = contour_left_tmp[0, 0, 0]
        for j_ in range(edge_tmp, column_tmp):
            img_reflected_tmp2[i_, j_] = img_reflected_tmp[i_, 2 * edge_tmp - j_]  # reflect

    return img_reflected_tmp2


def prepare_zapes_trim_filter(image_, filter_data_path_, version_trim_filter_):
    # trim along contour lines surrounding a signal area (replace values outside with fill_value)

    # load contour data
    # defined by the OpenCV contour format: [[[y, x]], ...]
    # defined by the pixel coordinate (not roi-pixel coordinate!)
    contour_tmp = np.load(filter_data_path_ + '/trim_filter_contour_total_' + version_trim_filter_ + '.npy')
    row_tmp, column_tmp = image_.shape
    img_trim_tmp = np.zeros((row_tmp, column_tmp))
    for i_ in range(row_tmp):
        for j_ in range(column_tmp):
            # judge whether each point is inside/outside/on a contour (return +1/-1/0 values)
            if cv2.pointPolygonTest(contour_tmp, (j_, i_), measureDist=False) > 0:
                img_trim_tmp[i_, j_] = 1
    return img_trim_tmp


############################################
# functions for integrating PES intensities
############################################
# list_box_pixel format:
# dEk_dk_pixel = 2.25  # pixel/pixel = no unit
# box_right_pixel = [k_origin_pixel - 15 - 10, 700, 950, 100, dEk_dk_pixel]
# box_left_pixel = [k_origin_pixel + 15 - 10, 700, 950, 100, -dEk_dk_pixel]
# # k0 (pixel), Ek0 (pixel), Ek1 (pixel), delta_k (pixel), dEk/dk


def draw_integration_box_pixel(list_box_pixel_, box_color_, ax_):
    e0_pixel = list_box_pixel_[1]
    e1_pixel = list_box_pixel_[2]
    n_edc = int(e1_pixel - e0_pixel + 1)
    k0_pixel = list_box_pixel_[0]
    delta_k_pixel = list_box_pixel_[3]
    gradient_tmp = list_box_pixel_[4]

    if gradient_tmp > 0:
        ax_.add_patch(patches.Polygon(xy=[(k0_pixel, e0_pixel), (k0_pixel + delta_k_pixel, e0_pixel),
                                          (k0_pixel + (n_edc - 1) / gradient_tmp + delta_k_pixel, e1_pixel),
                                          (k0_pixel + (n_edc - 1) / gradient_tmp, e1_pixel)],
                                      edgecolor=box_color_, fill=False))
    else:
        ax_.add_patch(patches.Polygon(xy=[(k0_pixel - delta_k_pixel, e0_pixel), (k0_pixel, e0_pixel),
                                          (k0_pixel + (n_edc - 1) / gradient_tmp, e1_pixel),
                                          (k0_pixel + (n_edc - 1) / gradient_tmp - delta_k_pixel, e1_pixel)],
                                      edgecolor=box_color_, fill=False))


def draw_integration_box(list_box_pixel_, box_color_, ax_, ek_, efactor_, ek_pixel_, k_origin_pixel_, ef_):
    e0_pixel = list_box_pixel_[1]
    e1_pixel = list_box_pixel_[2]
    n_edc = int(e1_pixel - e0_pixel + 1)
    k0_pixel = list_box_pixel_[0]
    delta_k_pixel = list_box_pixel_[3]
    gradient_tmp = list_box_pixel_[4]

    e0_tmp = ek_ + (e0_pixel - ek_pixel_) * efactor_ - ef_  # eV unit
    e1_tmp = ek_ + (e1_pixel - ek_pixel_) * efactor_ - ef_  # eV unit
    k0_tmp = 0.5123 * np.sqrt(ek_) * np.sin(
        (k0_pixel - k_origin_pixel_) * (26.0 / 351.0) * (np.pi / 180.0))  # A^-1 unit
    k0_tmp_plus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + delta_k_pixel - k_origin_pixel_)
                                                         * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp - k_origin_pixel_)
                                                     * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient_plus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp
                                                                   + delta_k_pixel - k_origin_pixel_)
                                                                  * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_minus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel - delta_k_pixel - k_origin_pixel_)
                                                          * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient_minus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp
                                                                    - delta_k_pixel - k_origin_pixel_)
                                                                   * (26.0 / 351.0) * (np.pi / 180.0))

    if gradient_tmp > 0:
        ax_.add_patch(patches.Polygon(xy=[(k0_tmp, e0_tmp), (k0_tmp_plus_delta_k, e0_tmp),
                                          (k0_tmp_gradient_plus_delta_k, e1_tmp),
                                          (k0_tmp_gradient, e1_tmp)],
                                      edgecolor=box_color_, fill=False))
    else:
        ax_.add_patch(patches.Polygon(xy=[(k0_tmp_minus_delta_k, e0_tmp), (k0_tmp, e0_tmp),
                                          (k0_tmp_gradient, e1_tmp),
                                          (k0_tmp_gradient_minus_delta_k, e1_tmp)],
                                      edgecolor=box_color_, fill=False))


def draw_integration_box_with_line(list_box_pixel_, box_color_, ax_, ek_, efactor_, ek_pixel_, k_origin_pixel_, ef_,
                                   energy_):
    e0_pixel = list_box_pixel_[1]
    e1_pixel = list_box_pixel_[2]
    n_edc = int(e1_pixel - e0_pixel + 1)
    k0_pixel = list_box_pixel_[0]
    delta_k_pixel = list_box_pixel_[3]
    gradient_tmp = list_box_pixel_[4]

    e0_tmp = ek_ + (e0_pixel - ek_pixel_) * efactor_ - ef_  # eV unit
    e1_tmp = ek_ + (e1_pixel - ek_pixel_) * efactor_ - ef_  # eV unit
    k0_tmp = 0.5123 * np.sqrt(ek_) * np.sin(
        (k0_pixel - k_origin_pixel_) * (26.0 / 351.0) * (np.pi / 180.0))  # A^-1 unit
    k0_tmp_plus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + delta_k_pixel - k_origin_pixel_)
                                                         * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp - k_origin_pixel_)
                                                     * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient_plus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp
                                                                   + delta_k_pixel - k_origin_pixel_)
                                                                  * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_minus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel - delta_k_pixel - k_origin_pixel_)
                                                          * (26.0 / 351.0) * (np.pi / 180.0))
    k0_tmp_gradient_minus_delta_k = 0.5123 * np.sqrt(ek_) * np.sin((k0_pixel + (n_edc - 1) / gradient_tmp
                                                                    - delta_k_pixel - k_origin_pixel_)
                                                                   * (26.0 / 351.0) * (np.pi / 180.0))

    if gradient_tmp > 0:
        ax_.add_patch(patches.Polygon(xy=[(k0_tmp, e0_tmp), (k0_tmp_plus_delta_k, e0_tmp),
                                          (k0_tmp_gradient_plus_delta_k, e1_tmp),
                                          (k0_tmp_gradient, e1_tmp)],
                                      edgecolor=box_color_, fill=False))
        k0_line_start_ = k0_tmp + (k0_tmp_plus_delta_k - k0_tmp) / (e1_tmp - e0_tmp) * (energy_ - e0_tmp)
        k0_line_end_ = k0_tmp_plus_delta_k \
                       + (k0_tmp_gradient_plus_delta_k - k0_tmp_plus_delta_k) / (e1_tmp - e0_tmp) * (energy_ - e0_tmp)
        ax_.hlines(energy_, k0_line_start_, k0_line_end_, color=box_color_, lw=3)
    else:
        ax_.add_patch(patches.Polygon(xy=[(k0_tmp_minus_delta_k, e0_tmp), (k0_tmp, e0_tmp),
                                          (k0_tmp_gradient, e1_tmp),
                                          (k0_tmp_gradient_minus_delta_k, e1_tmp)],
                                      edgecolor=box_color_, fill=False))
        k0_line_start_ = k0_tmp + (k0_tmp_minus_delta_k - k0_tmp) / (e1_tmp - e0_tmp) * (energy_ - e0_tmp)
        k0_line_end_ = k0_tmp_minus_delta_k \
                       + (k0_tmp_gradient_minus_delta_k - k0_tmp_gradient) / (e1_tmp - e0_tmp) * (energy_ - e0_tmp)
        ax_.hlines(energy_, k0_line_end_, k0_line_start_, color=box_color_, lw=3)


def integrate_in_box(image, list_box_pixel, kmin_pixel_, ekmin_pixel_, kmin_shift_, ekmin_shift_):
    # kmin_shift/ekmin_shift: the shift of the origin in momentum/kinetic-energy coordinates caused by padding
    e0_pixel_tmp = int(list_box_pixel[1])
    e1_pixel_tmp = int(list_box_pixel[2])
    n_edc_tmp = e1_pixel_tmp - e0_pixel_tmp + 1
    edc_tmp = np.empty(n_edc_tmp)

    k0_pixel_tmp = int(list_box_pixel[0])
    delta_k_pixel_tmp = int(list_box_pixel[3])
    gradient_tmp = list_box_pixel[4]

    if gradient_tmp > 0:
        for pos in range(n_edc_tmp):
            edc_tmp[pos] = np.sum(image[int(k0_pixel_tmp + pos / gradient_tmp - (kmin_pixel_ - kmin_shift_)):
                                        int(k0_pixel_tmp + pos / gradient_tmp
                                            + delta_k_pixel_tmp - (kmin_pixel_ - kmin_shift_)),
                                  int(e0_pixel_tmp + pos - (ekmin_pixel_ - ekmin_shift_))])
    else:
        for pos in range(n_edc_tmp):
            edc_tmp[pos] = np.sum(image[int(k0_pixel_tmp + pos / gradient_tmp
                                            - delta_k_pixel_tmp - (kmin_pixel_ - kmin_shift_)):
                                        int(k0_pixel_tmp + pos / gradient_tmp - (kmin_pixel_ - kmin_shift_)),
                                  int(e0_pixel_tmp + pos - (ekmin_pixel_ - ekmin_shift_))])
    return edc_tmp
