import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

# load homemade python moduels
import absplot_v03 as absplot
import zapes_utilities_v03 as zutil

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_field_dep_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.33
emax_plot = 0.32
tmin_plot = -115
tmax_plot = 20
t_offset_strong = -4.0
t_offset_weak = -4.0
list_delay_time_strong = [-100, -60, -50, -40]
list_delay_time_weak = [-96, -50, -42, -34]
colormap = 'gnu_v2_r'
ticks_color = 'black'
figs_clim_second_derivative_strong = [0.0055, 0.035]
figs_clim_second_derivative_weak = [0.0055, 0.03]

lw_dft_thick = 1.2
lw_dft_thin = 0.3
color_dft = 'gray'
color_dft_projection = 'gray'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=108)

#######################################################################################################################
# Plot structure 1 (Waveform panel, weak field)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=2, bottom=54, width=62, height=54)
# eframe1.visualize_frame()

# create single plotting frame
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1, left=10.5, bottom=16.25, width=45, height=30)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load waveform files
tmp_filepath = './waveform_extraction_2ppe7044_shift2d_n40_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset_weak
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel
E_parallel_deconvolved *= 1e-3
E_parallel_smooth_deconvolved *= 1e-3
E_parallel_fit_rescaled_deconvolved *= 1e-3

# set plot range
pframe11.ax.set_xlim(tmin_plot, tmax_plot)
scale_efield = 1.0  # np.max(np.abs(E_parallel_smooth_deconvolved)) * 1.5
pframe11.ax.set_ylim(-scale_efield, scale_efield)
pframe11.ax.set_ylim(-0.6, 0.6)

# plot data
pframe11.ax.plot(delay_time, E_parallel_deconvolved, color='black', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe11.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1)
for time_tmp in list_delay_time_weak:
    pframe11.ax.vlines(time_tmp + t_offset_weak, -scale_efield, scale_efield, color='blue', lw=1.5, alpha=0.3)

# set other attributes
pframe11.ax.text(0.5, 1.05, r'$E_{\mathrm{peak}}$ = 0.5' + r' MV cm$^{-1}$',
                 transform=pframe11.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe11.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe11.fix_yaxis_interval(0.5)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (ARPES images, weak field)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=59, bottom=54, width=150, height=54)
# eframe2.visualize_frame()

# create multiple plot frames
img_left = 12.5
img_bottom = 10
img_width = 30.5
img_hoffset = 32
img_height = 48 * (30.5/35)
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe22 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe22.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe23 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe23.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe24 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe24.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_2ppe7044_0_sc_2deriv_narrow3_shift2d_n40_v11.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe21.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_weak, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe7044_54_sc_2deriv_narrow3_shift2d_n40_v11.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe22.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_weak, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe7044_58_sc_2deriv_narrow3_shift2d_n40_v11.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe23.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_weak, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2ppe7044_62_sc_2deriv_narrow3_shift2d_n40_v11.png'
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.0)
pframe24.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_weak, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# load Floquet center frequency
tmp_filepath = './waveform_params_2ppe7044_shift2d_n40_v19.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_parallel = data_waveform_params[5]

# plot DFT band structures
for pframe in [pframe21, pframe22, pframe23, pframe24]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    if pframe == pframe21:
        pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)
    else:
        e_Floquet = hb * 2.0 * np.pi * freq_fit_parallel * 1e-3
        for n_Floquet in [-3, -2, -1, 1, 2, 3]:
            pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)
            pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)

# set attributes
pframe21.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe22, pframe23, pframe24]:
    pframe.ax.tick_params(labelleft=False)
for idx, pframe in enumerate([pframe21, pframe22, pframe23, pframe24]):
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color, lw=0.5, ls=(0, (8, 5)))
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    if list_delay_time_weak[idx] > 0:
        str_sign = '+'
    else:
        str_sign = ''
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{}{:.0f} fs'.format(str_sign, list_delay_time_weak[idx] + t_offset_weak),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color=ticks_color)
    pframe.fix_yaxis_interval(0.1)

# add colorbar
pframe24.add_colorbar(cleft=33, cbottom=8, cwidth=2, cheight=25)
pframe24.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color, show_ticks=False)
pframe24.cax.text(0.5, 1.05, 'max', transform=pframe24.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe24.cax.text(0.5, -0.13, 'min', transform=pframe24.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (Waveform panel, strong field)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=2, bottom=0, width=62, height=54)
# eframe3.visualize_frame()

# create single plotting frame
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3, left=10.5, bottom=16.25, width=45, height=30)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load waveform files
tmp_filepath = './waveform_extraction_2ppe8621_shift2d_n60_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset_strong
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel
E_parallel_deconvolved *= 1e-3
E_parallel_smooth_deconvolved *= 1e-3
E_parallel_fit_rescaled_deconvolved *= 1e-3

# set plot range
pframe31.ax.set_xlim(tmin_plot, tmax_plot)
scale_efield = np.max(np.abs(E_parallel_smooth_deconvolved)) * 1.25 * 1.1
pframe31.ax.set_ylim(-scale_efield, scale_efield)
pframe31.ax.set_ylim(-1, 1)

# plot data
pframe31.ax.plot(delay_time, E_parallel_deconvolved, color='black', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe31.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1)
for time_tmp in list_delay_time_strong:
    pframe31.ax.vlines(time_tmp, -scale_efield, scale_efield, color='blue', lw=1.5, alpha=0.3)

# set other attributes
pframe31.ax.text(0.5, 1.05, r'$E_{\mathrm{peak}}$ = 0.8' + ' MV cm$^{-1}$',
                 transform=pframe31.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe31.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe31.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe31.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Plot structure 4 (ARPES images, strong field)
#######################################################################################################################
# create empty frame for this structure
eframe4 = absplot.EmptyFrame(mainframe=mainframe, left=59, bottom=0, width=150, height=54)
# eframe4.visualize_frame()

# create multiple plot frames
img_left = 12.5
img_bottom = 10
img_width = 30.5
img_hoffset = 32
img_height = 42
pframe41 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe41.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe42 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe42.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe43 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe43.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe44 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe44.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_2ppe8621_0_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.02)
pframe41.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_strong, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe8621_1_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.02)
pframe42.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_strong, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe8621_8_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.02)
pframe43.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_strong, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2ppe8621_13_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.02)
pframe44.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_strong, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# load Floquet center frequency
tmp_filepath = './waveform_params_2ppe8621_shift2d_n60_v19.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_parallel = data_waveform_params[5]

# plot DFT band structures
for pframe in [pframe41, pframe42, pframe43, pframe44]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    if pframe == pframe41:
        pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)
    else:
        e_Floquet = hb * 2.0 * np.pi * freq_fit_parallel * 1e-3
        for n_Floquet in [-3, -2, -1, 1, 2, 3]:
            pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)
            pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)

# set attributes
pframe41.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe42, pframe43, pframe44]:
    pframe.ax.tick_params(labelleft=False)
for idx, pframe in enumerate([pframe41, pframe42, pframe43, pframe44]):
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color, lw=0.5, ls=(0, (8, 5)))
    if list_delay_time_strong[idx] > 0:
        str_sign = '+'
    else:
        str_sign = ''
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{}{:.0f} fs'.format(str_sign, list_delay_time_strong[idx]),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color=ticks_color)
    pframe.fix_yaxis_interval(0.1)

#######################################################################################################################
# End of Plot structure 4
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(0.5, 105, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(60, 105, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(0.5, 51, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(60, 51, 'd', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
