import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scipy import interpolate
import subprocess

# load homemade python moduels
import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_raw_data_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
figs_clim_second_derivative = [0.0055, 0.035]
figs_clim_raw = [350, 1250]

lw_dft_thick = 1
lw_dft_thin = 0.3
color_dft = 'gray'

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

# directory paths
outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=150)

#######################################################################################################################
# Plot structure 1 (2deriv and raw ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=15, bottom=1 + 132 - 48, width=100, height=65)
# eframe1.visualize_frame()

# create plot frames
img_left = 12
# img_bottom = 12
img_width = 35  # 35
img_hoffset = 36.75  # 36.75
img_height = 48
img_bottom = 12
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left + img_hoffset, bottom=img_bottom,
                             width=img_width, height=img_height)
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='white',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.3
emax_plot = 0.35
img_filepath1 = './image_processed_2ppe8621_13_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.02)
pframe11.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe8621_13_sc_raw_shift2d_own_n20_v07.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.02)
pframe12.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='inferno', clim=figs_clim_raw, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# load Floquet center frequency
tmp_filepath = './waveform_params_2ppe8621_shift2d_n60_v19.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_parallel = data_waveform_params[5]

# plot DFT band structures
for pframe in [pframe11]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    e_Floquet = hb * 2.0 * np.pi * freq_fit_parallel * 1e-3
    # print(e_Floquet)
    for n_Floquet in [-3, -2, -1, 1, 2, 3]:
        pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)

# load parameters for depicting the integration box
tmp_filepath = './curve_1d_box_params_v11.dat'
data_box_params = np.loadtxt(tmp_filepath, comments=';')
k1_box, e1_box = data_box_params[0:2]
k2_box, e2_box = data_box_params[2:4]
k3_box, e3_box = data_box_params[4:6]
k4_box, e4_box = data_box_params[6:8]

# depict a box for extracting 1D curves
points = [[k1_box, e1_box], [k2_box, e2_box], [k3_box, e3_box], [k4_box, e4_box], [k1_box, e1_box]]
pframe11.ax.add_patch(matplotlib.patches.Polygon(xy=points, closed=True,
                                                 edgecolor='black', facecolor='none', lw=1, ls=(0, (2, 2))))
pframe11.ax.text(0.1, 0.26, 'A', transform=pframe11.ax.transAxes, color='black', fontsize=fontsize_main, ha='center')
pframe11.ax.text(0.52, 0.45, 'B', transform=pframe11.ax.transAxes, color='black', fontsize=fontsize_main, ha='center')

# load EDC extraction parameters
tmp_filepath = './EDC_extraction_params_v11.dat'
data_EDC_params = np.loadtxt(tmp_filepath, comments=';')
kmin_EDC_extract = data_EDC_params[0]
kmax_EDC_extract = data_EDC_params[1]
emin_plot_EDC = data_EDC_params[2]
emax_plot_EDC = data_EDC_params[3]

# draw EDC extraction ranges
pframe11.ax.add_patch(matplotlib.patches.Rectangle(xy=(kmin_EDC_extract, emin_plot_EDC),
                                                   width=kmax_EDC_extract - kmin_EDC_extract,
                                                   height=emax_plot_EDC - emin_plot_EDC,
                                                   facecolor='none', edgecolor='black', lw=1))
# set other attributes
pframe11.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.hlines(0, kmin_plot, kmax_plot, color='black', lw=0.5, ls=(0, (8, 5)))
pframe11.ax.text(0.5, 0.89, r'$t=$ -40 fs', transform=pframe11.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
for pframe in [pframe11, pframe12]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.fix_yaxis_interval(0.1)
pframe12.ax.tick_params(labelleft=False)

# add colorbar for raw ARPES
pframe11.add_colorbar(cleft=75, cbottom=8, cwidth=1.5, cheight=20)
pframe11.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='black', show_ticks=True)
pframe11.cax.text(0.5, 1.05, 'max', transform=pframe11.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe11.cax.text(0.5, -0.12, 'min', transform=pframe11.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe11.cax.tick_params(axis='y', right=False, labelright=False)
# add colorbar for 2deriv ARPES
pframe12.add_colorbar(cleft=45, cbottom=8, cwidth=1.5, cheight=20)
pframe12.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='white', show_ticks=True)
pframe12.cax.text(0.5, 1.05, 'max', transform=pframe12.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe12.cax.text(0.5, -0.12, 'min', transform=pframe12.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe12.cax.tick_params(axis='y', right=False, labelright=False)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (EDC)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=15, bottom=1, width=95, height=85)
# eframe2.visualize_frame()

# create plot frames
pframe2 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                            left=12, bottom=12, width=35 + 36.75, height=(35 + 36.75) * (48 / 50))
pframe2.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                     ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe2.add_right_axis(ticks_fontsize=fontsize_main, ticks_color='black')

# load EDC extraction parameters
tmp_filepath = './EDC_extraction_params_v11.dat'
data_EDC_params = np.loadtxt(tmp_filepath, comments=';')
kmin_EDC_extract = data_EDC_params[0]
kmax_EDC_extract = data_EDC_params[1]
emin_plot_EDC = data_EDC_params[2]
emax_plot_EDC = data_EDC_params[3]

# load EDC analysis data
tmp_filepath = './EDC_extracted_v11.dat'
data_EDC_extracted = np.loadtxt(tmp_filepath, comments=';')
e_vec_EDC = data_EDC_extracted[:, 0]
EDC_raw = data_EDC_extracted[:, 1]
EDC_2deriv = data_EDC_extracted[:, 2]

# load EDC fitting data
tmp_filepath = './EDC_fitting_fixed_v11.dat'
data_EDC_fitting = np.loadtxt(tmp_filepath, comments=';')
e_vec_EDC_fit = data_EDC_fitting[:, 0]
EDC_fit = data_EDC_fitting[:, 1]
EDC_fit_gauss1 = data_EDC_fitting[:, 2]
EDC_fit_gauss2 = data_EDC_fitting[:, 3]
EDC_fit_bg = data_EDC_fitting[:, 4]

# set plot range
ymin_plot = 0.0
ymax_plot = 7
pframe2.ax.set_ylim(ymin_plot, ymax_plot)
pframe2.ax_right.set_ylim(-0.07, 0.6)

# plot data
offset_EDC_plot = -0.65
pframe2.ax.plot(e_vec_EDC, EDC_raw + offset_EDC_plot, color='orangered', lw=1.5)
pframe2.ax.plot(e_vec_EDC_fit, EDC_fit + offset_EDC_plot, color='black', lw=1)
pframe2.ax.plot(e_vec_EDC_fit, EDC_fit_bg + offset_EDC_plot, color='black', lw=0.5, ls=(0, (2, 2)))
pframe2.ax.fill_between(e_vec_EDC_fit, 0, EDC_fit_gauss1, facecolor='black', alpha=0.2)
pframe2.ax.fill_between(e_vec_EDC_fit, 0, EDC_fit_gauss2, facecolor='black', alpha=0.2)
pframe2.ax_right.plot(e_vec_EDC, EDC_2deriv, color='orangered', lw=1, ls=(0, (4, 4)))
pframe2.ax.set_xlim(emin_plot_EDC, emax_plot_EDC)

# set other attributes
pframe2.ax.set_ylabel(r'Intensity (a.u.)', fontsize=fontsize_main, labelpad=6)
pframe2.ax_right.set_ylabel(r'Curvature (a.u.)', fontsize=fontsize_main, labelpad=6)
pframe2.ax.set_xlabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)

# add guides for the energy splitting
e1 = -0.069 + 0.005  # manually adjusted
e2 = 0.03646 + 0.005  # manually adjusted
pframe2.ax.vlines(e1, ymin_plot, ymax_plot, color='gray', lw=3, alpha=0.2)
pframe2.ax.vlines(e2, ymin_plot, ymax_plot, color='gray', lw=3, alpha=0.2)

# depict arrows for splitting
arrow_start_rel = (e1 - emin_plot_EDC) / (emax_plot_EDC - emin_plot_EDC)
arrow_end_rel = (e2 - emin_plot_EDC) / (emax_plot_EDC - emin_plot_EDC)
arrow_vpos_rel = (5.8 - ymin_plot) / (ymax_plot - ymin_plot)
pframe2.ax.arrow(0.5 * (arrow_end_rel + arrow_start_rel), arrow_vpos_rel, 0.5 * (arrow_end_rel - arrow_start_rel), 0,
                 facecolor='black', edgecolor='black', lw=1,
                 head_width=0.025, head_length=0.04, overhang=0.0,
                 length_includes_head=True, clip_on=False, transform=pframe2.ax.transAxes)
pframe2.ax.arrow(0.5 * (arrow_end_rel + arrow_start_rel), arrow_vpos_rel, -0.5 * (arrow_end_rel - arrow_start_rel), 0,
                 facecolor='black', edgecolor='black', lw=1,
                 head_width=0.025, head_length=0.04, overhang=0.0,
                 length_includes_head=True, clip_on=False, transform=pframe2.ax.transAxes)
pframe2.ax.text(0.7, 0.8, '{:.3f} eV'.format(e2 - e1), fontsize=fontsize_main,
                transform=pframe2.ax.transAxes, horizontalalignment='center')

# depict arrows for peak width
idx_peak_gauss1 = np.argmax(EDC_fit_gauss1)
e_peak_gauss1 = e_vec_EDC_fit[idx_peak_gauss1]
peak_gauss1 = EDC_fit_gauss1[idx_peak_gauss1]
# print(peak_gauss1)
idx_e1_gauss1 = np.argmin(np.abs(EDC_fit_gauss1[:idx_peak_gauss1] - 0.5 * peak_gauss1))
idx_e2_gauss1 = np.argmin(np.abs(EDC_fit_gauss1[idx_peak_gauss1:] - 0.5 * peak_gauss1)) + idx_peak_gauss1
e1_gauss1 = e_vec_EDC_fit[idx_e1_gauss1]
e2_gauss1 = e_vec_EDC_fit[idx_e2_gauss1]
# print(e_peak_gauss1, e1_gauss1, e2_gauss1, e2_gauss1 - e1_gauss1)
arrow_start_rel = (e1_gauss1 - emin_plot_EDC) / (emax_plot_EDC - emin_plot_EDC)
arrow_end_rel = (e2_gauss1 - emin_plot_EDC) / (emax_plot_EDC - emin_plot_EDC)
arrow_vpos_rel = (0.5 * peak_gauss1 - ymin_plot) / (ymax_plot - ymin_plot)
pframe2.ax.arrow(0.5 * (arrow_end_rel + arrow_start_rel), arrow_vpos_rel, 0.5 * (arrow_end_rel - arrow_start_rel), 0,
                 facecolor='black', edgecolor='black', lw=1,
                 head_width=0.025, head_length=0.04, overhang=0.0,
                 length_includes_head=True, clip_on=False, transform=pframe2.ax.transAxes)
pframe2.ax.arrow(0.5 * (arrow_end_rel + arrow_start_rel), arrow_vpos_rel, -0.5 * (arrow_end_rel - arrow_start_rel), 0,
                 facecolor='black', edgecolor='black', lw=1,
                 head_width=0.025, head_length=0.04, overhang=0.0,
                 length_includes_head=True, clip_on=False, transform=pframe2.ax.transAxes)
pframe2.ax.text(0.28, 0.06, '{:.3f} eV'.format(e2_gauss1 - e1_gauss1), fontsize=fontsize_main,
                transform=pframe2.ax.transAxes, horizontalalignment='center')
pframe2.fix_xaxis_interval(0.1)

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (1D curves)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=120, bottom=1, width=80, height=149)
# eframe3.visualize_frame()

pframe3 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                            left=15, bottom=12, width=45, height=132)
pframe3.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                     ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

y_offset = 0.025

# depict a guide to the eye using a spline curve
x_guide = [-0.074, -0.083, -0.090, -0.098, -0.098]
y_guide = [0.2, 0.3, 0.4, 0.5, 0.62]
func_spline = interpolate.UnivariateSpline(y_guide, x_guide)
func_spline.set_smoothing_factor(100)
y_guide_spline = np.linspace(0.12, 0.68, 100)
x_guide_spline = func_spline(y_guide_spline)
pframe3.ax.plot(x_guide_spline, y_guide_spline + y_offset, color='orangered', lw=1, ls=(0, (4, 4)))

# load raw 1D curves (averaged in the box and normalized)
list_curve_1d_raw = []
tmp_filepath = './curve_1d_average_norm_v11.dat'
data_curve_1d = np.loadtxt(tmp_filepath, comments=';')
k_vec_curve_1d_plot = data_curve_1d[:, 0]
list_curve_1d_raw.append(data_curve_1d[:, 1])
list_curve_1d_raw.append(data_curve_1d[:, 2])
list_curve_1d_raw.append(data_curve_1d[:, 3])
list_curve_1d_raw.append(data_curve_1d[:, 4])
list_curve_1d_raw.append(data_curve_1d[:, 5])

# load raw 1D curves (averaged in the box and normalized)
list_curve_1d_smooth = []
tmp_filepath = './curve_1d_average_norm_smooth_v11.dat'
data_curve_1d_smooth = np.loadtxt(tmp_filepath, comments=';')
list_curve_1d_smooth.append(data_curve_1d_smooth[:, 1])
list_curve_1d_smooth.append(data_curve_1d_smooth[:, 2])
list_curve_1d_smooth.append(data_curve_1d_smooth[:, 3])
list_curve_1d_smooth.append(data_curve_1d_smooth[:, 4])
list_curve_1d_smooth.append(data_curve_1d_smooth[:, 5])

# set plot range
kmin_plot_1d = k_vec_curve_1d_plot[0]
kmax_plot_1d = k_vec_curve_1d_plot[-1]
pframe3.ax.set_xlim(kmin_plot_1d, kmax_plot_1d)
ymin_plot = 0.0
ymax_plot = 0.74
pframe3.ax.set_ylim(ymin_plot, ymax_plot)

offset_plot_curves = 0.11
factor_plot_curves = 0.23
list_label = ['-60 fs', '-50 fs', '-48 fs', '-40 fs', '-34 fs']
list_peak1 = [-0.074, -0.065, -0.065, -0.068, -0.068]
list_peak2 = [-0.098, -0.083, -0.092, -0.098, -0.098]
list_pos_manual_adjust = [-0.01, 0.0, -0.012, 0.0, 0.0]
for idx_plot in range(len(list_curve_1d_raw)):
    pframe3.ax.plot(k_vec_curve_1d_plot,
                    list_curve_1d_raw[idx_plot] * factor_plot_curves
                    + offset_plot_curves * idx_plot + list_pos_manual_adjust[idx_plot] + y_offset,
                    color='orangered', lw=1.5)
    pframe3.ax.plot(k_vec_curve_1d_plot,
                    list_curve_1d_smooth[idx_plot] * factor_plot_curves
                    + offset_plot_curves * idx_plot + list_pos_manual_adjust[idx_plot] + y_offset,
                    color='black', lw=1)
    idx_peak = np.argmin(np.abs(k_vec_curve_1d_plot - list_peak1[idx_plot]))
    marker_bottom = list_curve_1d_smooth[idx_plot][idx_peak] * factor_plot_curves \
                    + offset_plot_curves * idx_plot + 0.015 + list_pos_manual_adjust[idx_plot] + y_offset
    # depict arrows
    if idx_plot != 0:
        offset_arrow = 0.015
        arrow_start_rel = (marker_bottom + 0.02 - ymin_plot) / (ymax_plot - ymin_plot)
        arrow_end_rel = (marker_bottom - ymin_plot) / (ymax_plot - ymin_plot)
        arrow_hpos_rel = (list_peak1[idx_plot] - kmin_plot_1d) / (kmax_plot_1d - kmin_plot_1d)
        pframe3.ax.arrow(arrow_hpos_rel, arrow_start_rel + offset_arrow,
                         0, arrow_end_rel - arrow_start_rel,
                         facecolor='darkblue', edgecolor='darkblue', lw=1.5,
                         head_width=0.0, head_length=0.0, overhang=0.0,
                         length_includes_head=True, clip_on=False, transform=pframe3.ax.transAxes, zorder=3)
        pframe3.ax.arrow(arrow_hpos_rel + 0.002, arrow_start_rel + offset_arrow - 0.013,
                         0, arrow_end_rel - arrow_start_rel,
                         facecolor='darkblue', edgecolor='none', lw=0,
                         head_width=0.05, head_length=0.025, overhang=0.5,
                         length_includes_head=True, clip_on=False, transform=pframe3.ax.transAxes, zorder=3)
    # put texts
    right_edge = list_curve_1d_smooth[idx_plot][-1] * factor_plot_curves \
                 + offset_plot_curves * idx_plot + list_pos_manual_adjust[idx_plot] + y_offset
    pframe3.ax.text(1.05, (right_edge - ymin_plot) / (ymax_plot - ymin_plot),
                    r'$t= $' + list_label[idx_plot], transform=pframe3.ax.transAxes,
                    fontsize=fontsize_main, color='black')

# set other attributes
pframe3.ax.set_ylabel(r'Intensity (a.u.)', fontsize=fontsize_main, labelpad=4)
pframe3.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe3.fix_yaxis_interval(0.2)
pframe3.ax.text(0.0, 1.01, r'A', fontsize=fontsize_main + 2, ha='center', transform=pframe3.ax.transAxes)
pframe3.ax.text(1.0, 1.01, r'B', fontsize=fontsize_main + 2, ha='center', transform=pframe3.ax.transAxes)

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(15, 145, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(15, 81, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(123, 145, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
