import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

# load homemade python moduels
import absplot_v03 as absplot
import zapes_utilities_v03 as zutil

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_volkov_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.33
emax_plot = 0.32
colormap = 'gnu_v2_r'
figs_clim_second_derivative = [0.005, 0.035]
ticks_color = 'black'

lw_dft_thick = 1.5
lw_dft_thin = 0.4
color_dft = 'gray'
color_dft_projection = 'gray'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

# directory paths
outdir = '.'  # output directory

matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=106)

#######################################################################################################################
# Plot structure 1 (waveforms)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=26, bottom=65, width=150, height=40)
# eframe1.visualize_frame()

# create single plotting frame
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1, left=15, bottom=10, width=55, height=27)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1, left=90, bottom=10, width=55, height=27)
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load waveform files
tmp_filepath = './waveform_extraction_2ppe7176_shift2d_n40_v18.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
factor_uv_deconvolve_perp = 0.8
delay_time = np.real(data_waveform[:, 0])
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_fit_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel * 1e-3
E_perp_deconvolved = np.real(data_waveform[:, 12]) / factor_uv_deconvolve_perp * 1e-3
E_perp_smooth_deconvolved = np.real(data_waveform[:, 14]) / factor_uv_deconvolve_perp * 1e-3
E_perp_fit_deconvolved = np.real(data_waveform[:, 16]) / factor_uv_deconvolve_perp * 1e-3
E_perp_fit_rescaled_deconvolved = np.real(data_waveform[:, 18]) / factor_uv_deconvolve_perp * 1e-3

# set plot range
pframe11.ax.set_xlim(-120, 120)
pframe12.ax.set_xlim(-120, 120)
scale_efield1 = 0.5  # np.max(np.abs(E_parallel_deconvolved)) * 1.25
pframe11.ax.set_ylim(-scale_efield1, scale_efield1)
scale_efield2 = np.max(np.abs(E_perp_deconvolved)) * 1.25
pframe12.ax.set_ylim(-scale_efield2, scale_efield2)

# plot data
pframe11.ax.plot(delay_time, E_parallel_deconvolved, color='black', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe11.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1.5)
pframe12.ax.plot(delay_time, E_perp_deconvolved, color='gray', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe12.ax.plot(delay_time, E_perp_fit_rescaled_deconvolved, color='gray', lw=1.5)
for idx_data in [33, 74, 76, 94, 123]:
    pframe11.ax.vlines(delay_time[idx_data], -scale_efield1, scale_efield1, color='blue', lw=1.5, alpha=0.3)
    pframe12.ax.vlines(delay_time[idx_data], -scale_efield2, scale_efield2, color='blue', lw=1.5, alpha=0.3)

# set other attributes
pframe11.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe11.fix_yaxis_interval(0.5)
pframe12.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe12.ax.set_ylabel(r'$E_z$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe12.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (experimental ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=3.5, bottom=0, width=200, height=65)
# eframe2.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 35
img_hoffset = 37
img_height = 48
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe22 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe22.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe23 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe23.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe24 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe24.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe25 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 4, bottom=img_bottom, width=img_width, height=img_height)
pframe25.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
list_idx_data = [33, 74, 76, 94, 123]
img_filepath1 = './image_processed_2ppe7176_30_sc_2deriv_narrow3_shift2d_n40_v10.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.04)
pframe21.assign_image(image=image_processed1.image,
                      extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe7176_74_sc_2deriv_narrow3_shift2d_n40_v10.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.04)
pframe22.assign_image(image=image_processed2.image,
                      extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe7176_76_sc_2deriv_narrow3_shift2d_n40_v10.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.04)
pframe23.assign_image(image=image_processed3.image,
                      extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2ppe7176_94_sc_2deriv_narrow3_shift2d_n40_v10.png'
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.04)
pframe24.assign_image(image=image_processed4.image,
                      extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath5 = './image_processed_2ppe7176_123_sc_2deriv_narrow3_shift2d_n40_v10.png'
image_processed5 = zutil.ImageProcessed(img_filepath5, ef_shift=0.04)
pframe25.assign_image(image=image_processed5.image,
                      extent=image_processed5.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load waveform files
tmp_filepath = './waveform_extraction_2ppe7176_shift2d_n40_v18.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
delay_time = np.real(data_waveform[:, 0])
tmp_filepath = './waveform_params_2ppe7176_shift2d_n40_v18.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_perp = data_waveform_params[13]

for idx, pframe in enumerate([pframe21, pframe22, pframe23, pframe24, pframe25]):
    pframe.ax.text(0.5, 1.03, r'$t = $' + '{:.0f} fs'.format(delay_time[list_idx_data[idx]]),
                   transform=pframe.ax.transAxes,
                   fontsize=fontsize_main, color=ticks_color, ha='center')

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT band structures
for pframe in [pframe21, pframe22, pframe23, pframe24, pframe25]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    if pframe == pframe21:
        pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)
    elif pframe != pframe25:
        e_Floquet = hb * 2.0 * np.pi * freq_fit_perp * 1e-3
        for n_Floquet in [-3, -2, -1, 1, 2, 3]:
            pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)
            pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)

# set attributes
pframe21.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe21, pframe22, pframe23, pframe24, pframe25]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color, lw=0.5, ls=(0, (8, 5)))
    pframe.fix_yaxis_interval(0.1)
for pframe in [pframe22, pframe23, pframe24, pframe25]:
    pframe.ax.tick_params(labelleft=False)

# add colorbar
pframe25.add_colorbar(cleft=37.5, cbottom=12, cwidth=2, cheight=25)
pframe25.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color, show_ticks=False)
pframe25.cax.text(0.5, 1.05, 'max', transform=pframe25.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe25.cax.text(0.5, -0.13, 'min', transform=pframe25.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################
mainframe.ax.text(26, 101, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(102, 101, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(2, 61, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
