import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from scipy import interpolate, ndimage
import subprocess

# load homemade python moduels
import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_2ppe_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
ticks_color = 'white'
lw_dft_thick = 1.5
lw_dft_thin = 1.0
color_dft = 'gray'
color_dft_projection = 'white'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

# directory paths
outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=101)

#######################################################################################################################
# Plot structure 1 (UV spectra and 2PPE intensity)
#######################################################################################################################

# create empty frame for capsulation
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=1, width=68, height=98)
# eframe1.visualize_frame()

# create multiple plotting frame1
img_left = 12
img_bottom = 9
img_width = 52
img_height = 35
img_voffset = 46
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left, bottom=img_bottom + img_voffset, width=img_width, height=img_height)
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12.add_top_axis(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=1)

# load 2ppe intensity data
filepath1 = './data_2ppe_count_v1.dat'
data_2ppe_count = np.loadtxt(filepath1, comments=';')
wavelength_ps_2ppe = data_2ppe_count[:, 0]
intensity_ps_2ppe = data_2ppe_count[:, 1] * 1e-7

# spectra measurement data
list_norm_intensity_spec = []
for idx in range(15, 32):
    tmp_filepath = './spec2_FLMS156171_00' + str(idx) + '.txt'
    tmp_data = pd.read_csv(tmp_filepath, sep="\t", decimal=",", header=13)
    wavelength_spec = tmp_data.T.values[0]
    tmp_intensity = tmp_data.T.values[1]
    tmp_intensity -= np.min(tmp_intensity[500:600])
    tmp_intensity /= np.max(tmp_intensity[520:690])
    list_norm_intensity_spec.append(tmp_intensity)

filepath_spec_broad_UV = './UV.txt'
tmp_data = np.loadtxt(filepath_spec_broad_UV)
wavelength_spec_broad_UV = tmp_data[:, 0]
norm_intensity_spec_broad_UV = tmp_data[:, 1]
norm_intensity_spec_broad_UV -= np.min(norm_intensity_spec_broad_UV)
norm_intensity_spec_broad_UV /= np.max(norm_intensity_spec_broad_UV)

# prepare energy axis
light_velocity = 299792458  # m/sec unit
hb = 0.65821  # Dirac constant, eV•fs unit
wavelength_plot_spec = np.linspace(375.7, 438, 1000)
frequency_plot_spec = light_velocity / wavelength_plot_spec * 1e-6  # PHz unit
energy_plot_spec = hb * 2.0 * np.pi * frequency_plot_spec

# set plot range
pframe11.ax.set_xlim(378, wavelength_ps_2ppe[-1])
pframe11.ax.set_ylim(0, np.max(intensity_ps_2ppe) * 1.2)
pframe12.ax.set_xlim(wavelength_plot_spec[0], wavelength_plot_spec[-1])
pframe12.ax.set_ylim(0, 1.1*150)
pframe12.ax_top.set_xlim(energy_plot_spec[0], energy_plot_spec[-1])

# calculate a smooth guide to eye
interp_func = interpolate.interp1d(wavelength_ps_2ppe, intensity_ps_2ppe, fill_value='extrapolate')
wavelength_smooth = np.linspace(375, 445, 1000)
intensity_ps_2ppe_smooth_pre = interp_func(wavelength_smooth)
intensity_ps_2ppe_smooth = ndimage.gaussian_filter1d(intensity_ps_2ppe_smooth_pre, sigma=60, mode='nearest')

# plot data
pframe11.ax.plot(wavelength_ps_2ppe, intensity_ps_2ppe,
                 color='blue', markerfacecolor='blue', marker='o', ms=5, lw=0)
pframe11.ax.plot(wavelength_smooth, intensity_ps_2ppe_smooth, color='blue', lw=2)
pframe12.ax.plot(wavelength_spec_broad_UV, norm_intensity_spec_broad_UV*150,
                 color='orange', lw=1)
pframe12.ax.fill_between(wavelength_spec_broad_UV, norm_intensity_spec_broad_UV*150, 0,
                         color='orange', alpha=0.3)
list_wavelength = [380.0, 390.0, 400.0, 396.0, 393.0,
                   386.0, 384.0, 403.0, 406.0, 410.0,
                   413.0, 416.0, 420.0, 423.0, 426.0,
                   430.0, 433.0]  # nm unit, in the order of spectrum measurements (see logbook)
for idx_plot in range(len(list_norm_intensity_spec)):
    target_power = 150.0 * (400.0 / float(list_wavelength[idx_plot]))  # targeted power for constant photon flux
    pframe12.ax.plot(wavelength_spec, np.array(list_norm_intensity_spec[idx_plot], dtype=np.float32) * target_power,
                     color='blue', lw=1)

# set other attributes
pframe11.ax.set_ylabel(r'$I_{\mathrm{2PPE}}$ (a.u.)', fontsize=fontsize_main, labelpad=5)
pframe11.ax.set_xlabel('Wavelength (nm)', fontsize=fontsize_main, labelpad=3)
pframe11.fix_xaxis_interval(20)
pframe12.ax.set_ylabel('Intensity (a.u.)', fontsize=fontsize_main, labelpad=3)
pframe12.ax.set_xlabel('Wavelength (nm)', fontsize=fontsize_main, labelpad=3)
pframe12.ax_top.set_xlabel('Energy (eV)', fontsize=fontsize_main, labelpad=6)
pframe12.fix_xaxis_top_interval(0.1)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (2PPE ARPES images)
#######################################################################################################################

# create empty frame for capsulation
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=66, bottom=1, width=138, height=92)
# eframe2.visualize_frame()

# create multiple plotting frame2
img_left = 12.5
img_bottom = 9
img_width = 34.5
img_hoffset = 42.5
img_height = 35
img_voffset = 46
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom + img_voffset,
                             width=img_width, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe22 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset, bottom=img_bottom + img_voffset,
                             width=img_width, height=img_height)
pframe22.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe23 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 2, bottom=img_bottom + img_voffset,
                             width=img_width, height=img_height)
pframe23.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe24 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom,
                             width=img_width, height=img_height)
pframe24.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe25 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset, bottom=img_bottom,
                             width=img_width, height=img_height)
pframe25.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe26 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left + img_hoffset * 2, bottom=img_bottom,
                             width=img_width, height=img_height)
pframe26.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load zapes3 2ppe image data
list_wavelength_2ppe = np.array([430.0, 420.0, 413.0, 400.0, 393.0, 386.0])
list_data_tag = ['0268', '0267', '0265', '0253', '0256', '0257']
emin_plot = -0.25
emax_plot = 0.07
kmin_plot = -0.1
kmax_plot = 0.1

# prepare energy axis
light_velocity = 299792458  # m/sec unit
def func_convert_wavelength_energy(wavelength_):
    frequency_ = light_velocity / wavelength_ * 1e-6  # PHz unit
    return hb * 2.0 * np.pi * frequency_

# assign image data
for idx, pframe in enumerate([pframe21, pframe22, pframe23, pframe24, pframe25, pframe26]):
    img_filepath1 = './image_processed_2ppe{}_raw_converted_v1.png'.format(list_data_tag[idx])
    image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
    pframe.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                        xmin=kmin_plot, xmax=kmax_plot,
                        ymin=emin_plot, ymax=emax_plot,
                        colormap='inferno', clim=[0.0, 150], transpose=True, inversex=True)

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT band structures
for pframe in [pframe21, pframe22, pframe23, pframe24, pframe25, pframe26]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)
    pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)
    pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
    pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)

# set other attributes
for pframe in [pframe21, pframe24]:
    pframe.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)

for pframe in [pframe21, pframe22, pframe23, pframe24, pframe25, pframe26]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax, color=ticks_color, lw=0.5, ls=(0, (8, 5)))
for idx, pframe in enumerate([pframe21, pframe22, pframe23, pframe24, pframe25, pframe26]):
    pframe.ax.text(0.5, 0.87,
                   '{:d} nm ({:.2f} eV)'.format(int(list_wavelength_2ppe[idx]),
                                                func_convert_wavelength_energy(list_wavelength_2ppe[idx])),
                   transform=pframe.ax.transAxes, fontsize=fontsize_main, color=ticks_color,
                   horizontalalignment='center')
    pframe.fix_xaxis_interval(0.1)

# add colorbar
pframe23.add_colorbar(cleft=37, cbottom=5, cwidth=2, cheight=22)
pframe23.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color=ticks_color, show_ticks=False)
pframe23.cax.text(0.5, 1.05, 'max', transform=pframe23.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')
pframe23.cax.text(0.5, -0.13, 'min', transform=pframe23.cax.transAxes, fontsize=fontsize_colorbar,
                  horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(0.5, 97, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(68, 97, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(0.5, 43, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
