import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

# load homemade python moduels
import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(edf_theory_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# figure parameters
colormap = 'gnu_v2_r'
ticks_color = 'black'
colormap2 = 'inferno'
ticks_color2 = 'white'
freq_Floquet = 25
lw_band_thick = 1.5
lw_band_thin = 0.3

kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.3
emax_plot = 0.35

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

offset_vertical = 47
offset_vertical2 = 8
n_img_panel = 3

# directory paths
outdir = '.'  # output directory

matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=109 + offset_vertical * (n_img_panel - 1) + offset_vertical2)

#######################################################################################################################
# Plot structure 1 (model band structure)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=1,
                             bottom=53 + offset_vertical * (n_img_panel - 1) + offset_vertical2, width=75, height=54)
# eframe1.visualize_frame()

# create single plotting frame
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=11, bottom=11, width=29.5, height=29.5 * (48 / 35))
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=11 + 29.5 * (37 / 35), bottom=11, width=29.5, height=29.5 * (48 / 35))
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# set plot range
for pframe in [pframe11, pframe12]:
    pframe.ax.set_xlim(kmin_plot, kmax_plot)
    pframe.ax.set_ylim(emin_plot, emax_plot)
    band_plot_eoffset = 0.0
    pframe.ax.set_ylim(emin_plot + band_plot_eoffset, emax_plot + band_plot_eoffset)

pframe11.ax.plot(k_model, e_surf11, color='orangered', lw=2, label='surface state')
pframe11.ax.plot(k_model, e_surf12, color='orangered', lw=2)
pframe12.ax.plot(k_model, e_surf11, color='orangered', lw=2, label='surface state')
pframe12.ax.plot(k_model, e_surf12, color='orangered', lw=2)
pframe12.ax.plot(k_model, e_bulk, color='gray', lw=1.5, label='pure bulk state')
pframe12.ax.fill_between(k_model, projection1_bulk, projection2_bulk, color='gray', alpha=0.3, lw=0)
pframe12.ax.fill_between(k_model, projection1_surf, projection2_surf, color='orangered', alpha=0.3, lw=0)

# set other attributes
pframe11.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.hlines(0, kmin_plot, kmax_plot, color='black', lw=0.5, ls=(0, (8, 5)))
pframe11.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe12.ax.tick_params(labelleft=False)
pframe12.ax.hlines(0, kmin_plot, kmax_plot, color='black', lw=0.5, ls=(0, (8, 5)))
pframe12.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe11.ax.text(0.5, 0.9, 'only TSS',
                 transform=pframe11.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe12.ax.text(0.5, 0.9, 'TSS + bulk',
                 transform=pframe12.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (waveform)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=1 + 29.5 * (37 / 35) * 3,
                             bottom=53 + offset_vertical * (n_img_panel - 1) + offset_vertical2,
                             width=105, height=54)
# eframe2.visualize_frame()

# create single plotting frame
pframe2 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                            left=11, bottom=11, width=29.5 * (37 / 35) * 2 + 29.5, height=29.5 * (48 / 35))
pframe2.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                     ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load calc waveform data
calc_script_version = 'v19'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 28.0
params = '_params3'
matel = ''
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
list_delay_time = [-100, -60, -50, -40, -30, -20, -10, 0, 20, 100]
list_idx_plot = [0, 1, 2, 3, 4, 5]
idx_show_probe = 1

t_offset = 0.0
filepath1 = './driving_waveform_Schueler2_{}_{}.dat'.format(analysis_tag, calc_script_version)
data_waveform_plot = np.loadtxt(filepath1, comments=';')
delay_time_plot = data_waveform_plot[:, 0] + t_offset
E_field_plot = data_waveform_plot[:, 1]
filepath2 = './driving_waveform_selected_Schueler2_{}_{}.dat'.format(analysis_tag, calc_script_version)
data_waveform_selected = np.loadtxt(filepath2, comments=';')
delay_time_selected = data_waveform_selected[:, 0] + t_offset
E_field_selected = data_waveform_selected[:, 1]
filepath2 = './probe_envelope_Schueler2_{}_{}.dat'.format(analysis_tag, calc_script_version)
data_probe_envelope = np.loadtxt(filepath2, comments=';')
delay_time_probe = data_probe_envelope[:, 0] + t_offset
probe_envelope = data_probe_envelope[:, 1]

tmin_plot = -130
tmax_plot = 130

# set plot range
pframe2.ax.set_xlim(tmin_plot, tmax_plot)
scale_efield1 = 1.0
pframe2.ax.set_ylim(-scale_efield1, scale_efield1)

# plot data
pframe2.ax.plot(delay_time_plot, E_field_plot, color='black', lw=2)
for idx_data, delay_time in enumerate(list_idx_plot):
    pframe2.ax.vlines(list_delay_time[list_idx_plot[idx_data]], -scale_efield1, scale_efield1, color='blue', lw=1.5,
                      alpha=0.3)
pframe2.ax.plot(delay_time_probe + list_delay_time[idx_show_probe], probe_envelope * 0.6 * scale_efield1,
                color='blue', lw=0.5, alpha=1)
pframe2.ax.fill_between(delay_time_probe + list_delay_time[idx_show_probe], probe_envelope * 0.6 * scale_efield1,
                        color='blue', alpha=0.3)

# set other attributes
pframe2.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe2.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (theoretical ARPES images, sequential 2deriv)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=1,
                             bottom=1 + offset_vertical * (n_img_panel - 1) + offset_vertical2, width=203, height=52)
# eframe3.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 29.5  # 35
img_hoffset = img_width * (37 / 35)  # 36.75
img_height = img_width * (48 / 35)
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe32 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe32.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe33 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe33.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe34 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe34.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe35 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 4, bottom=img_bottom, width=img_width, height=img_height)
pframe35.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe36 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 5, bottom=img_bottom, width=img_width, height=img_height)
pframe36.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
calc_script_version = 'v19'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 28.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.00, 0.08]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
list_delay_time = [-100, -60, -50, -40, -30, -20, -10, 0, 20, 100]
list_idx_plot = [0, 1, 2, 3, 4, 5]
img_filepath1 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[0], calc_script_version)
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe31.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[1], calc_script_version)
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe32.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[2], calc_script_version)
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe33.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[3], calc_script_version)
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.0)
pframe34.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath5 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[4], calc_script_version)
image_processed5 = zutil.ImageProcessed(img_filepath5, ef_shift=0.0)
pframe35.assign_image(image_processed5.image, extent=image_processed5.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath6 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[5], calc_script_version)
image_processed6 = zutil.ImageProcessed(img_filepath6, ef_shift=0.0)
pframe36.assign_image(image_processed6.image, extent=image_processed6.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# plot model band structures
for pframe in [pframe31, pframe32, pframe33, pframe34, pframe35, pframe36]:
    pframe.ax.plot(k_model, e_surf11, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_surf12, color='gray', lw=lw_band_thick)
    if not pframe in [pframe31]:
        e_Floquet = hb * 2.0 * np.pi * freq_Floquet * 1e-3
        for n_Floquet in [-3, -2, -1, 0, 1, 2, 3]:
            pframe.ax.plot(k_model, e_surf11 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)
            pframe.ax.plot(k_model, e_surf12 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)

# set attributes
pframe31.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe32, pframe33, pframe34, pframe35, pframe36]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe31, pframe32, pframe33, pframe34, pframe35, pframe36]:
    # pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color, lw=0.5, ls=(0, (8, 5)))
    # pframe.fix_yaxis_interval(0.1)
for idx, pframe in enumerate([pframe31, pframe32, pframe33, pframe34, pframe35, pframe36]):
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(list_delay_time[list_idx_plot[idx]]),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color='black')
    pframe.fix_yaxis_interval(0.1)

# add colorbar
pframe36.add_colorbar(cleft=img_width * (38 / 35), cbottom=8, cwidth=2, cheight=25)
pframe36.modify_colorbar(ticks_fontsize=fontsize_colorbar, ticks_color=ticks_color, show_ticks=False)
pframe36.cax.text(0.5, 1.04, 'max', transform=pframe36.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe36.cax.text(0.5, -0.1, 'min', transform=pframe36.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Plot structure 4 (theoretical ARPES images, sequential raw)
#######################################################################################################################
# create empty frame for this structure
eframe4 = absplot.EmptyFrame(mainframe=mainframe, left=1,
                             bottom=1 + offset_vertical * (n_img_panel - 2) + offset_vertical2 + 1,
                             width=203, height=52)
# eframe4.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 29.5  # 35
img_hoffset = img_width * (37 / 35)  # 36.75
img_height = img_width * (48 / 35)
pframe41 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe41.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe42 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe42.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe43 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe43.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe44 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe44.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe45 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 4, bottom=img_bottom, width=img_width, height=img_height)
pframe45.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe46 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4,
                             left=img_left + img_hoffset * 5, bottom=img_bottom, width=img_width, height=img_height)
pframe46.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color2,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
calc_script_version = 'v19'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 28.0
params = '_params3'
matel = ''
figs_clim_raw_calc = [0, 500000]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
list_delay_time = [-100, -60, -50, -40, -30, -20, -10, 0, 20, 100]
list_idx_plot = [0, 1, 2, 3, 4, 5]
img_filepath1 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[0], calc_script_version)
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe41.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[1], calc_script_version)
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe42.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[2], calc_script_version)
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe43.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[3], calc_script_version)
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.0)
pframe44.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath5 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[4], calc_script_version)
image_processed5 = zutil.ImageProcessed(img_filepath5, ef_shift=0.0)
pframe45.assign_image(image_processed5.image, extent=image_processed5.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath6 = './image_processed_raw_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[5], calc_script_version)
image_processed6 = zutil.ImageProcessed(img_filepath6, ef_shift=0.0)
pframe46.assign_image(image_processed6.image, extent=image_processed6.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap2, clim=figs_clim_raw_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# plot model band structures
for pframe in [pframe41, pframe42, pframe43, pframe44, pframe45, pframe46]:
    pframe.ax.plot(k_model, e_surf11, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_surf12, color='gray', lw=lw_band_thick)
    if not pframe in [pframe41]:
        e_Floquet = hb * 2.0 * np.pi * freq_Floquet * 1e-3
        for n_Floquet in [-3, -2, -1, 0, 1, 2, 3]:
            pframe.ax.plot(k_model, e_surf11 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)
            pframe.ax.plot(k_model, e_surf12 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)

# set attributes
pframe41.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe42, pframe43, pframe44, pframe45, pframe46]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe41, pframe42, pframe43, pframe44, pframe45, pframe46]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color2, lw=0.5, ls=(0, (8, 5)))
    # pframe.fix_yaxis_interval(0.1)
for idx, pframe in enumerate([pframe41, pframe42, pframe43, pframe44, pframe45, pframe46]):
    pframe.fix_yaxis_interval(0.1)

# add colorbar
pframe46.add_colorbar(cleft=img_width * (38 / 35), cbottom=8, cwidth=2, cheight=25)
pframe46.modify_colorbar(ticks_fontsize=fontsize_colorbar, ticks_color=ticks_color2, show_ticks=False)
pframe46.cax.text(0.5, 1.04, 'max', transform=pframe46.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe46.cax.text(0.5, -0.1, 'min', transform=pframe46.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 4
#######################################################################################################################


#######################################################################################################################
# Plot structure 5 (theoretical ARPES images, bandwidth and pulse form)
#######################################################################################################################
# create empty frame for this structure
eframe5 = absplot.EmptyFrame(mainframe=mainframe, left=1, bottom=1 + offset_vertical * (n_img_panel - 3),
                             width=203, height=54)
# eframe5.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 29.5 * 0.97  # 35
img_hoffset = img_width * (37 / 35)  # 36.75
img_height = img_width * (48 / 35)
offset_additional = 6
pframe51 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe51.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe52 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe52.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe53 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe53.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe54 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe54.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe55 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left + img_hoffset * 4 + offset_additional, bottom=img_bottom, width=img_width,
                             height=img_height)
pframe55.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe56 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe5,
                             left=img_left + img_hoffset * 5 + offset_additional, bottom=img_bottom, width=img_width,
                             height=img_height)
pframe56.modify_ticks(ticks_fontsize=fontsize_main, ticks_color=ticks_color,
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
calc_script_version = 'v15'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 18.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.00, 0.08]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath1 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe51.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe51.ax.text(0.5, 0.89, r'$\Delta E = $72 meV',
                 transform=pframe51.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe51.ax.text(0.5, 0.79, r'sech2',
                 transform=pframe51.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

calc_script_version = 'v15'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 22.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.00, 0.08]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath2 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe52.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe52.ax.text(0.5, 0.89, r'$\Delta E = $59 meV',
                 transform=pframe52.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe52.ax.text(0.5, 0.79, r'sech2',
                 transform=pframe52.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

calc_script_version = 'v19'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 28.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.00, 0.08]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath3 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe53.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe53.ax.text(0.5, 0.89, r'$\Delta E = $47 meV',
                 transform=pframe53.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe53.ax.text(0.5, 0.79, r'sech2',
                 transform=pframe53.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

calc_script_version = 'v15'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 50.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.00, 0.08]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath4 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.0)
pframe54.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe54.ax.text(0.5, 0.89, r'$\Delta E = $26 meV',
                 transform=pframe54.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe54.ax.text(0.5, 0.79, r'sech2',
                 transform=pframe54.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

calc_script_version = 'v13'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 24.0
params = '_params1'
matel = ''
figs_clim_second_derivative_calc = [0.002, 0.1]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath5 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed5 = zutil.ImageProcessed(img_filepath5, ef_shift=0.0)
pframe55.assign_image(image_processed5.image, extent=image_processed5.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe55.ax.text(0.5, 0.89, r'$\Delta E = $24 meV',
                 transform=pframe55.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe55.ax.text(0.5, 0.79, r'Lorentzian',
                 transform=pframe55.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

calc_script_version = 'v17'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 36.0
params = '_params3'
matel = ''
figs_clim_second_derivative_calc = [0.002, 0.1]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
img_filepath6 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, 3, calc_script_version)
image_processed6 = zutil.ImageProcessed(img_filepath6, ef_shift=0.0)
pframe56.assign_image(image_processed6.image, extent=image_processed6.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
pframe56.ax.text(0.5, 0.89, r'$\Delta E = $ 51 meV',
                 transform=pframe56.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe56.ax.text(0.5, 0.79, r'Gaussian',
                 transform=pframe56.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# plot model band structures
for pframe in [pframe51, pframe52, pframe53, pframe54, pframe55, pframe56]:
    pframe.ax.plot(k_model, e_surf11, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_surf12, color='gray', lw=lw_band_thick)
    e_Floquet = hb * 2.0 * np.pi * freq_Floquet * 1e-3
    for n_Floquet in [-3, -2, -1, 0, 1, 2, 3]:
        pframe.ax.plot(k_model, e_surf11 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)
        pframe.ax.plot(k_model, e_surf12 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)

# set attributes
pframe51.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe52, pframe53, pframe54, pframe55, pframe56]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe51, pframe52, pframe53, pframe54, pframe55, pframe56]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color=ticks_color, lw=0.5, ls=(0, (8, 5)))
    # pframe.fix_yaxis_interval(0.1)
for idx, pframe in enumerate([pframe51, pframe52, pframe53, pframe54, pframe55, pframe56]):
    pframe.fix_yaxis_interval(0.1)
pframe51.ax.text(0.5, 1.05, r'$t = $-40 fs',
                 transform=pframe51.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

#######################################################################################################################
# End of Plot structure 5
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(0.5, 104 + offset_vertical * (n_img_panel - 1) + offset_vertical2, 'a',
                  weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(92, 104 + offset_vertical * (n_img_panel - 1) + offset_vertical2, 'b',
                  weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(0.5, 51 + offset_vertical * (n_img_panel - 1) + offset_vertical2, 'c',
                  weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(0.5, 52 + offset_vertical * (n_img_panel - 2) + offset_vertical2, 'd',
                  weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(0.5, 54 + offset_vertical * (n_img_panel - 3), 'e',
                  weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = 'convert {}.png {}.tiff'.format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
