import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import subprocess

import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(fig2_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
t_offset = -4.0
tmin_plot = -120
tmax_plot = 120
figs_clim_second_derivative = [0.0055, 0.035]
figs_clim_kstreaking_2deriv = [0.005, 0.021]

lw_dft_thick = 1.5
lw_dft_thin = 0.3
color_dft = 'gray'
color_dft_projection = 'gray'
alpha_dft_projection = 0.2

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

# directory paths
outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=115)

#######################################################################################################################
# Plot structure 1 (2deriv streaking images)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=65, width=105, height=48)
# eframe1.visualize_frame()

# create plot frames
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1, left=11, bottom=10,
                             width=37*2+35*0.5, height=35)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_kstreaking_2ppe8621_shift2d_n60_v10_streaking.png'
image_processed1 = zutil.ImageProcessedTemporal(img_filepath1, temporal_axis='x', other_axis='momentum')
extent_tmp = image_processed1.get_extent()
k_offset = 0.0
extent = np.array(extent_tmp) + np.array([t_offset, t_offset, k_offset, k_offset])
pframe11.assign_image(image_processed1.image, extent=extent,
                      xmin=tmin_plot, xmax=tmax_plot, ymin=-0.1, ymax=0.1,
                      colormap='gnu_v2_r', clim=figs_clim_kstreaking_2deriv, transpose=False, inversex=False,
                      func_colormap=zutil.util_cmap)

# set attributes
pframe11.ax.set_ylabel(r'$\Delta k_e$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=2)
pframe11.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe11.fix_yaxis_interval(0.1)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (waveform)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=37*3, bottom=65, width=93, height=48)
# eframe2.visualize_frame()

# create single plotting frame
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2, left=11, bottom=10, width=37+35, height=35)
pframe21.add_right_axis()
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# load waveform files
# indices in the original streaking analysis
# these can be checked based on the process_image file names and descriptions in corresponding source files
# list_idx_data = [30, 50, 55, 60, 72]
list_idx_data = [30, 50, 55, 60, 71]  # fine adjustment
tmp_filepath = './waveform_extraction_2ppe8621_shift2d_n60_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel
E_parallel_fit_rescaled_deconvolved = np.real(data_waveform[:, 10]) / factor_uv_deconvolve_parallel
E_parallel_deconvolved *= 1e-3
E_parallel_smooth_deconvolved *= 1e-3
E_parallel_fit_rescaled_deconvolved *= 1e-3

# load current waveform files
tmp_filepath = './current_waveform_extraction_2ppe8621_shift2d_n60_v07.dat'
data_current_waveform = np.loadtxt(tmp_filepath,
                                   dtype=complex,
                                   converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                                   comments=';')
factor_uv_deconvolve_current = 0.8
delay_time_current = np.real(data_current_waveform[:, 0])  # t_offset is already considered in "current_analysis_v07"
current_deconvolved = np.real(data_current_waveform[:, 14]) / factor_uv_deconvolve_current
current_Boltzmann = np.real(data_current_waveform[:, 19]) / factor_uv_deconvolve_current

# set plot range
tmin_plot = -110
tmax_plot = 20
pframe21.ax_right.set_xlim(tmin_plot, tmax_plot)
pframe21.ax.set_xlim(tmin_plot, tmax_plot)
scale_current = 35
pframe21.ax_right.set_ylim(-scale_current, scale_current)
scale_efield = 1.0
pframe21.ax.set_ylim(-scale_efield, scale_efield)

# plot data
color_current = 'orangered'  # (255 / 255, 64 / 255, 128 / 255)
pframe21.ax_right.plot(delay_time_current, current_deconvolved, color=color_current, lw=0, marker='o', ms=4,
                       markerfacecolor='none')
pframe21.ax_right.plot(delay_time_current, current_Boltzmann, color=color_current, lw=1, ls=(0, (1, 1)))
pframe21.ax.plot(delay_time, E_parallel_deconvolved, color='black', lw=0, marker='o', ms=4, markerfacecolor='none')
pframe21.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1)
for idx_data in list_idx_data:
    pframe21.ax.vlines(delay_time[idx_data], -scale_efield, scale_efield, color='blue', lw=1.25, alpha=0.3)

# set other attributes
pframe21.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe21.ax_right.set_ylabel(r'$j_y$ (A cm$^{-1}$)', fontsize=fontsize_main, labelpad=0)
pframe21.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe21.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe21.ax_right.tick_params(axis='y', colors=color_current)
pframe21.ax_right.yaxis.label.set_color(color_current)
pframe21.fix_xaxis_interval(20)

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (experimental ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=0, width=202, height=60)
# eframe3.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 35  # 35
img_hoffset = 37  # 36.75
img_height = 48
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe32 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe32.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe33 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width, height=img_height)
pframe33.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe34 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 3, bottom=img_bottom, width=img_width, height=img_height)
pframe34.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe35 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset * 4, bottom=img_bottom, width=img_width, height=img_height)
pframe35.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
# indices in the original streaking analysis
# these can be checked based on the process_image file names and descriptions in corresponding source files
# list_idx_data = [30, 50, 55, 60, 72]
list_idx_data = [30, 50, 55, 60, 71]  # fine adjustment
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.3
emax_plot = 0.35
img_filepath1 = './image_processed_2ppe8621_0_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.02)
pframe31.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe8621_1_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.02)
pframe32.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe8621_8_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.02)
pframe33.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2ppe8621_13_sc_2deriv_narrow6_shift2d_own_n20_v07.png'
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.02)
pframe34.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath5 = './image_processed_2ppe8621_4_sc_2deriv_narrow8_shift2d_own_n20_v07_3.png'
image_processed5 = zutil.ImageProcessed(img_filepath5, ef_shift=0.02)
pframe35.assign_image(image_processed5.image, extent=image_processed5.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load waveform files
tmp_filepath = './waveform_extraction_2ppe8621_shift2d_n60_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
delay_time = np.real(data_waveform[:, 0]) + t_offset

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# load Floquet center frequency
tmp_filepath = './waveform_params_2ppe8621_shift2d_n60_v19.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_parallel = data_waveform_params[5]

# plot DFT band structures
for pframe in [pframe31, pframe32, pframe33, pframe34, pframe35]:
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    if pframe == pframe31:
        pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2, 
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2, 
                               facecolor=color_dft_projection, alpha=alpha_dft_projection)
        pframe.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)
    else:
        e_Floquet = hb * 2.0 * np.pi * freq_fit_parallel * 1e-3
        # print(e_Floquet)
        for n_Floquet in [-3, -2, -1, 1, 2, 3]:
            pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)
            pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft, lw=lw_dft_thin)

# set attributes
pframe31.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe32, pframe33, pframe34, pframe35]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe31, pframe32, pframe33, pframe34, pframe35]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color='black', lw=0.5, ls=(0, (8, 5)))
for idx, pframe in enumerate([pframe31, pframe32, pframe33, pframe34, pframe35]):
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(delay_time[list_idx_data[idx]]),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color='black')
    pframe.fix_yaxis_interval(0.1)
pframe31.ax.text(0.5, 0.72, 'BCB',
                 transform=pframe31.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe31.ax.text(0.49, 0.38, 'TSS',
                 transform=pframe31.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')
pframe31.ax.text(0.5, 0.07, 'BVB',
                 transform=pframe31.ax.transAxes, horizontalalignment='center',
                 fontsize=fontsize_main, color='black')

# add colorbar
pframe35.add_colorbar(cleft=38, cbottom=8, cwidth=2, cheight=25)
pframe35.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='black', show_ticks=True)
pframe35.cax.text(0.5, 1.05, 'max', transform=pframe35.cax.transAxes, 
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe35.cax.text(0.5, -0.12, 'min', transform=pframe35.cax.transAxes, 
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe35.cax.tick_params(axis='y', right=False, labelright=False)

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(0.5, 109, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(108, 109, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(0.5, 63, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
