import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import PIL
from scipy import interpolate
import subprocess

import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(fig3_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
colormap = 'gnu_v2_r'
ticks_color = 'black'
lw_band_thick = 1.5
lw_band_thin = 0.3

kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.3
emax_plot = 0.35

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

# directory paths
outdir = '.'  # output directory


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=102)

#######################################################################################################################
# Plot structure 1 (schematics based on DFT)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=24, bottom=63, width=80, height=37)
# eframe1.visualize_frame()

img_schematics = np.array(PIL.Image.open('./schematics_duality_v8.png'))
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=0, bottom=0, width=105 * 0.77, height=48 * 0.77)
pframe11.hide_axis()
pframe11.hide_frame()
pframe11.assign_general_image(img_schematics)
color_schematics_text = 'black'
pframe11.ax.text(0.15, 0.8, 'Intraband \nacceleration', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.text(0.835, 0.8, 'Floquet \nsidebands', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.text(0.168, 0.665, r'$ϵ$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.text(0.28, 0.355, r'$k$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
offset_schematics1 = 0.343
pframe11.ax.text(0.168 + offset_schematics1, 0.665, r'$ϵ$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.text(0.28 + offset_schematics1, 0.355, r'$k$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
offset_schematics2 = 0.686
pframe11.ax.text(0.168 + offset_schematics2, 0.665, r'$ϵ$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.text(0.28 + offset_schematics2, 0.355, r'$k$', fontsize=fontsize_main,
                 transform=pframe11.ax.transAxes, horizontalalignment='center', color=color_schematics_text)
pframe11.ax.set_zorder(-1)

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (Floquet cycle map)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=16, bottom=0, width=87, height=61)
# eframe2.visualize_frame()

# create plot frames
img_height = 50.5
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=11, bottom=10, width=73, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)

# assign image data
calc_script_version = 'v23'
offset_phase = 0.0
max_cycle = 6
tmp_filepath = './analytical_model_Floquet_cycle_map_coord' \
               + '_E600kVcm_25THz_offset{}pi_mcycle{}_{}.dat'.format(offset_phase, max_cycle, calc_script_version)
data_cycle_map_coord = np.loadtxt(tmp_filepath, comments=';')
energy_cycle_map = data_cycle_map_coord[:, 1]
n_cycle_array = data_cycle_map_coord[:, 2]
energy_excursion = data_cycle_map_coord[:, 3]
integral_Floquet1 = data_cycle_map_coord[:, 4]
integral_Floquet1_smooth = data_cycle_map_coord[:, 5]
extent = [n_cycle_array[0], n_cycle_array[-1], energy_cycle_map[0], energy_cycle_map[-1]]
tmp_filepath = 'analytical_model_Floquet_cycle_map' \
               + '_E600kVcm_25THz_offset{}pi_mcycle{}_{}.dat'.format(offset_phase, max_cycle, calc_script_version)
img_cycle_map = np.loadtxt(tmp_filepath)
n_cycle_plot_min = 0
n_cycle_plot_max = 4
emin_cycle_plot = -0.27
emax_cycle_plot = 0.27
idx_start = np.argmin(np.abs(energy_cycle_map - 0.05))
idx_cycle_norm = np.argmin(np.abs(n_cycle_array - 4))
norm_factor = np.max(img_cycle_map[idx_cycle_norm, idx_start:])
img_cycle_map /= norm_factor
figs_clim_cycle_map = [0.0, 1.5]
pframe21.assign_image(img_cycle_map, extent=extent,
                      xmin=n_cycle_plot_min, xmax=n_cycle_plot_max, ymin=emin_cycle_plot, ymax=emax_cycle_plot,
                      colormap='gnu_v2_r', clim=figs_clim_cycle_map, transpose=True, inversex=False,
                      func_colormap=zutil.util_cmap)

# set attributes
pframe21.ax.set_xlabel(r'Number of cycles, $N_{\mathrm{cycle}}$', fontsize=fontsize_main, labelpad=3)
pframe21.ax.set_ylabel(r'Energy, $ϵ$ (eV)', fontsize=fontsize_main, labelpad=3)
pframe21.fix_xaxis_interval(1)
pframe21.fix_yaxis_interval(0.1)

# add colorbar
pframe21.add_colorbar(cleft=40, cbottom=5, cwidth=30, cheight=2,
                      orientation='horizontal', ticks_label_xpad=3)
pframe21.modify_colorbar(ticks_fontsize=fontsize_colorbar, ticks_color='black', show_ticks=True)
pframe21.cax.xaxis.set_major_locator(matplotlib.ticker.MultipleLocator(0.5))
# pframe21.cax.xaxis.set_ticks_position('top')
pframe21.cax.set_xlabel(r'$|\Psi(ϵ)|^2/N_{\mathrm{cycle}}$', fontsize=fontsize_colorbar, labelpad=-27)

# add energy excursion waveform
hb_vf = 4.1 * hb  # band gradient, eV•AA unit
hb_omega = 0.10339  # pump photon energy, eV unit, 0.12924 eV = 31.25 THz = 32 fs periodicity
E0 = 6.0e-3  # field strength, V/AA unit = 10^5 kV/cm
offset = 0.0 * np.pi
n_grad = 1000
n_cycle_grad_plot = np.linspace(n_cycle_plot_min, n_cycle_plot_max, n_grad)

alpha_array = (1.0 - integral_Floquet1_smooth) * np.where(n_cycle_array > 2, 0, 1)
alpha_array = alpha_array ** 1.4  # overlap of markers makes the line less transparent, manually compensate this

# interpolate
n_cycle_array_plot = np.linspace(n_cycle_array[0], n_cycle_array[-1], 2000)
interp_func = interpolate.interp1d(n_cycle_array, alpha_array, fill_value='extrapolate')
alpha_array_plot = interp_func(n_cycle_array_plot)
interp_func = interpolate.interp1d(n_cycle_array, energy_excursion, fill_value='extrapolate')
energy_excursion_plot = interp_func(n_cycle_array_plot)

color_excursion = 'black'
pframe21.ax.scatter(n_cycle_array_plot, energy_excursion_plot, s=0.3, color=color_excursion, alpha=alpha_array_plot)

# draw intraband current marker and arrow

def draw_excursion_arrow(n_cycle_pos_, color_, pframe_, alpha_array_):
    idx_pos_ = np.argmin(np.abs(n_cycle_array - n_cycle_pos_))
    dk_arrow_ = 0.01
    list_arrow_ = [(n_cycle_array[idx_pos_] - n_cycle_plot_min) / (n_cycle_plot_max - n_cycle_plot_min),
                   (energy_excursion[idx_pos_] - emin_cycle_plot) / (emax_cycle_plot - emin_cycle_plot),
                   dk_arrow_,
                   dk_arrow_ * ((energy_excursion[idx_pos_ + 1] - energy_excursion[idx_pos_])
                                / (emax_cycle_plot - emin_cycle_plot))
                   / ((n_cycle_array[idx_pos_ + 1] - n_cycle_array[idx_pos_])
                      / (n_cycle_plot_max - n_cycle_plot_min))]
    # alpha_ = alpha_array_[idx_pos_]
    alpha_ = alpha_array_[idx_pos_] ** 0.7
    # print(alpha_)
    pframe_.ax.arrow(list_arrow_[0], list_arrow_[1], list_arrow_[2], list_arrow_[3],
                     facecolor=color_, edgecolor=None, lw=0,
                     head_width=0.04, head_length=0.08, overhang=0.4, alpha=alpha_,
                     length_includes_head=True, clip_on=False, transform=pframe_.ax.transAxes, zorder=2)

draw_excursion_arrow(0.125, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(0.375, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(0.5, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(0.625, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(0.875, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(1.0, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(1.125, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(1.375, color_excursion, pframe21, alpha_array)
draw_excursion_arrow(1.5, color_excursion, pframe21, alpha_array)

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (theoretical ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=111, bottom=0, width=84, height=99)
# eframe3.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 10
img_width = 35 * 0.9  # 35
img_hoffset = 37 * 0.9  # 36.75
img_height = 48 * 0.9
img_woffset = 50 * 0.9
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom + img_woffset, width=img_width, height=img_height)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe32 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset, bottom=img_bottom + img_woffset,
                             width=img_width, height=img_height)
pframe32.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe33 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe33.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)
pframe34 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width, height=img_height)
pframe34.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black',
                      ticks_label_xpad=ticks_xpad, ticks_label_ypad=ticks_ypad)

# assign image data
freq_Floquet = 25
calc_script_version = 'v19'
calc_tag = 'snaphots_trace'
F0 = 800
fwhm = 28.0
params = "_params3"
matel = ""
figs_clim_second_derivative_calc = [0.00, 0.08]
figs_clim_raw_calc = [0, 500000]
analysis_tag = '{}{}_F{}_fwhm{}{}'.format(calc_tag, matel, F0, int(fwhm), params)
list_delay_time = [-100, -60, -50, -40, -30, -20, -10, 0, 20, 100]
list_idx_plot = [1, 2, 3, 5]
img_filepath1 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[0], calc_script_version)
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe31.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[1], calc_script_version)
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe32.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[2], calc_script_version)
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.0)
pframe33.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath4 = './image_processed_2deriv_calc_Schueler_' \
                + '{}_idx{}_{}.png'.format(analysis_tag, list_idx_plot[3], calc_script_version)
image_processed4 = zutil.ImageProcessed(img_filepath4, ef_shift=0.0)
pframe34.assign_image(image_processed4.image, extent=image_processed4.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap=colormap, clim=figs_clim_second_derivative_calc, transpose=True, inversex=True,
                      func_colormap=zutil.util_cmap)

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# plot model band structures
for pframe in [pframe31, pframe32, pframe33, pframe34]:
    pframe.ax.plot(k_model, e_surf11, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_surf12, color='gray', lw=lw_band_thick)
    e_Floquet = hb * 2.0 * np.pi * freq_Floquet * 1e-3
    for n_Floquet in [-3, -2, -1, 0, 1, 2, 3]:
        pframe.ax.plot(k_model, e_surf11 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)
        pframe.ax.plot(k_model, e_surf12 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)

# set attributes
for pframe in [pframe31, pframe33]:
    pframe.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe32, pframe34]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe31, pframe32]:
    pframe.ax.tick_params(labelbottom=False)
for idx, pframe in enumerate([pframe33, pframe34]):
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
for idx, pframe in enumerate([pframe31, pframe32, pframe33, pframe34]):
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color='black', lw=0.5, ls=(0, (8, 5)))
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{:.0f} fs'.format(list_delay_time[list_idx_plot[idx]]),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color='black')
    pframe.fix_yaxis_interval(0.1)

# add colorbar
pframe34.add_colorbar(cleft=38 * 0.9, cbottom=8 * 0.9, cwidth=2 * 0.9, cheight=25 * 0.9)
pframe34.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='black', show_ticks=False)
pframe34.cax.text(0.5, 1.05, 'max', transform=pframe34.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe34.cax.text(0.5, -0.11, 'min', transform=pframe34.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(16, 97, 'a', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(16, 58, 'b', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.text(109.5, 97, 'c', weight='bold', fontsize=fontsize_symbol, color='black')
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
