import os
import re
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scipy import interpolate, ndimage
import subprocess

# load homemade python moduels
import zapes_utilities_v03 as zutil
import absplot_v03 as absplot

# matplotlib backend selection
matplotlib.use('TkAgg')

# file version check
this_file_name = os.path.basename(__file__)
regular_expression = re.compile(r'(fig4_)(\w+)(.py)')
regular_group = regular_expression.match(this_file_name).groups()
script_name = str(regular_group[0])
version = str(regular_group[1])
# print('file version:', version)

# physical constant
hb = 0.65821  # Dirac constant, eV•fs unit
me = 5.68563e-2  # electron mass, eV•fs^2/A^2 unit
kb = 8.61733e-5  # Boltzmann constant, eV/K unit
elem_charge = 1.602176634e-19  # elementary charge, C unit

# global figure parameters
t_offset = -4.0
figs_clim_second_derivative = [0.0055, 0.035]

lw_dft_thick = 1.5
lw_dft_thin = 0.3
color_dft = 'gray'
color_dft_projection = 'gray'
alpha_dft_projection = 0.2

lw_band_thick = 1.5
lw_band_thin = 0.3

tmin_plot = -130
tmax_plot = 130
kmin_plot = -0.15
kmax_plot = 0.15
emin_plot = -0.3
emax_plot = 0.35

fontsize_main = 8
fontsize_colorbar = 7
fontsize_symbol = 9

ticks_xpad = 4
ticks_ypad = 3

outdir = '.'


matplotlib.rcParams['font.family'] = 'Arial'
mainframe = absplot.MainFrame(width=205, height=122)

#######################################################################################################################
# Plot structure 1 (experimental ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe1 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=60, width=120, height=55)
# eframe1.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 9
img_width = 35 * (44 / 48)
img_hoffset = 34
img_height = 44
pframe11 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe11.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe12 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width,
                             height=img_height)
pframe12.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe13 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe1,
                             left=img_left + img_hoffset * 2, bottom=img_bottom, width=img_width,
                             height=img_height)
pframe13.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_2ppe8621_27_sc_2deriv_narrow8_shift2d_own_n20_v07.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.02)
pframe11.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2ppe8621_91_sc_2deriv_narrow9_shift2d_n60_v12.png'
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.02)
pframe12.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath3 = './image_processed_2ppe8621_103_sc_2deriv_narrow9_shift2d_n60_v12.png'
image_processed3 = zutil.ImageProcessed(img_filepath3, ef_shift=0.02)
pframe13.assign_image(image_processed3.image, extent=image_processed3.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# load Floquet center frequency
tmp_filepath = './waveform_params_2ppe8621_shift2d_n60_v19.dat'
data_waveform_params = np.loadtxt(tmp_filepath, comments=';')
freq_fit_parallel = data_waveform_params[5]

# plot DFT calculation data
for idx, pframe in enumerate([pframe11, pframe12, pframe13]):
    pframe.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
    pframe.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
    pframe.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)
    pframe.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                           facecolor=color_dft_projection, alpha=alpha_dft_projection)
    e_Floquet = hb * 2.0 * np.pi * freq_fit_parallel * 1e-3
    # print(e_Floquet)
    for n_Floquet in [-3, -2, -1, 1, 2, 3]:
        pframe.ax.plot(k_plot_dft, band_dft_SS1 + n_Floquet * e_Floquet, color=color_dft,
                       lw=lw_dft_thin)
        pframe.ax.plot(k_plot_dft, band_dft_SS2 + n_Floquet * e_Floquet, color=color_dft,
                       lw=lw_dft_thin)

# set attributes
pframe11.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
for pframe in [pframe12, pframe13]:
    pframe.ax.tick_params(labelleft=False)
list_delay_time = [-12, 22, 46]
for idx, pframe in enumerate([pframe11, pframe12, pframe13]):
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                     color='black', lw=0.5, ls=(0, (8, 5)))
    str_sign = ''
    pframe.ax.text(0.5, 0.9, r'$t = $' + '{}{:.0f} fs'.format(str_sign, list_delay_time[idx]),
                   transform=pframe.ax.transAxes, horizontalalignment='center',
                   fontsize=fontsize_main, color='black')
    pframe.fix_yaxis_interval(0.1)

# add colorbar
pframe13.add_colorbar(cleft=35, cbottom=7, cwidth=2, cheight=24)
pframe13.modify_colorbar(ticks_fontsize=fontsize_main, ticks_color='black', show_ticks=False)
pframe13.cax.text(0.5, 1.04, 'max', transform=pframe13.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')
pframe13.cax.text(0.5, -0.11, 'min', transform=pframe13.cax.transAxes,
                  fontsize=fontsize_colorbar, horizontalalignment='center')

#######################################################################################################################
# End of Plot structure 1
#######################################################################################################################


#######################################################################################################################
# Plot structure 2 (experimental ARPES image, after transient)
#######################################################################################################################
# create empty frame for this structure
eframe2 = absplot.EmptyFrame(mainframe=mainframe, left=0, bottom=0, width=45, height=55)
# eframe2.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 9
img_width = 35 * (44 / 48)
img_height = 44
pframe21 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe2,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe21.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)

# assign image data
img_filepath1 = './image_processed_2ppe8621_11_sc_2deriv_narrow6_shift2d_own_n20_v07_2.png'
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.02)
pframe21.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_second_derivative, transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)

# load integration contours
tmp_filepath = './contour0_definition_2ppe8621_shift2d_n60_v12_wide.dat'
data_contour0 = np.loadtxt(tmp_filepath, comments=';')
contour0_k = data_contour0[:, 2]
contour0_e = data_contour0[:, 3]

pframe21.ax.plot(contour0_k, contour0_e, color='black', lw=0.75, ls=(0, (5, 2)))

# load DFT calculation data
data_dft_plot = np.loadtxt('./dft_plot_data_v1.dat', comments=';')
k_plot_dft = data_dft_plot[:, 0]
band_dft_SS1 = data_dft_plot[:, 1]
band_dft_SS2 = data_dft_plot[:, 2]
projection_dft_VB1 = data_dft_plot[:, 3]
projection_dft_VB2 = data_dft_plot[:, 4]
projection_dft_CB1 = data_dft_plot[:, 5]
projection_dft_CB2 = data_dft_plot[:, 6]
band_dft_VB1 = data_dft_plot[:, 7]
band_dft_VB2 = data_dft_plot[:, 8]
band_dft_VB3 = data_dft_plot[:, 9]
band_dft_VB4 = data_dft_plot[:, 10]
band_dft_CB1 = data_dft_plot[:, 11]
band_dft_CB2 = data_dft_plot[:, 12]
band_dft_CB3 = data_dft_plot[:, 13]
band_dft_CB4 = data_dft_plot[:, 14]

# plot DFT calculatoin data
pframe21.ax.plot(k_plot_dft, band_dft_SS1, color=color_dft, lw=lw_dft_thick)
pframe21.ax.plot(k_plot_dft, band_dft_SS2, color=color_dft, lw=lw_dft_thick)
pframe21.ax.fill_between(k_plot_dft, projection_dft_VB1, projection_dft_VB2,
                         facecolor=color_dft_projection, alpha=alpha_dft_projection)
pframe21.ax.fill_between(k_plot_dft, projection_dft_CB1, projection_dft_CB2,
                         facecolor=color_dft_projection, alpha=alpha_dft_projection)
pframe21.ax.plot(k_plot_dft, band_dft_VB1, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_VB2, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_VB3, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_VB4, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_CB1, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_CB2, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_CB3, color=color_dft, lw=lw_dft_thin)
pframe21.ax.plot(k_plot_dft, band_dft_CB4, color=color_dft, lw=lw_dft_thin)

# set attributes
pframe21.ax.set_ylabel(r'$ϵ-ϵ_{\mathrm{F}}$ (eV)', fontsize=fontsize_main, labelpad=3)
pframe21.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe21.ax.hlines(0, image_processed1.kmin, image_processed1.kmax,
                   color='black', lw=0.5, ls=(0, (8, 5)))
pframe21.ax.text(0.5, 0.9, r'$t = $' + '120 fs', transform=pframe21.ax.transAxes, ha='center',
                 fontsize=fontsize_main, color='black')
pframe21.fix_yaxis_interval(0.1)

#######################################################################################################################
# End of Plot structure 2
#######################################################################################################################


#######################################################################################################################
# Plot structure 3 (theoretical ARPES images)
#######################################################################################################################
# create empty frame for this structure
eframe3 = absplot.EmptyFrame(mainframe=mainframe, left=42, bottom=0, width=80, height=55)
# eframe3.visualize_frame()

# create plot frames
img_left = 11
img_bottom = 9
img_width = 35 * (44 / 48)
img_hoffset = 34
img_height = 44
pframe31 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left, bottom=img_bottom, width=img_width, height=img_height)
pframe31.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe32 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe3,
                             left=img_left + img_hoffset, bottom=img_bottom, width=img_width,
                             height=img_height)
pframe32.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)

# assign image data
freq_Floquet = 25
tag = '3bnd_gkba_mel_snapexp'
calc_script_version = 'sb_v10'
F0 = 800
fwhm = 18
g0_new = 0.0005
vsb_new = 0.1
Ub = 0.2
tag_filter_params = "_params10"
figs_clim_2deriv_calc2 = [0.00, 0.05]
# tag_filter_params = '_params3'
analysis_tag = '_{}_F{}_fwhm{}_g{}_vsb{}_Ub{}{}'.format(tag, F0, fwhm, g0_new, vsb_new, Ub,
                                                        tag_filter_params)
img_filepath1 = './image_processed_2deriv_calc_Schueler{}_idx1_{}.png'.format(analysis_tag, calc_script_version)
image_processed1 = zutil.ImageProcessed(img_filepath1, ef_shift=0.0)
pframe31.assign_image(image_processed1.image, extent=image_processed1.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_2deriv_calc2,
                      transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)
img_filepath2 = './image_processed_2deriv_calc_Schueler{}_idx2_{}.png'.format(analysis_tag, calc_script_version)
image_processed2 = zutil.ImageProcessed(img_filepath2, ef_shift=0.0)
pframe32.assign_image(image_processed2.image, extent=image_processed2.get_extent(),
                      xmin=kmin_plot, xmax=kmax_plot, ymin=emin_plot, ymax=emax_plot,
                      colormap='gnu_v2_r', clim=figs_clim_2deriv_calc2,
                      transpose=True,
                      inversex=True,
                      func_colormap=zutil.util_cmap)

# load model band structure
e_offset_model = 0.0
calc_script_version_band = 'band_v04'
filepath1 = './band_3bnd_{}.dat'.format(calc_script_version_band)
data_model_band = np.loadtxt(filepath1, comments=';')
k_model = data_model_band[:, 0]
e_surf11 = data_model_band[:, 1] + e_offset_model
e_surf12 = data_model_band[:, 2] + e_offset_model
e_bulk = data_model_band[:, 3] + e_offset_model
projection1_bulk = data_model_band[:, 4] + e_offset_model
projection2_bulk = data_model_band[:, 5] + e_offset_model
projection1_surf = data_model_band[:, 6] + e_offset_model
projection2_surf = data_model_band[:, 7] + e_offset_model

# plot model band structures
for pframe in [pframe31, pframe32]:
    pframe.ax.plot(k_model, e_surf11, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_surf12, color='gray', lw=lw_band_thick)
    pframe.ax.plot(k_model, e_bulk, color='gray', lw=lw_band_thick)
    pframe.ax.fill_between(k_model, projection1_bulk, projection2_bulk, color='gray', alpha=0.2, lw=0)
    pframe.ax.fill_between(k_model, projection1_surf, projection2_surf, color='gray', alpha=0.2, lw=0)
    e_Floquet = hb * 2.0 * np.pi * freq_Floquet * 1e-3
    for n_Floquet in [-3, -2, -1, 1, 2, 3]:
        pframe.ax.plot(k_model, e_surf11 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)
        pframe.ax.plot(k_model, e_surf12 + n_Floquet * e_Floquet, color='gray', lw=lw_band_thin)

# set attributes
for pframe in [pframe32]:
    pframe.ax.tick_params(labelleft=False)
for pframe in [pframe31, pframe32]:
    pframe.ax.set_xlabel(r'$k_y$ (Å$^{-1}$)', fontsize=fontsize_main, labelpad=3)
    pframe.ax.hlines(0, image_processed1.kmin, image_processed1.kmax, color='black', lw=0.5,
                     ls=(0, (8, 5)))
    pframe.fix_yaxis_interval(0.1)
pframe31.ax.text(0.5, 0.9, r'$t = $' + '22 fs', transform=pframe31.ax.transAxes,
                 ha='center', fontsize=fontsize_main, color='black')
pframe32.ax.text(0.5, 0.9, r'$t = $' + '46 fs', transform=pframe32.ax.transAxes,
                 ha='center', fontsize=fontsize_main, color='black')

#######################################################################################################################
# End of Plot structure 3
#######################################################################################################################


#######################################################################################################################
# Plot structure 4 (occupations)
#######################################################################################################################
# create empty frame for this structure
eframe4 = absplot.EmptyFrame(mainframe=mainframe, left=124, bottom=0, width=79, height=121)
# eframe4.visualize_frame()

whole_left = 125
whole_width = 57.5
whole_height = 113 - 9
occupation_height = 38
spacing2 = 2
spacing1 = 2
waveform_height = whole_height - (occupation_height * 2 + spacing1 + spacing2)
waveform_bottom_offset = occupation_height * 2 + spacing1 + spacing2
exp_occupation_offset = occupation_height + spacing2

# create single plotting frame
pframe40 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4, left=12,
                             bottom=9 + waveform_bottom_offset, width=whole_width,
                             height=waveform_height)
pframe40.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe40.add_top_axis(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=2)
pframe41 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4, left=12,
                             bottom=9 + exp_occupation_offset, width=whole_width,
                             height=occupation_height)
pframe41.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe41.add_right_axis(ticks_fontsize=fontsize_main, ticks_color='black')
pframe42 = absplot.PlotFrame(mainframe=mainframe, parentframe=eframe4, left=12,
                             bottom=9, width=whole_width, height=occupation_height)
pframe42.modify_ticks(ticks_fontsize=fontsize_main, ticks_color='black', ticks_label_xpad=ticks_xpad,
                      ticks_label_ypad=ticks_ypad)
pframe42.add_right_axis(ticks_fontsize=fontsize_main, ticks_color='black')

# load exp waveform files
tmp_filepath = './waveform_extraction_2ppe8621_shift2d_n60_v19.dat'
data_waveform = np.loadtxt(tmp_filepath,
                           dtype=complex,
                           converters={1: lambda s: complex(s.decode().replace('i', 'j'))},
                           comments=';')
factor_uv_deconvolve_parallel = 0.8
delay_time = np.real(data_waveform[:, 0]) + t_offset
E_parallel_deconvolved = np.real(data_waveform[:, 4]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_smooth_deconvolved = np.real(data_waveform[:, 6]) / factor_uv_deconvolve_parallel * 1e-3
E_parallel_fit_rescaled_deconvolved = np.real(
    data_waveform[:, 10]) / factor_uv_deconvolve_parallel * 1e-3

# load calc occupation data
tag = '3bnd_gkba_mel_dk0.06_steps_T10.0'
calc_script_version = 'pop_v09'
F0 = 800
fwhm = 18
g0_new = 0.0005
vsb_new = 0.1
Ub = 0.2
tag_filter_params = "_params10"
analysis_tag = '_{}_F{}_fwhm{}_g{}_vsb{}_Ub{}'.format(tag, F0, fwhm, g0_new, vsb_new, Ub)
filepath1 = './intensity_interp_Schueler{}_{}.dat'.format(analysis_tag, calc_script_version)
data_intensity_interp = np.loadtxt(filepath1, comments=';')
delay_time_int_interp = data_intensity_interp[:, 0]
int_floq_pure_bg_subt_interp = data_intensity_interp[:, 1]
int_floq_sb_bg_subt_interp = data_intensity_interp[:, 2]
int_bulk_interp = data_intensity_interp[:, 3]
plot_factor = 0.13 / np.max(int_floq_pure_bg_subt_interp)
int_floq_pure_bg_subt_interp *= plot_factor
int_floq_sb_bg_subt_interp *= plot_factor

# load calc occupation data
filepath1 = './occupation_Schueler{}_{}.dat'.format(analysis_tag, calc_script_version)
data_occ = np.loadtxt(filepath1, comments=';')
delay_time_occ = data_occ[:, 0]
occb = data_occ[:, 1]

# load exp intensity data
tmp_filepath = './contour_intensity_bg_subt_2ppe8621_shift2d_n60_v12_wide.dat'
data_exp_intensity = np.loadtxt(tmp_filepath, comments=';')
delay_time_int_exp = data_exp_intensity[:, 0]
bg_int_exp = data_exp_intensity[:, 1]
int_exp_bulk = data_exp_intensity[:, 2]

# interpolation and smoothing
delay_time_int_exp2 = np.linspace(tmin_plot, tmax_plot, 1000)
interp_func = interpolate.interp1d(delay_time_int_exp, int_exp_bulk, fill_value='extrapolate')
int_exp_bulk_interp = interp_func(delay_time_int_exp2)
int_exp_bulk_smooth = ndimage.gaussian_filter1d(int_exp_bulk_interp, sigma=6)

# set plot range
pframe40.ax.set_xlim(tmin_plot, tmax_plot)
pframe40.ax_top.set_xlim(tmin_plot, tmax_plot)
scale_efield1 = 1
pframe40.ax.set_ylim(-scale_efield1, scale_efield1)
pframe41.ax.set_xlim(tmin_plot, tmax_plot)
pframe41.ax_right.set_ylim(0, 7.5)
pframe42.ax.set_xlim(tmin_plot, tmax_plot)
pframe42.ax_right.set_xlim(tmin_plot, tmax_plot)
pframe42.ax.set_ylim(0, np.max(int_floq_pure_bg_subt_interp) * 1.05)
pframe42.ax_right.set_ylim(0, np.max(occb) * 1.2)
pframe42.ax_right.set_ylim(0, 0.399)

# plot data and set attributes
pframe40.ax.plot(delay_time, E_parallel_deconvolved, color='darkgray', lw=0, marker='o', ms=4,
                 markerfacecolor='none', alpha=1)
pframe40.ax.plot(delay_time, E_parallel_fit_rescaled_deconvolved, color='black', lw=1.5)
list_delay_time = [-12, 22, 46, 120]
for delay_time_tmp in list_delay_time:
    pframe40.ax.vlines(delay_time_tmp, -scale_efield1, scale_efield1, color='blue', lw=1.25, alpha=0.3)
pframe40.ax.hlines(0.0, delay_time[0], delay_time[-1], color='black', lw=0.25)
pframe40.ax.tick_params(labelbottom=False)
pframe40.ax_top.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=5)
pframe40.ax.set_ylabel(r'$E_y$ (MV cm$^{-1}$)', fontsize=fontsize_main, labelpad=3)
pframe40.fix_xaxis_top_interval(100)
pframe40.fix_yaxis_interval(1.0)

color2 = 'gray'
pframe41.ax.tick_params(labelbottom=False)
pframe41.ax_right.plot(delay_time_int_exp2, int_exp_bulk_interp, color=color2, lw=2, alpha=0.3)
pframe41.ax_right.plot(delay_time_int_exp2, int_exp_bulk_smooth, color='black', lw=1)
pframe41.ax_right.fill_between(delay_time_int_exp2, 0, int_exp_bulk_interp, facecolor=color2, alpha=0.1)
pframe41.ax_right.set_ylabel(r'$I_{\mathrm{BCB}}$ (a.u.)', fontsize=fontsize_main, labelpad=4)
pframe41.ax.tick_params(labelleft=False)

color1 = 'orangered'
color2 = 'gray'
pframe42.ax.plot(delay_time_int_interp, int_floq_pure_bg_subt_interp, color=color1, lw=1, ls=(0, (2, 2)))
pframe42.ax.plot(delay_time_int_interp, int_floq_sb_bg_subt_interp, color=color1, lw=1)
pframe42.ax_right.plot(delay_time_occ, occb - occb[0], color='gray', lw=1)
pframe42.ax_right.fill_between(delay_time_occ, 0, occb - occb[0], facecolor='gray', alpha=0.1)
pframe42.ax.set_xlabel(r'Delay time, $t$ (fs)', fontsize=fontsize_main, labelpad=3)
pframe42.ax.set_ylabel(r'$n_{\mathrm{Floquet}}$', fontsize=fontsize_main, labelpad=3)
pframe42.ax_right.set_ylabel(r'$n_{\mathrm{BCB}}$', fontsize=fontsize_main, labelpad=3)
pframe42.ax.tick_params(axis='y', colors=color1)
pframe42.ax.yaxis.label.set_color(color1)
pframe42.fix_yaxis_interval(0.05)

#######################################################################################################################
# End of Plot structure 4
#######################################################################################################################


#######################################################################################################################
# Figure symbols
#######################################################################################################################

mainframe.ax.text(1, 118, 'a', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(124, 118, 'b', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(1, 56, 'c', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(45, 56, 'd', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(124, 85, 'e', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.text(124, 44, 'f', weight='bold', fontsize=fontsize_symbol)
mainframe.ax.set_zorder(1)

#######################################################################################################################
# export image files
#######################################################################################################################

plt.savefig(outdir + '/' + script_name + version + '.png', dpi=160)
# plt.savefig(outdir + '/' + script_name + version + '.png', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.jpg', dpi=600)
# # plt.savefig(outdir + '/' + script_name + version + '.tiff', dpi=600)  # this creates a huge TIFF file!
# cmd = "convert {}.png {}.tiff".format(outdir + '/' + script_name + version, outdir + '/' + script_name + version)
# subprocess.call(cmd.split())  # execute the command from this script
# plt.savefig(outdir + '/' + script_name + version + '.pdf', dpi=600)

plt.show()
